#define __STDC_FORMAT_MACROS
#include <climits>
#include <inttypes.h>

#include <sched.h>

#include "base/array.h"
#include "base/assert.h"
#include "base/stlutil.h"
#include "bench/clock.h"
#include "bench/randomgenerator.h"
#include "bench/tpccclient.h"
#include "bench/tpccgenerator.h"
#include "bench/tpcctables.h"
#include "benchmark/benchmark.h"

using std::vector;

static const int CPU_MAP[] = {
// "no mapping"
    //~ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15

// Other core first (best for opterons)
    //~ 0, 4, 8, 12, 2, 6, 10, 14, 1, 5, 9, 13, 3, 7, 11, 15,

// Unshared L2, same core first
    //~ 0, 2, 4, 6, 1, 3, 5, 7, 
    //~ 0, 2, 4, 6, 8, 10, 12, 14, 1, 3, 5, 7, 9, 11, 13, 15

// Same core, unshared L2 first (best for Core2s)
    //~ 0, 2, 1, 3, 4, 6, 5, 7,
    0, 2, 1, 3, 4, 6, 5, 7, 8, 10, 9, 11, 12, 14, 13, 15,
};

// TODO: Unify this with tpcc.cc
class TPCCThread : public benchmark::BenchmarkWorker {
public:
    TPCCThread(int num_warehouses, int id) :
            num_warehouses_(num_warehouses), id_(id), txn_count_(0) {}

    ~TPCCThread() {
        delete client_;
    }

    virtual void initialize() {
#ifdef __linux__
        // bind to the specific CPU
        cpu_set_t cpuset;
        CPU_ZERO(&cpuset);
        CHECK(id_ < base::arraySize(CPU_MAP));
        CPU_SET(CPU_MAP[id_], &cpuset);
        int error = sched_setaffinity(0, sizeof(cpuset), &cpuset);
        ASSERT(error == 0);
#elif defined(__APPLE__)
        // TODO: http://developer.apple.com/releasenotes/Performance/RN-AffinityAPI/
        printf("WARNING: thread affinity not yet implemented\n");
#else
#error Unsupported platform for thread affinity?
#endif

        clock_ = new SystemClock();

        // Create a generator for filling the database.
        random_ = new tpcc::RealRandomGenerator();
        random_->seed((unsigned int) clock_->getMicroseconds());
        tpcc::NURandC cLoad = tpcc::NURandC::makeRandom(random_);
        random_->setC(cLoad);

        // Generate the data
        char now[Clock::DATETIME_SIZE+1];
        clock_->getDateTimestamp(now);
        TPCCGenerator generator(random_, now, Item::NUM_ITEMS, District::NUM_PER_WAREHOUSE,
                Customer::NUM_PER_DISTRICT, NewOrder::INITIAL_NUM_PER_DISTRICT);
        TPCCTables* tables = new TPCCTables();
        generator.makeItemsTable(tables);
        for (int i = 0; i < num_warehouses_; ++i) {
            generator.makeWarehouse(tables, i+1);
        }

        // Change the constants for run
        random_ = new tpcc::RealRandomGenerator();
        random_->setC(tpcc::NURandC::makeRandomForRun(random_, cLoad));

        // Client owns all the parameters
        client_ = new TPCCClient(clock_, random_, tables, Item::NUM_ITEMS, num_warehouses_,
                District::NUM_PER_WAREHOUSE, Customer::NUM_PER_DISTRICT);
    }

    virtual void work(bool measure) {
        client_->doOne();
        if (measure) {
            txn_count_ += 1;
        }
    }

    virtual int result() { return txn_count(); }

    int txn_count() const { return txn_count_; }

private:
    int num_warehouses_;
    int id_;

    int txn_count_;

    SystemClock* clock_;
    tpcc::RealRandomGenerator* random_;

    TPCCClient* client_;
};

static const int WARMUP_SECS = 5;
static const int MEASURE_SECS = 30;

int main(int argc, const char* argv[]) {
    if (argc != 3) {
        fprintf(stderr, "tpcc-scale [warehouses/CPU] [cpus]\n");
        exit(1);
    }

    long num_warehouses = strtol(argv[1], NULL, 10);
    if (num_warehouses == LONG_MIN || num_warehouses == LONG_MAX) {
        fprintf(stderr, "Bad warehouse number (%s)\n", argv[1]);
        exit(1);
    }
    if (num_warehouses <= 0) {
        fprintf(stderr, "Number of warehouses must be > 0 (was %ld)\n", num_warehouses);
        exit(1);
    }
    if (num_warehouses > Warehouse::MAX_WAREHOUSE_ID) {
        fprintf(stderr, "Number of warehouses must be <= %d (was %ld)\n", Warehouse::MAX_WAREHOUSE_ID, num_warehouses);
        exit(1);
    }

    int cpus = atoi(argv[2]);
    ASSERT(cpus > 0);
    CHECK(cpus <= base::arraySize(CPU_MAP));

    vector<TPCCThread*> threads;
    for (int i = 0; i < cpus; ++i) {
        threads.push_back(new TPCCThread((int) num_warehouses, i));
    }

    //~ int64_t us = benchmark::runThreadBench(WARMUP_SECS, MEASURE_SECS, threads);

    //~ int txns = 0;
    //~ for (size_t i = 0; i < threads.size(); ++i) {
        //~ txns += threads[i]->txn_count();
        //~ delete threads[i];
    //~ }

    //~ double seconds = (double) us / 1000000.0;
    //~ printf("%d txns in %lld us = %f txns/s\n", txns, us, txns/seconds);

    double txns_per_sec = benchmark::runForkBench(WARMUP_SECS, MEASURE_SECS, threads);
    printf("%f txns/s\n", txns_per_sec);
    STLDeleteElements(&threads);

    return 0;
}
