#include "benchmark/benchmark.h"

#include <cassert>
#include <tr1/functional>

#include <unistd.h>

#include "base/function_thread.h"
#include "base/mutex.h"

using std::vector;

namespace {

using benchmark::BenchmarkWorker;

/** Contains the state for executing performance tests with multiple threads.
This effectively manages the state machine for the test. */
class ThreadBench {
public:
    ThreadBench(int num_threads) :
            num_threads_(num_threads),
            state_(INVALID),
            cond_(&mu_),
            ready_count_(0) {
        assert(num_threads_ > 0);
    }
    
    enum State {
        INVALID,
        WARMUP,
        MEASURE,
        DONE,
        EXIT,
    };

    State state() const {
        // See stateChange about volatile
        const volatile State* s = &state_;
        return *s;
    }

    /** Wait for num_threads to call this. Returns once all the threads have entered. */
    void blockForStart() {
        mu_.lock();
        assert(state_ == INVALID);
        assert(0 <= ready_count_ && ready_count_ < num_threads_);
        ready_count_ += 1;
        if (ready_count_ == num_threads_) {
            // we are the last one to enter! Time to exit.
            stateChange(WARMUP);
            cond_.broadcast();
        } else {
            // Wait for the broadcast
            while (ready_count_ < num_threads_) {
                cond_.wait();
            }
        }
        assert(state_ == WARMUP);
        assert(ready_count_ == num_threads_);
        mu_.unlock();
    }

    void startMeasure() {
        assert(state_ == WARMUP);
        stateChange(MEASURE);
    }

    void doneMeasure() {
        assert(state_ == MEASURE);
        stateChange(DONE);
        // The master thread must also signal that it is done
        signalDone();
    }

    /** Notify that this thread has entered the done state. */
    void signalDone() {
        mu_.lock();
        assert(state_ == DONE);
        assert(0 < ready_count_ && ready_count_ <= num_threads_);
        ready_count_ -= 1;
        if (ready_count_ == 0) {
            // We are the last thread to observe the test finishing
            stateChange(EXIT);
        }
        mu_.unlock();
    }

private:
    const int num_threads_;
    State state_;

    void stateChange(State next_state) {
        // Cast to volatile because we want all reads/writes to actually go to memory
        // TODO: On x86 this should be fine, but it may need memory fences on
        // other platforms to unsure atomic uncached reads and writes.
        //~ printf("changing state %d\n", next_state);
        volatile State* s = &state_;
        *s = next_state;
    }

    // The following members are protected by the mutex.
    Mutex mu_;
    Condition cond_;
    int ready_count_;
};

class ThreadBenchThread {
public:
    ThreadBenchThread(ThreadBench* test_state, BenchmarkWorker* work) :
            test_state_(test_state),
            work_(work),
            // Spawn the thread
            thread_(std::tr1::bind(&ThreadBenchThread::run, this)) {
        assert(test_state_ != NULL);
        assert(work_ != NULL);
    }

    void join() {
        thread_.join();
    }

private:
    void run() {
        // initialize in parallel
        work_->initialize();

        // wait for everyone to start running
        test_state_->blockForStart();

        bool seen_done = false;

        ThreadBench::State state = test_state_->state();
        while (state != ThreadBench::EXIT) {
            if (state == ThreadBench::DONE && !seen_done) {
                // This is the first time we have observed that the test is done
                // notify the global test state, then continue applying load
                seen_done = true;
                test_state_->signalDone();
            }

            // apply load
            work_->work(state == ThreadBench::MEASURE);
            state = test_state_->state();
        }

        // Clean up in the thread
        work_->finish();
    }

    ThreadBench* test_state_;
    BenchmarkWorker* work_;
    FunctionThread thread_;
};

}  // namespace

namespace benchmark {

/** @return the number of microseconds the measurement period ran for. */
int64_t runThreadBench(
        int warmup_secs,
        int measure_secs,
        const vector<BenchmarkWorker*>& workers) {
    // + 1 to include this "master" thread
    ThreadBench test(static_cast<int>(workers.size() + 1));

    vector<ThreadBenchThread*> threads_;
    for (size_t i = 0; i < workers.size(); ++i) {
        threads_.push_back(new ThreadBenchThread(&test, workers[i]));
    }

    // Wait for all the threads to start
    test.blockForStart();

    // Running: wait for the warm up period
    sleep(warmup_secs);

    // Run the measurement period
    //~ printf("measure begin\n");
    struct timeval start;
    int error = gettimeofday(&start, NULL);
    assert(error == 0);

    test.startMeasure();
    sleep(measure_secs);
    test.doneMeasure();
    //~ printf("measure end\n");

    struct timeval end;
    error = gettimeofday(&end, NULL);
    assert(error == 0);

    // Wait for the workers to finish
    for (size_t i = 0; i < threads_.size(); ++i) {
        threads_[i]->join();
        delete threads_[i];
    }

    return end.tv_usec - start.tv_usec + (end.tv_sec - start.tv_sec) * 1000000;
}

}  // namespace benchmark
