#ifndef BUFFER_H__
#define BUFFER_H__

#include <cassert>
#include <cstring>
#include <vector>

#include "base/cast.h"

// An interface for input streams of data.
class InputStream {
public:
    virtual ~InputStream() {}

    // Read length bytes of data from the stream into buffer. Returns the number of bytes read.
    // If it returns 0, the stream is closed and there is no more data. If it returns -1, the
    // stream is non-blocking and there is no data available.
    virtual int read(char* buffer, size_t length) = 0;
};

// Creates an InputStream from a buffer in memory. It does not own the buffer.
class MemoryInputStream : public InputStream {
public:
    MemoryInputStream(const char* buffer, int size) : buffer_(buffer), size_(size), current_(0) {
        assert(buffer != 0);
        assert(size > 0);
    }

    virtual int read(char* output, size_t length) {
        assert(output != NULL);
        assert(length > 0);

        int copy_bytes = std::min(assert_range_cast<int>(length), size_ - current_);
        memcpy(output, buffer_+current_, copy_bytes);
        current_ += copy_bytes;
        assert(current_ <= size_);
        return copy_bytes;
    }

private:
    const char* buffer_;
    int size_;
    int current_;
};

class ReadBuffer {
public:
    ReadBuffer();
    ~ReadBuffer();

    // Returns true if there is more data. Returns false if the connection is closed.
    bool readAll(InputStream* connection);

    int available() const { return available_; }

    int read(char* buffer, size_t length);

    // Discard all the data in the buffer.
    void clear();

    // read 16 kB blocks at a time
    static const int BLOCK_SIZE = 1 << 14;

private:
    enum ReadStatus {
        // Read ok, there is no more data
        READ_OK_NO_MORE,
        // Read ok, there might be more data
        READ_OK_MAYBE_MORE,
        // Connection is closed/file is at EOF
        READ_DONE,
    };

    ReadStatus readOnce(InputStream* connection);

    struct Block {
        Block() : filled(0) {}

        int filled;
        char data[BLOCK_SIZE];
    };

    // TODO: This should be changed to use a circular array buffer
    // TODO: Does it help if socket buffers are not length prefixed?
    std::vector<Block*> blocks_;
    int read_position_;
    int available_;
};

#endif
