#include <cstdlib>

#include "messageconnection.h"
#include "stupidunit/stupidunit.h"

using std::string;
using std::vector;

class MemoryInputStreamTest : public Test {
public:
    MemoryInputStreamTest();

protected:
    static const char data_[];
    char buffer_[256];

    MemoryInputStream stream_;
};
const char MemoryInputStreamTest::data_[] = "hello world";
MemoryInputStreamTest::MemoryInputStreamTest() : stream_(data_, sizeof(data_)) {}

TEST_F(MemoryInputStreamTest, CreateDeath) {
    EXPECT_DEATH(MemoryInputStream(NULL, 42));
    EXPECT_DEATH(MemoryInputStream(buffer_, -1));
    EXPECT_DEATH(MemoryInputStream(buffer_, 0));
}

TEST_F(MemoryInputStreamTest, ReadBadParameters) {
    EXPECT_DEATH(stream_.read(NULL, 42));
    EXPECT_DEATH(stream_.read(buffer_, 0));
}

TEST_F(MemoryInputStreamTest, Simple) {
    int bytes = stream_.read(buffer_, 1);
    EXPECT_EQ(1, bytes);
    EXPECT_EQ(data_[0], buffer_[0]);

    bytes = stream_.read(buffer_, sizeof(buffer_));
    EXPECT_EQ(sizeof(data_) - 1, bytes);
    EXPECT_EQ(0, memcmp(data_ + 1, buffer_, bytes));

    EXPECT_EQ(0, stream_.read(buffer_, sizeof(buffer_)));
    EXPECT_EQ(0, stream_.read(buffer_, sizeof(buffer_)));
};

class MockInputStream : public InputStream {
public:
    MockInputStream() : read_count_(0), available_(0) {}

    virtual int read(char* buffer, size_t length) {
        read_count_ += 1;

        int bytes_read = std::min(available_, assert_range_cast<int>(length));
        available_ -= bytes_read;
        assert(available_ >= 0);

        if (bytes_read == 0) bytes_read = -1;
        return bytes_read;
    }

    void addFakeData(int amount) {
        available_ += amount;
    }

    int read_count_;

private:
    int available_;
};

class ReadBufferTest : public Test {
public:
    ReadBuffer buffer_;
    MockInputStream input_;
    char bytes_[ReadBuffer::BLOCK_SIZE];
};

TEST_F(ReadBufferTest, Simple) {
    input_.addFakeData(ReadBuffer::BLOCK_SIZE - 1);

    bool result = buffer_.readAll(&input_);
    EXPECT_EQ(true, result);

    int count = buffer_.read(bytes_, sizeof(bytes_));
    EXPECT_EQ(ReadBuffer::BLOCK_SIZE - 1, count);
}

TEST_F(ReadBufferTest, CrossBlockRead) {
    input_.addFakeData(ReadBuffer::BLOCK_SIZE*2);

    bool result = buffer_.readAll(&input_);
    EXPECT_EQ(true, result);

    int count = buffer_.read(bytes_, sizeof(bytes_)-1);
    EXPECT_EQ(sizeof(bytes_)-1, count);
    count = buffer_.read(bytes_, 2);
    EXPECT_EQ(2, count);
}

TEST_F(ReadBufferTest, NoMoreDataRead) {
    input_.addFakeData(ReadBuffer::BLOCK_SIZE + 100);

    bool result = buffer_.readAll(&input_);
    EXPECT_EQ(true, result);

    // Call read twice: once per block
    EXPECT_EQ(2, input_.read_count_);
    EXPECT_EQ(ReadBuffer::BLOCK_SIZE + 100, buffer_.available());

    // Fill the 2nd block
    input_.addFakeData(ReadBuffer::BLOCK_SIZE - 100);
    result = buffer_.readAll(&input_);
    EXPECT_EQ(true, result);

    // Calls read twice: once to fill the block, again to check for more data
    EXPECT_EQ(4, input_.read_count_);
    EXPECT_EQ(ReadBuffer::BLOCK_SIZE * 2, buffer_.available());

    input_.addFakeData(100);
    result = buffer_.readAll(&input_);
    EXPECT_EQ(true, result);

    // Calls read once
    EXPECT_EQ(5, input_.read_count_);
    EXPECT_EQ(ReadBuffer::BLOCK_SIZE * 2 + 100, buffer_.available());
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
