#ifndef DTXNDISTRIBUTOR_H
#define DTXNDISTRIBUTOR_H

#include <cassert>
#include <string>
#include <vector>
#include <stdint.h>
#include <stdint.h>

#include "dtxn/executionengine.h"  // ExecutionEngine::Status

namespace dtxn {

/** Records the state of each partition that may or may not be involved in a transaction. */
class PartitionState {
public:
    PartitionState() : status_(0) {}

    bool involved() const { return status_ & INVOLVED; }
    bool done() const { return status_ & DONE; }

    void setInvolved() {
        assert(!done());
        status_ |= INVOLVED;
    }

    void setDone() {
        assert(!done());
        status_ |= DONE;
    }

private:
    // true = have done work for the txn
    static const int8_t INVOLVED = 0x1;
    // true = have received the last work unit for the txn
    static const int8_t DONE = 0x2;

    int8_t status_;
};

/** Tracks information about a transaction that is being coordinated by the distributor. */
class CoordinatedTransaction {
public:
    CoordinatedTransaction(size_t num_partitions, const std::string& initial_client_message) :
            status_(ExecutionEngine::OK),
            partition_status_(num_partitions),
            received_messages_(1, std::make_pair(CLIENT_PARTITION, initial_client_message)),
            multiple_partitions_(false),
            state_(NULL) {
        assert(partition_status_.size() > 0);
    }

    // Special index for the client partition.
    static const int CLIENT_PARTITION = -1;

    typedef std::vector<std::pair<int, std::string> > MessageList;

    /** Returns the list of messages received in the last round. */ 
    const MessageList& received() const { return received_messages_; }

    /** Returns the list of messages sent in the last round. */
    const MessageList& sent() const { return sent_messages_; }

    void setMultiplePartitions() {
        assert(!allDone() || sent_messages_.size() > 1);
        assert(!multiple_partitions_);
        multiple_partitions_ = true;
    }

    bool multiple_partitions() const {
        return multiple_partitions_;
    }

    void sentMessages() {
        assert(!sent_messages_.empty());
        received_messages_.clear();
    }

    void setDone(int index) {
        assert(0 <= index && index < partition_status_.size());
        partition_status_[index].setDone();
    }
    
    bool isDone(int index) const {
        assert(0 <= index && index < partition_status_.size());
        return partition_status_[index].done();
    }

    bool allDone() const {
        for (int i = 0; i < partition_status_.size(); ++i) {
            if (!partition_status_[i].done()) {
                return false;
            }
        }
        return true;
    }

    // Returns true if the coordinator has provided the final client response.
    bool haveClientResponse() const {
        if (sent_messages_.size() == 1 && sent_messages_[0].first == CLIENT_PARTITION) {
            assert(allDone());
            return true;
        }
        return false;
    }

    void setAllDone() {
        for (int i = 0; i < partition_status_.size(); ++i) {
            if (!partition_status_[i].done()) {
                partition_status_[i].setDone();
            }
        }
    }

    ExecutionEngine::Status status() const { return status_; }

    void commit(const std::string& client_response) {
        // NOTE: client_response might be empty if transaction has no results (just commit/abort)
        assert(sent_messages_.empty());
        assert(status_ == ExecutionEngine::OK);
        setAllDone();
        sent_messages_.push_back(std::make_pair(CLIENT_PARTITION, client_response));
    }

    void abort(const std::string& client_response) {
        // Can only abort or commit once, but we can abort while sending or receiving
        assert(!haveClientResponse());
        assert(status_ != ExecutionEngine::ABORT_DEADLOCK || client_response.empty());
        setAllDone();
        sent_messages_.clear();
        sent_messages_.push_back(std::make_pair(CLIENT_PARTITION, client_response));
        if (status_ == ExecutionEngine::OK) status_ = ExecutionEngine::ABORT_USER;
    }
    
    void send(int index, const std::string& message) {
        assert(0 <= index && index < partition_status_.size());
        assert(!message.empty());
        assert(!containsKey(sent_messages_, index));
        assert(!received_messages_.empty());
        assert(!partition_status_[index].done());
        assert(status_ == ExecutionEngine::OK);
        partition_status_[index].setInvolved();
        sent_messages_.push_back(std::make_pair(index, message));
    }
    
    void receive(int index, const std::string& message, ExecutionEngine::Status status) {
        assert(0 <= index && index < partition_status_.size());
        assert(!containsKey(received_messages_, index));
        assert(containsKey(sent_messages_, index));
        assert(status != ExecutionEngine::ABORT_DEADLOCK || message.empty());
        assert(status == ExecutionEngine::OK || status == ExecutionEngine::ABORT_USER ||
                status == ExecutionEngine::ABORT_DEADLOCK);

        // Priority, from lowest to highest: ExecutionEngine::OK, ExecutionEngine::ABORT_USER,
        // ExecutionEngine::ABORT_DEADLOCK
        if ((status_ == ExecutionEngine::OK && status != ExecutionEngine::OK) ||
                (status_ == ExecutionEngine::ABORT_USER &&
                status == ExecutionEngine::ABORT_DEADLOCK)) {
            // forget any messages we have received
            for (size_t i = 0; i < received_messages_.size(); ++i) {
                received_messages_[i].second.clear();
            }
            status_ = status;
        }

        // Keep the message only if the status matches our current status
        received_messages_.push_back(std::make_pair(index, std::string()));        
        if (status_ == status) {
            received_messages_.back().second = message;
        }

        // We have received all the messages: prepare for the next messages
        if (received_messages_.size() == sent_messages_.size()) {
            sent_messages_.clear();
        } else {
            assert(received_messages_.size() <= sent_messages_.size());
        }
    }

    // TODO: This might return true at the wrong times I think.
    bool receivedAll() const { return sent_messages_.empty(); }

    int numInvolved() const {
        int result = 0;
        for (int i = 0; i < partition_status_.size(); ++i) {
            result += partition_status_[i].involved();
        }
        return result;
    }

    std::vector<int> involvedPartitions() const {
        std::vector<int> result;
        for (int i = 0; i < partition_status_.size(); ++i) {
            if (partition_status_[i].involved()) result.push_back(i);
        }
        return result;
    }

    void state(void* next_state) { state_ = next_state; }
    void* state() { return state_; }

private:
    bool containsKey(const MessageList& messages, int index) const {
        for (int i = 0; i < messages.size(); ++i) {
            if (messages[i].first == index) return true;
        }
        return false;
    }

    // The current status of the transaction
    ExecutionEngine::Status status_;

    // Stores the status of each individual partition.
    std::vector<PartitionState> partition_status_;

    MessageList received_messages_;

    // Next round of messages to be sent out
    MessageList sent_messages_;

    bool multiple_partitions_;

    // Opaque pointer for use by the Coordinator.
    void* state_;
};

/** Interface for the transaction coordinator. This decides how to distribute transactions to
partitions of the database. */
class Coordinator {
public:
    virtual ~Coordinator() {}

    // A new transaction has been received from the client. The coordinator must supply the messages
    // to go out.
    virtual void begin(CoordinatedTransaction* transaction) = 0;

    // Responses for all the previous messages have been received. The coordinaor must supply the
    // next messages.
    // TODO: This is currently *not* called if a remote partition aborts. This does not work for
    // TPC-C new order, for example. In that case, an aborted transaction still returns data!
    virtual void nextRound(CoordinatedTransaction* transaction) = 0;

    // The transaction has completed, so the coordinator should perform any necessary clean-up.
    virtual void done(CoordinatedTransaction* transaction) = 0;
};

}  // namespace dtxn

#endif
