#include "dtxn/dtxnserver.h"

#include <cstdlib>
#include <string>

#include "base/stlutil.h"
#include "dtxn/executionengine.h"
#include "dtxn/messages.h"
#include "dtxn/transaction.h"
#include "dtxn/transactionscheduler.h"
#include "io/eventloop.h"
#include "net/messageserver.h"
#include "replication/faulttolerantlog.h"

using std::make_pair;
using std::pair;
using std::string;
using std::vector;

namespace dtxn {

class Transaction;

/** Stores the state of the transaction that the server must track. */
class ServerTransactionState {
public:
    ServerTransactionState(net::ConnectionHandle* client, int32_t id)
            : client_(client), id_(id), log_sequence_(INVALID_LOG_SEQUENCE), replicated_(false) {}

    net::ConnectionHandle* client() { return client_; }
    int32_t id() { return id_; }

    void log_sequence(int sequence) {
        assert(log_sequence_ == INVALID_LOG_SEQUENCE);
        assert(sequence >= 0);
        log_sequence_ = sequence;
    }
    int log_sequence() const { return log_sequence_; }

    void setReplicated() {
        assert(!replicated_);
        replicated_ = true;
    }
    bool replicated() const { return replicated_; }

    static const int INVALID_LOG_SEQUENCE = -1;

private:
    net::ConnectionHandle* client_;
    int32_t id_;
    int log_sequence_;
    bool replicated_;
};

DtxnServer::DtxnServer(TransactionScheduler* scheduler, io::EventLoop* event_loop,
        net::MessageServer* msg_server, replication::FaultTolerantLog* log) :

        scheduler_(scheduler), event_loop_(event_loop), msg_server_(msg_server), log_(log) {
    assert(scheduler_ != NULL);
    assert(event_loop_ != NULL);
    msg_server_->addCallback(&DtxnServer::fragmentReceived, this);
    msg_server_->addCallback(&DtxnServer::decisionReceived, this);
    event_loop_->addIdleCallback(&idleCallback, this);

    if (log_ != NULL) {
        log_->entry_callback(this);
    }
}

DtxnServer::~DtxnServer() {
    // Unregister message callbacks
    msg_server_->removeCallback(&DtxnServer::fragmentReceived);
    msg_server_->removeCallback(&DtxnServer::decisionReceived);

    // Unregister the idle callback so it does not get called if the event loop is still used
    event_loop_->removeIdleCallback(&idleCallback);

    for (TransactionIdMap::iterator it = transaction_id_map_.begin();
            it != transaction_id_map_.end(); ++it) {
        Transaction* txn = it->second;
        ServerTransactionState* state = (ServerTransactionState*) txn->external_state();
        delete state;
        delete txn;
    }

    STLDeleteElements(&backup_undecided_);

    delete scheduler_;

    // TODO: If we exit while there are transactions pending in the fault tolerant log we will
    // leak them, because we are relying on the callback to "find" them again. Do we care?
}

void DtxnServer::fragmentReceived(
        net::ConnectionHandle* connection, const Fragment& fragment) {
    // Look up or create the transaction state
    pair<net::ConnectionHandle*, int> key(connection, fragment.id);
    pair<TransactionIdMap::iterator, bool> result = transaction_id_map_.insert(
            make_pair(key, (Transaction*) NULL));
    if (result.second) {
        // We did an insert: this is a new transactioon
        assert(result.first->second == NULL);

        // Start a new transaction
        ServerTransactionState* state = new ServerTransactionState(connection, fragment.id);

        Transaction* transaction = scheduler_->begin(state, fragment.transaction,
                fragment.last_fragment, fragment.multiple_partitions);

        // register this transaction
        result.first->second = transaction;
    } else {
        // This had better be another message for an in-progress multiple partition transaction
        assert(fragment.multiple_partitions);
        assert(result.first->second != NULL);
        Transaction* transaction = result.first->second;
        transaction->addWorkUnit(fragment.transaction, fragment.last_fragment);
        // tell the scheduler we have some more work
        scheduler_->workUnitAdded(transaction);
    }

    // try to send results
    // TODO: Only call this occasionally?
    pollForResults();
}

void DtxnServer::decisionReceived(
        net::ConnectionHandle* connection, const CommitDecision& decision) {
    // Look up the transaction state
    pair<net::ConnectionHandle*, int> key(connection, decision.id);
    TransactionIdMap::iterator it = transaction_id_map_.find(key);
    assert(it != transaction_id_map_.end());

    Transaction* transaction = it->second;
    assert(transaction->multiple_partitions());
    assert(transaction->result_sent() || !decision.commit);

    // If we replicated the transaction, replicate the decision. 
    ServerTransactionState* state = reinterpret_cast<ServerTransactionState*>(
            transaction->external_state());
    if (log_ != NULL && state->log_sequence() != ServerTransactionState::INVALID_LOG_SEQUENCE) {
        LogEntry entry;
        entry.multiple_partitions = true;
        entry.decision_log_entry = state->log_sequence();
        entry.commit = decision.commit;
        log_->submit(entry, NULL);

        if (!state->replicated()) {
            // Replication is still pending: this is a timeout abort
            // Set the callback arg to NULL so we know this transaction has aborted
            assert(!decision.commit);
            log_->setCallbackArg(state->log_sequence(), NULL);
        }
    }

    finishTransaction(transaction, decision.commit);
    pollForResults();
}

void DtxnServer::nextLogEntry(int sequence, const std::string& serialized_entry, void* argument) {
    Transaction* transaction = (Transaction*) argument;
    assert(log_ != NULL);
    LogEntry* entry = new LogEntry();
    entry->parseFromString(serialized_entry);

    if (log_->isPrimary()) {
        if (entry->decision_log_entry == -1) {
            if (transaction == NULL) {
                // This was a two-phase commit transaction that was timeout aborted: do nothing
                assert(entry->multiple_partitions);
                delete entry;
                return;
            }

            // This is a transaction that has been replicated: send the response out
            CHECK(transaction->work_units() == entry->fragments);
            CHECK(transaction->multiple_partitions() == entry->multiple_partitions);
            if (transaction->multiple_partitions()) {
                // record that replication completed
                ServerTransactionState* state = reinterpret_cast<ServerTransactionState*>(
                        transaction->external_state());
                state->setReplicated();
            }
            sendResponse(transaction);
        } else {
            // Decision for a multi-partition transaction
            // TODO: We need to ACK this: inform the coordinator it can forget about the transaction
            assert(0 <= entry->decision_log_entry);
            assert(entry->decision_log_entry < sequence);
        }
    } else {
        assert(transaction == NULL);
        if (entry->decision_log_entry == -1) {
            // Queue transactions until later
            // We abuse the log entry's fields to record decisions
            if (!entry->multiple_partitions) {
                // mark this transactions as decided
                entry->decision_log_entry = 1;
                entry->commit = true;
            }
            assert(sequence >= backup_undecided_.nextIndex());
            backup_undecided_.insert(sequence, entry);
            entry = NULL;  // ownership transferred to backup_undecided_ queue
        } else {
            // Record the decision
            assert(0 <= entry->decision_log_entry);
            assert(entry->decision_log_entry < sequence);
            LogEntry* e = backup_undecided_.at(entry->decision_log_entry);
            assert(e->decision_log_entry == -1);
            e->decision_log_entry = 1;
            e->commit = entry->commit;
        }
    }

    delete entry;
}

void DtxnServer::pollForResults() {
    Transaction* transaction = NULL;
    while ((transaction = scheduler_->getWorkUnitResult()) != NULL) {
        if (log_ != NULL) {
            // Non-primaries cannot (should not?) be executing transactions!
            assert(log_->isPrimary());
            // Cases:
            // * Abort: return results; do not replicate (no need!)
            // * Single partition commit: return results; replicate
            // * Multi-partition commit not last: return results; do not replicate
            // * Multi-partition commit last fragment: do not return results; replicate
            // Rules:
            // * Replicate if we are committing the last fragment
            // * Return results if we are *not* replicating
            if (transaction->last_fragment() && transaction->last_status() == ExecutionEngine::OK) {
                // This transaction is committing: Replicate!
                int sequence = log_->submit(transaction->log_entry(), transaction);
                ((ServerTransactionState*) transaction->external_state())->log_sequence(sequence);
            } else {
                // Either aborting or an unfinished multi-partition transaction: return immediately
                assert(transaction->last_status() == ExecutionEngine::ABORT_USER ||
                        transaction->last_status() == ExecutionEngine::ABORT_DEADLOCK || 
                        !transaction->last_fragment());
                sendResponse(transaction);
            }
        } else {
            // Non-replicated: just send the response out
            sendResponse(transaction);
        }
    }
}

void DtxnServer::sendResponse(Transaction* transaction) {
    if (transaction->multiple_partitions()) {
        realSendResponse(transaction);
    } else {
        sp_pending_sends_.push_back(transaction);
    }
}

void DtxnServer::realSendResponse(Transaction* transaction) {
    assert(transaction->last_status() == ExecutionEngine::OK ||
            transaction->last_status() == ExecutionEngine::ABORT_USER ||
            transaction->last_status() == ExecutionEngine::ABORT_DEADLOCK);
    ServerTransactionState* state = reinterpret_cast<ServerTransactionState*>(
            transaction->external_state());

    FragmentResponse response;
    response.id = state->id();
    response.status = transaction->last_status();
    // hack to avoid a copy
    response.result.clear();
    response.result.swap(*transaction->output());
    msg_server_->send(state->client(), response);
    transaction->resultSent();

    // If this is a single partition transaction result, we are done: delete it
    if (!transaction->multiple_partitions()) {
        finishTransaction(transaction, transaction->last_status() == ExecutionEngine::OK);
    }
}

void DtxnServer::finishTransaction(Transaction* transaction, bool commit) {
    // Unregister the transaction
    ServerTransactionState* state = reinterpret_cast<ServerTransactionState*>(
            transaction->external_state());
    pair<net::ConnectionHandle*, int> key(state->client(), state->id());
    size_t count = transaction_id_map_.erase(key);
    ASSERT(count == 1);

    if (commit) {
        scheduler_->commit(transaction);
    } else {
        scheduler_->abort(transaction);
    }
    
    delete state;
}

bool DtxnServer::idle() {
    // let the scheduler do idle processing, which could include deadlock detection.
    // let it decide if we should block or not
    bool poll_network = scheduler_->idle();
    pollForResults();
    // flush any buffered replication messages
    // TODO: Use a more intelligent policy?
    if (log_ != NULL) {
        log_->flush();
    }

    // Send any queued SP transactions
    while (!sp_pending_sends_.empty()) {
        realSendResponse(sp_pending_sends_.dequeue());
    }

    // Apply decided backup transactions
    int count = 0;
    bool has_more = false;
    while (!backup_undecided_.empty()) {
        LogEntry* undecided = backup_undecided_.front();
        if (undecided != NULL) {
            if (undecided->decision_log_entry == -1) {
                break;
            }
            assert(undecided->decision_log_entry == 1);
            // If the transaction aborted, just skip it
            if (undecided->commit) {
                if (count > 0) {
                    has_more = true;
                    break;
                }
                count += 1;
                scheduler_->backupApplyTransaction(undecided->fragments);
            }
            delete undecided;
        }
        backup_undecided_.pop_front();
    }

    return poll_network | has_more;
}

bool DtxnServer::idleCallback(void* argument) {
    DtxnServer* server = reinterpret_cast<DtxnServer*>(argument);
    return server->idle();
}

}  // namespace dtxn
