#ifndef DTXN_DTXNSERVER_H__
#define DTXN_DTXNSERVER_H__

#include <tr1/functional>  // for std::tr1::hash
#include <vector>

#include "base/circularbuffer.h"
#include "base/slidingarray.h"
#include "base/unordered_map.h"
#include "replication/faulttolerantlog.h"

namespace io {
class EventLoop;
}

namespace net {
class ConnectionHandle;
class MessageServer;
}

namespace replication {
class FaultTolerantLog;
}

// TODO: Move this hash function into stlutil or similar?
// TODO: Should this be in std::tr1? Technically probably not, but it is convenient
template <typename T, typename U>
class pairhash {
public:
    size_t operator()(const std::pair<T, U>& value) const {
        size_t h = t_hash_(value.first);
        // Bitwise rotation: order matters (hash(x, y) != hash(y, x)) and equality generates
        // different hashes (hash(x, x) != hash(y, y))
        static const int BITS = 19;
        h = (h << BITS) | (h >> (sizeof(h)*8 - BITS));
        return h ^ u_hash_(value.second);
    }

private:
    std::tr1::hash<T> t_hash_;
    std::tr1::hash<U> u_hash_;
};

namespace dtxn {

class Fragment;
class CommitDecision;
class LogEntry;
class Transaction;
class TransactionScheduler;

/** Provides the network interface to a single partition.

NOTE: This no longer manages client connections itself. The disadvantage is that we no longer
crash when a multi-partition transaction coordinator connection closes. We need to fix this with
timeouts in the future, plus the replicated 2PC log service.
*/
class DtxnServer : public replication::FaultTolerantLogCallback {
public:
    // Owns scheduler and log. Does not own event_loop, msg_server or log. log is optional.
    // TODO: Does this ownership policy make sense? Maybe this should own everything?
    DtxnServer(TransactionScheduler* scheduler, io::EventLoop* event_loop,
            net::MessageServer* msg_server, replication::FaultTolerantLog* log);
    ~DtxnServer();

    // fragment arrives from connection. Used to deliver new transactions to this server.
    void fragmentReceived(net::ConnectionHandle* connection, const Fragment& fragment);

    // A commit/abort decision arrives from connection. A multi-partition transaction must be
    // active.
    void decisionReceived(net::ConnectionHandle* connection, const CommitDecision& decision);

    virtual void nextLogEntry(int sequence, const std::string& entry, void* argument);

private:
    // Sends the results of transaction back to the client.
    void sendResponse(Transaction* transaction);
    void realSendResponse(Transaction* transaction);

    // Checks if the scheduler has any results. If so, they get returned.
    void pollForResults();

    // Called when there are no more network events to process.
    bool idle();

    void finishTransaction(Transaction* transaction, bool commit);

    // Callback used by the event loop to notify that all network events have been processed.
    static bool idleCallback(void* argument);

    TransactionScheduler* scheduler_;
    // TODO: We should be able to replace event_loop_ with msg_server_
    io::EventLoop* event_loop_;
    net::MessageServer* msg_server_;
    replication::FaultTolerantLog* log_;

    typedef base::unordered_map<
            std::pair<net::ConnectionHandle*, int>,
            Transaction*,
            pairhash<net::ConnectionHandle*, int> > TransactionIdMap;
    TransactionIdMap transaction_id_map_;

    base::SlidingArray<LogEntry*> backup_undecided_;
    CircularBuffer<Transaction*> sp_pending_sends_;
};

}  // namespace dtxn

#endif
