#include "dtxn/dtxnserver.h"
#include "dtxn/executionengine.h"
#include "dtxn/messages.h"
#include "dtxn/transaction.h"
#include "dtxn/transactionscheduler.h"
#include "io/eventloop.h"
#include "mockmessageconnection.h"
#include "net/messageserver.h"
#include "replication/mockfaulttolerantlog.h"
#include "stupidunit/stupidunit.h"


using namespace dtxn;
using std::string;
using std::vector;

class MockTransactionScheduler : public TransactionScheduler {
public:
    MockTransactionScheduler() : idle_(false),
            work_count_(0), commit_count_(0), abort_count_(0), backup_count_(0) {}

    virtual ~MockTransactionScheduler() {}
    virtual Transaction* begin(void* external_state, const string& initial_work_unit,
            bool last_fragment, bool multiple_partitions) {
        Transaction* txn = new NoLockTransaction(
                external_state, initial_work_unit, last_fragment, multiple_partitions);
        begun_.push_back(txn);
        return txn;
    }

    virtual void workUnitAdded(Transaction* transaction) {
        work_count_ += 1;
    }

    virtual void commit(Transaction* transaction) {
        commit_count_ += 1;
        delete transaction;
    }

    virtual void abort(Transaction* transaction) {        
        // Remove from the output queue: this is the scheduler's responsability
        outputs_.eraseValue(transaction);

        abort_count_ += 1;
        delete transaction;
    }

    virtual Transaction* getWorkUnitResult() {
        if (outputs_.empty()) return NULL;
        return outputs_.dequeue();
    }

    virtual bool idle() {
        idle_ = true;
        return false;
    }

    virtual void backupApplyTransaction(const std::vector<std::string>& work_units) {
        assert(!work_units.empty());
        backup_count_ += (int) work_units.size();
    }

    bool idle_;

    int work_count_;
    int commit_count_;
    int abort_count_;
    int backup_count_;

    CircularBuffer<Transaction*> begun_;
    CircularBuffer<Transaction*> outputs_;
};

class BaseDtxnServerTest : public Test {
public:
    BaseDtxnServerTest() :
            server_(NULL),
            scheduler_(new MockTransactionScheduler()),
            client_(new MockMessageConnection()) {
        client_handle_ = msg_server_.addConnection(client_);
        fragment_.transaction = "hello world";
    }

    ~BaseDtxnServerTest() {
        delete server_;
    }

    DtxnServer* server_;
    MockTransactionScheduler* scheduler_;
    io::MockEventLoop event_loop_;
    MockMessageConnection* client_;
    net::ConnectionHandle* client_handle_;
    net::MessageServer msg_server_;
    Fragment fragment_;
    FragmentResponse response_;
    CommitDecision decision_;
};

class DtxnServerTest : public BaseDtxnServerTest {
public:
    DtxnServerTest() {
        server_ = new DtxnServer(scheduler_, &event_loop_, &msg_server_, NULL);
    }
};

TEST_F(DtxnServerTest, NullCreate) {
    EXPECT_DEATH(DtxnServer(NULL, &event_loop_, &msg_server_, NULL));
    EXPECT_DEATH(DtxnServer(scheduler_, NULL, &msg_server_, NULL));
    EXPECT_DEATH(DtxnServer(scheduler_, &event_loop_, NULL, NULL));
}

TEST_F(DtxnServerTest, Idle) {
    // Nothing happening: calls the scheduler
    EXPECT_FALSE(scheduler_->idle_);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_TRUE(scheduler_->idle_);
}

TEST_F(DtxnServerTest, UnregisterCallbacksInDestructor) {
    io::MockEventLoop loop;
    net::MessageServer msg_server_;
    {
        DtxnServer s(new MockTransactionScheduler(), &loop, &msg_server_, NULL);
        EXPECT_TRUE(loop.idle_callback_enabled_);
        EXPECT_TRUE(msg_server_.registered<Fragment>());
    }
    // Make sure the idle handler is unregistered in the destructor
    EXPECT_FALSE(loop.idle_callback_enabled_);
    EXPECT_FALSE(msg_server_.registered<Fragment>());
}

TEST_F(DtxnServerTest, SinglePartition) {
    fragment_.multiple_partitions = false;

    fragment_.id = 0;
    server_->fragmentReceived(client_handle_, fragment_);

    EXPECT_EQ(1, scheduler_->begun_.size());
    EXPECT_FALSE(client_->hasMessage());

    // Start another transaction: skipping transaction ids is permitted
    EXPECT_DEATH(server_->fragmentReceived(client_handle_, fragment_));
    fragment_.id = 15;
    server_->fragmentReceived(client_handle_, fragment_);
    EXPECT_EQ(2, scheduler_->begun_.size());
    EXPECT_FALSE(client_->hasMessage());

    // respond to the latest client message
    Transaction* txn = scheduler_->begun_.at(1);
    txn->output()->assign("out");
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);

    // The idle handler gives a chance to sent results
    event_loop_.idle();
    client_->getMessage(&response_);
    EXPECT_EQ(15, response_.id);
    EXPECT_EQ(ExecutionEngine::OK, response_.status);
    EXPECT_EQ("out", response_.result);
    EXPECT_EQ(1, scheduler_->commit_count_);

    txn = scheduler_->begun_.at(0);
    txn->output()->assign("out2");
    txn->setWorkUnitStatus(ExecutionEngine::ABORT_USER);
    scheduler_->outputs_.push_back(txn);

    event_loop_.idle();
    client_->getMessage(&response_);
    EXPECT_EQ(0, response_.id);
    EXPECT_EQ(ExecutionEngine::ABORT_USER, response_.status);
    EXPECT_EQ("out2", response_.result);
    EXPECT_EQ(1, scheduler_->abort_count_);
}

TEST_F(DtxnServerTest, OutOfOrderDestruct) {
    // This transaction is pending
    server_->fragmentReceived(client_handle_, fragment_);
    /*Transaction* pending = */scheduler_->begun_.dequeue();

    // This transaction completes
    fragment_.id = 2;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* completed = scheduler_->begun_.dequeue();
    completed->setWorkUnitStatus(ExecutionEngine::ABORT_DEADLOCK);
    scheduler_->outputs_.push_back(completed);
    event_loop_.idle();
    EXPECT_TRUE(client_->hasMessage());

    // the destructor should skip the completed transaction
}

TEST_F(DtxnServerTest, DeadlockAbort) {
    //~ fragment_.multiple_partitions = false;
    //~ fragment_.id = 0;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.front();
    txn->setWorkUnitStatus(ExecutionEngine::ABORT_DEADLOCK);
    scheduler_->outputs_.push_back(txn);

    event_loop_.idle();
    client_->getMessage(&response_);
    EXPECT_EQ(0, response_.id);
    EXPECT_EQ(ExecutionEngine::ABORT_DEADLOCK, response_.status);
    EXPECT_EQ("", response_.result);
    EXPECT_EQ(1, scheduler_->abort_count_);
}

TEST_F(DtxnServerTest, BadDecision) {
    // Decision for a transaction that has not started
    EXPECT_DEATH(server_->decisionReceived(client_handle_, decision_));
    server_->fragmentReceived(client_handle_, fragment_);
    // Decision for a queued single partition transaction
    EXPECT_DEATH(server_->decisionReceived(client_handle_, decision_));

    Transaction* txn = scheduler_->begun_.at(0);
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    EXPECT_TRUE(client_->hasMessage());

    // Decision for a finished transaction
    EXPECT_DEATH(server_->decisionReceived(client_handle_, decision_));
}

TEST_F(DtxnServerTest, MultiplePhaseCommit) {
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = false;
    server_->fragmentReceived(client_handle_, fragment_);

    Transaction* txn = scheduler_->begun_.dequeue();
    EXPECT_TRUE(txn->multiple_partitions());
    EXPECT_FALSE(txn->last_fragment());

    // cannot receive another part until we have responded
    fragment_.transaction = "part2";
    EXPECT_DEATH(server_->fragmentReceived(client_handle_, fragment_));

    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    client_->getMessage(&response_);
    EXPECT_EQ(0, response_.id);
    EXPECT_EQ(ExecutionEngine::OK, response_.status);
    EXPECT_EQ("", response_.result);

    // receive part 2
    server_->fragmentReceived(client_handle_, fragment_);
    EXPECT_EQ("part2", txn->last_work_unit());
    EXPECT_EQ(1, scheduler_->work_count_);

    // can't commit until we sent out results!
    decision_.commit = true;
    EXPECT_DEATH(server_->decisionReceived(client_handle_, decision_));

    txn->setWorkUnitStatus(ExecutionEngine::OK);
    txn->output()->assign("foo");
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    client_->getMessage(&response_);
    EXPECT_EQ(0, response_.id);
    EXPECT_EQ(ExecutionEngine::OK, response_.status);
    EXPECT_EQ("foo", response_.result);

    server_->decisionReceived(client_handle_, decision_);
    EXPECT_EQ(1, scheduler_->commit_count_);
    EXPECT_EQ(false, client_->hasMessage());
}

TEST_F(DtxnServerTest, BadDoneMessage) {
    // Begin a multiple partition transaction
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = false;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    EXPECT_TRUE(client_->hasMessage());

    // Execute a fragment with the done bit set
    fragment_.last_fragment = true;
    server_->fragmentReceived(client_handle_, fragment_);
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();

    // Add another piece of work for this transaction: NOT PERMITTED
    EXPECT_DEATH(server_->fragmentReceived(client_handle_, fragment_));
}


class DtxnServerLogTest : public BaseDtxnServerTest {
public:
    DtxnServerLogTest() {
        server_ = new DtxnServer(scheduler_, &event_loop_, &msg_server_, &log_);
    }

    void nextBackupLog(int index) {
        string serialized;
        log_entry_.appendToString(&serialized);
        log_.nextLogEntryCallback(index, serialized, NULL);
    }

    replication::MockFaultTolerantLog log_;
    LogEntry log_entry_;
};

TEST_F(DtxnServerLogTest, PrimarySinglePartition) {
    log_.is_primary(true);

    // Receive a single partition transaction and give it to the scheduler.
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    EXPECT_EQ(0, log_.flush_count_);
    event_loop_.idle();
    EXPECT_EQ(1, log_.flush_count_);

    // Client can't get message yet: must be replicated
    EXPECT_FALSE(client_->hasMessage());
    ASSERT_EQ(1, log_.submitted_.size());

    // Queue a second transaction
    fragment_.id = 2;
    server_->fragmentReceived(client_handle_, fragment_);
    txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    EXPECT_EQ(2, log_.submitted_.size());

    // Replication completes for first transaction; client gets results after idle
    log_.callbackSubmitted(0);
    EXPECT_FALSE(client_->hasMessage());
    event_loop_.idle();
    EXPECT_TRUE(client_->hasMessage());
    client_->message_.clear();

    // A transaction begins and aborts: not replicated
    fragment_.id = 3;
    server_->fragmentReceived(client_handle_, fragment_);
    txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::ABORT_USER);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    EXPECT_EQ(2, log_.submitted_.size());
    EXPECT_TRUE(client_->hasMessage());
}

TEST_F(DtxnServerLogTest, BackupSinglePartition) {
    log_.is_primary(false);

    log_entry_.fragments.push_back("foo");
    log_entry_.multiple_partitions = false;
    nextBackupLog(0);
    // Not applied until idle
    EXPECT_EQ(0, scheduler_->backup_count_);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(1, scheduler_->backup_count_);
}

TEST_F(DtxnServerLogTest, PrimaryMultiPartition) {
    log_.is_primary(true);

    // Receive a multi-partition transaction (not finished!) and give it to the scheduler.
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = false;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();

    // Client *should* get the message. Nothing is replicated.
    EXPECT_TRUE(client_->hasMessage());
    client_->message_.clear();
    EXPECT_EQ(0, log_.submitted_.size());

    // Prepare the transaction
    fragment_.transaction = "foo";
    fragment_.last_fragment = true;
    server_->fragmentReceived(client_handle_, fragment_);
    EXPECT_EQ(1, scheduler_->work_count_);
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();

    // No answer to the client; NOW we replicate
    EXPECT_FALSE(client_->hasMessage());
    client_->message_.clear();
    ASSERT_EQ(1, log_.submitted_.size());

    log_entry_.parseFromString(log_.submitted_[0]);
    EXPECT_TRUE(log_entry_.multiple_partitions);
    ASSERT_EQ(2, log_entry_.fragments.size());
    EXPECT_EQ("hello world", log_entry_.fragments[0]);
    EXPECT_EQ("foo", log_entry_.fragments[1]);

    // Replication completes: client gets final results
    log_.callbackSubmitted(0);
    EXPECT_TRUE(client_->hasMessage());
    client_->message_.clear();

    // decision is received and replicated
    decision_.id = fragment_.id;
    decision_.commit = true;
    server_->decisionReceived(client_handle_, decision_);
    ASSERT_EQ(2, log_.submitted_.size());

    log_entry_.parseFromString(log_.submitted_[1]);
    EXPECT_TRUE(log_entry_.multiple_partitions);
    EXPECT_EQ(0, log_entry_.fragments.size());
    EXPECT_EQ(0, log_entry_.decision_log_entry);
    EXPECT_TRUE(log_entry_.commit);

    // decision finalized, nothing visible happens
    log_.callbackSubmitted(1);

    // A transaction begins and aborts: not replicated
    fragment_.id = 2;
    server_->fragmentReceived(client_handle_, fragment_);
    txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::ABORT_USER);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();
    EXPECT_EQ(2, log_.submitted_.size());
    EXPECT_TRUE(client_->hasMessage());

    // Decision arrives for the aborted transaction
    // TODO: Should the coordinator not send this message?
    decision_.id = 2;
    decision_.commit = false;
    server_->decisionReceived(client_handle_, decision_);
    // Nothing replicated
    EXPECT_EQ(2, log_.submitted_.size());
}

TEST_F(DtxnServerLogTest, PrimaryMultiPartitionTimeoutBeforeReplication) {
    log_.is_primary(true);

    // Receive a multi-partition transaction that finishes
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = true;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();

    // Replication occurs
    EXPECT_EQ(1, log_.submitted_.size());

    // Now an abort is received
    decision_.commit = false;
    server_->decisionReceived(client_handle_, decision_);

    // The abort is replicated, even without the acks for the replication
    EXPECT_EQ(2, log_.submitted_.size());
    EXPECT_FALSE(client_->hasMessage());

    // Ack the original replication: no output (already aborted)
    log_.callbackSubmitted(0);
    EXPECT_FALSE(client_->hasMessage());

    // Ack the abort: still okay
    log_.callbackSubmitted(1);
    EXPECT_FALSE(client_->hasMessage());
}

TEST_F(DtxnServerLogTest, PrimaryMultiPartitionTimeoutWhileQueued) {
    log_.is_primary(true);

    // Receive a multi-partition transaction that finishes
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = true;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);

    // Receive an abort *before* we try to replicate the entry
    decision_.commit = false;
    server_->decisionReceived(client_handle_, decision_);

    // Removed from the output queue: this is the scheduler's responsability
    EXPECT_EQ(0, scheduler_->outputs_.size());

    // No replication occurs: we just forget about the transaction
    event_loop_.idle();
    event_loop_.idle();
    EXPECT_EQ(0, log_.submitted_.size());
    EXPECT_FALSE(client_->hasMessage());
}

TEST_F(DtxnServerLogTest, PrimaryMultiPartitionAbortBeforePrepare) {
    log_.is_primary(true);

    // Unfinished multi-partition transaction
    fragment_.multiple_partitions = true;
    fragment_.last_fragment = false;
    server_->fragmentReceived(client_handle_, fragment_);
    Transaction* txn = scheduler_->begun_.dequeue();
    txn->setWorkUnitStatus(ExecutionEngine::OK);
    scheduler_->outputs_.push_back(txn);
    event_loop_.idle();

    // No replication
    EXPECT_EQ(0, log_.submitted_.size());
    EXPECT_TRUE(client_->hasMessage());
    client_->message_.clear();

    // Abort decision is received: no replication or responses
    decision_.commit = false;
    server_->decisionReceived(client_handle_, decision_);
    EXPECT_EQ(0, log_.submitted_.size());
    EXPECT_FALSE(client_->hasMessage());
}

TEST_F(DtxnServerLogTest, BackupMultiPartition) {
    log_.is_primary(false);

    // receive a pair of fragments
    log_entry_.fragments.push_back("foo");
    log_entry_.fragments.push_back("bar");
    log_entry_.multiple_partitions = true;
    nextBackupLog(0);
    EXPECT_FALSE(event_loop_.idle());
    // not executed yet: still not decided
    EXPECT_EQ(0, scheduler_->backup_count_);

    // A single-partition transaction arrives: still nothing executed
    log_entry_.fragments.clear();
    log_entry_.fragments.push_back("baz");
    log_entry_.multiple_partitions = false;
    nextBackupLog(1);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(0, scheduler_->backup_count_);

    // A multi-partition transaction arrives: again not executed
    log_entry_.multiple_partitions = true;
    EXPECT_DEATH(nextBackupLog(1));  // bad sequence
    nextBackupLog(2);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(0, scheduler_->backup_count_);

    // decision arrives for the 2nd mp txn: abort
    log_entry_.fragments.clear();
    log_entry_.decision_log_entry = 2;
    log_entry_.commit = false;
    nextBackupLog(3);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(0, scheduler_->backup_count_);

    // Another single partition transaction: this puts a "gap" in the sequence numbers
    log_entry_.fragments.push_back("foo");
    log_entry_.multiple_partitions = false;
    log_entry_.decision_log_entry = -1;
    nextBackupLog(4);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(0, scheduler_->backup_count_);

    // decision arrives for the 1st mp txn: everything is executed
    log_entry_.fragments.clear();
    log_entry_.multiple_partitions = true;
    log_entry_.decision_log_entry = 0;
    log_entry_.commit = true;
    // Only apply one entry on each idle call
    nextBackupLog(5);
    EXPECT_TRUE(event_loop_.idle());
    EXPECT_EQ(2, scheduler_->backup_count_);
    EXPECT_TRUE(event_loop_.idle());
    EXPECT_EQ(3, scheduler_->backup_count_);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(4, scheduler_->backup_count_);
    EXPECT_FALSE(event_loop_.idle());
    EXPECT_EQ(4, scheduler_->backup_count_);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
