#include "dtxn/ordered/ordereddistributor.h"

#include <netinet/in.h>
#include <sys/time.h>

#include <algorithm>

#include "base/stlutil.h"
#include "dtxn/configparser.h"
#include "dtxn/dtxndistributor.h"
#include "dtxn/executionengine.h"
#include "dtxn/messages.h"
#include "io/libeventloop.h"
#include "messageconnection.h"
#include "net/messageserver.h"
#include "strings/utils.h"

using std::string;
using std::vector;

namespace dtxn {

// Contains the state for a pending transaction.
class OrderedDistributor::TransactionState {
public:
    TransactionState(net::ConnectionHandle* client, size_t num_partitions,
            int32_t distributor_id, int32_t request_id, const string& initial_client_message) :

            client_(client),
            distributor_id_(distributor_id),
            request_id_(request_id),
            transaction_(num_partitions, initial_client_message),
            timer_(NULL),
            distributor_(NULL) {
        assert(client_ != NULL);
        assert(request_id_ >= 0);
    }

    ~TransactionState() {
        if (timer_ != NULL) {
            distributor_->event_loop()->cancelTimeOut(timer_);
        }
    }

    // Returns true if we have all the responses for this round
    bool responseReceived(int partition_index, const std::string& response,
            ExecutionEngine::Status status) {
        // TODO: Abort early?
        transaction_.receive(partition_index, response, status);
        return transaction_.receivedAll();
    }

    net::ConnectionHandle* client() { return client_; }
    void clearClient() { assert(client_ != NULL); client_ = NULL; }
    int32_t distributor_id() const { return distributor_id_; }
    int32_t request_id() const { return request_id_; }

    void startResponseTimer(OrderedDistributor* distributor, int timeout_ms) {
        distributor_ = distributor;

        if (timer_ == NULL) {
            timer_ = distributor_->event_loop()->createTimeOut(timeout_ms, timerCallback, this);
        } else {
            distributor_->event_loop()->resetTimeOut(timer_, timeout_ms);
        }
    }

    CoordinatedTransaction* transaction() { return &transaction_; }

private:
    static void timerCallback(void* argument) {
        TransactionState* transaction = reinterpret_cast<TransactionState*>(argument);
        transaction->distributor_->responseTimeout(transaction);
    }

    net::ConnectionHandle* client_;
    int32_t distributor_id_;
    int32_t request_id_;

    CoordinatedTransaction transaction_;

    // Used to time out this transaction if a round takes too long to complete
    void* timer_;
    OrderedDistributor* distributor_;
};

OrderedDistributor::OrderedDistributor(const vector<net::ConnectionHandle*>& partitions,
        Coordinator* coordinator, io::EventLoop* event_loop, net::MessageServer* msg_server) :
        partitions_(partitions),
        coordinator_(coordinator),
        event_loop_(event_loop),
        msg_server_(msg_server) {
    assert(!partitions_.empty());
    assert(coordinator_ != NULL);
    assert(event_loop_ != NULL);

    msg_server_->addCallback(&OrderedDistributor::requestReceived, this);
    msg_server_->addCallback(&OrderedDistributor::responseReceived, this);
}

OrderedDistributor::~OrderedDistributor() {
    //~ STLDeleteElements(&partitions_);
    // Close all partition connections
    for (size_t i = 0; i < partitions_.size(); ++i) {
        msg_server_->closeConnection(partitions_[i]);
    }

    // Clean up queued messages
    for (size_t i = queue_.firstIndex(); i < queue_.nextIndex(); ++i) {
        // TODO: is this the "correct" way to clean up queued messages?
        if (queue_.at(i) != NULL) {
            coordinator_->done(queue_.at(i)->transaction());
            delete queue_.at(i);
        }
    }

    msg_server_->removeCallback(&OrderedDistributor::requestReceived);
    msg_server_->removeCallback(&OrderedDistributor::responseReceived);
}

void OrderedDistributor::requestReceived(net::ConnectionHandle* connection,
        const Fragment& request) {
    // TODO: Verify that the request ids are being generated correctly?

    // Queue and/or dispatch the transaction
    bool last_done = queue_.empty() || queue_.back() == NULL ||
            queue_.back()->transaction()->allDone();
    TransactionState* state = new TransactionState(
            connection, partitions_.size(), assert_range_cast<int32_t>(queue_.nextIndex()),
            request.id, request.transaction);
    queue_.push_back(state);
    assert(queue_.at(state->distributor_id()) == state);
    if (last_done) {
        dispatchTransaction(queue_.back());
    }
}

void OrderedDistributor::responseReceived(net::ConnectionHandle* connection,
        const FragmentResponse& response) {
    // response must be for the current transaction, or the previous transaction if it aborted
    if (response.id < assert_range_cast<int32_t>(queue_.firstIndex())) {
        // ignore this response: it is for an old transaction
        // TODO: Verify that we aborted this transaction due to a timeout?
        return;
    }
    TransactionState* state = queue_.at(response.id);
    assert(state != NULL);

    // Find the partition index
    int partition_index = -1;
    for (partition_index = 0; partition_index < partitions_.size(); ++partition_index) {
        if (partitions_[partition_index] == connection) break;
    }
    assert(0 <= partition_index && partition_index < partitions_.size());

    state->transaction()->receive(partition_index, response.result,
            (ExecutionEngine::Status) response.status);
    if (state->transaction()->receivedAll()) {
        if (state->transaction()->status() != ExecutionEngine::OK) {
            // We are aborting: send an abort to the client because the coordinator has no choice
            // Select the first message non-empty message we received
            // TODO: We need to run "user code" to decide what to do about the message.
            const string* message = NULL;
            assert(!state->transaction()->received().empty());
            for (size_t i = 0; i < state->transaction()->received().size(); ++i) {
                message = &state->transaction()->received()[i].second;
                if (!message->empty()) {
                    break;
                }
            }
            state->transaction()->abort(*message);
        } else {
            // We have received all responses: ask the coordinator what to do
            coordinator_->nextRound(state->transaction());
        }
        processFragments(state);
    }
}

void OrderedDistributor::responseTimeout(TransactionState* state) {
    assert(base::contains(queue_, state));

    // If the transaction times out, we abort it unconditionally: should indicate deadlock
    // TODO: Indicate a specific timeout code or message?
    state->transaction()->abort("");
    finishTransaction(state);
}

void OrderedDistributor::dispatchTransaction(TransactionState* state) {
    // every transaction except this one must be all done
#ifndef NDEBUG
    for (size_t i = queue_.firstIndex(); i < state->distributor_id(); ++i) {
        assert(queue_.at(i) == NULL || queue_.at(i)->transaction()->allDone());
    }
#endif

    coordinator_->begin(state->transaction());
    if (!state->transaction()->allDone() || state->transaction()->sent().size() > 1) {
        state->transaction()->setMultiplePartitions();
    }
    processFragments(state);
}

void OrderedDistributor::processFragments(TransactionState* state) {
    CoordinatedTransaction* transaction = state->transaction();
    assert(!transaction->sent().empty());

    // If we should send a response to the client
    if (transaction->haveClientResponse()) {
        finishTransaction(state);
        return;
    }

    // Send out messages to partitions
    Fragment request;
    request.id = state->distributor_id();
    request.multiple_partitions = transaction->multiple_partitions();
    const CoordinatedTransaction::MessageList& messages = transaction->sent();
    for (int i = 0; i < messages.size(); ++i) {
        int partition_index = messages[i].first;
        request.transaction = messages[i].second;
        request.last_fragment = transaction->isDone(partition_index);
        msg_server_->send(partitions_[partition_index], request);
    }

    // start the deadlock timer for multi-partition transactions
    // TODO: Don't do this for the "ordered request" mode?
    if (request.multiple_partitions) {
        // TODO: Does adding a small random variation reduce probability of simultaneous aborts?
        // TODO: Tune this better?
        state->startResponseTimer(this, 800);
    }
    transaction->sentMessages();

    // If this is the last round, dispatch the next transaction
    if (state->transaction()->allDone() && state->distributor_id() +1 < queue_.nextIndex()) {
        dispatchTransaction(queue_.at(state->distributor_id() + 1));
    }
}

void OrderedDistributor::finishTransaction(TransactionState* state) {
    assert(state->transaction()->haveClientResponse());
    if (state->transaction()->multiple_partitions()) {
        CommitDecision decision;
        decision.id = state->distributor_id();
        decision.commit = state->transaction()->status() == ExecutionEngine::OK;

        vector<int> involved_partitions = state->transaction()->involvedPartitions();
        for (int i = 0; i < involved_partitions.size(); ++i) {
            int index = involved_partitions[i];
            msg_server_->send(partitions_[index], decision);
        }
    }

    // Return the results to the client
    FragmentResponse client_response;
    client_response.id = state->request_id();
    client_response.status = state->transaction()->status();
    assert(state->transaction()->sent()[0].first ==
            CoordinatedTransaction::CLIENT_PARTITION);
    client_response.result = state->transaction()->sent()[0].second;
    msg_server_->send(state->client(), client_response);

    // Clean up this transaction, since we are DONE
    coordinator_->done(state->transaction());

    // Remove the request from the queue
    assert(queue_.at(state->distributor_id()) == state);
    queue_.at(state->distributor_id()) = NULL;
    while (!queue_.empty() && queue_.front() == NULL) {
        queue_.pop_front();
    }
    delete state;
}

}  // namespace dtxn
