#include "dtxn/ordered/orderedscheduler.h"

#include <cstdlib>
#include <string>

#include "base/assert.h"
#include "base/stlutil.h"
#include "dtxn/transaction.h"

using std::string;

namespace dtxn {

OrderedScheduler::~OrderedScheduler() {
    STLDeleteElements(&single_partition_queue_);
    STLDeleteElements(&multi_partition_queue_);
    STLDeleteElements(&result_queue_);
    STLDeleteElements(&unfinished_queue_);
}

void OrderedScheduler::cc_mode(ConcurrencyControl mode) {
    assert(mode == CC_BLOCK || mode == CC_SPECULATIVE);

    // Check that there is no work underway
    assert(unfinished_queue_.empty());
    cc_mode_ = mode;
}

Transaction* OrderedScheduler::begin(void* external_state, const std::string& initial_work_unit,
        bool last_fragment, bool multiple_partitions) {
    Transaction* state = new NoLockTransaction(
            external_state, initial_work_unit, last_fragment, multiple_partitions);

    if (unfinished_queue_.empty() && multiple_partitions) {
        // New multi-partition transaction: execute it
        executeFragment(state);
    } else {
        // Queue this transaction for later
        if (multiple_partitions) {
            multi_partition_queue_.push_back(state);
        } else {
            single_partition_queue_.push_back(state);
        }
    }

    return state;
}

void OrderedScheduler::workUnitAdded(Transaction* transaction) {
    assert(transaction->multiple_partitions());
    assert(transaction->last_status() == ExecutionEngine::INVALID);
    assert(transaction == unfinished_queue_.front());
    executeFragment(transaction);
}

void OrderedScheduler::commit(Transaction* transaction) {
    assert(transaction->last_status() == ExecutionEngine::OK);

    if (!unfinished_queue_.empty() && transaction == unfinished_queue_.front()) {
        unfinished_queue_.pop_front();

        if (*transaction->undo() != NULL) {
            engine_->freeUndo(*transaction->undo());
        }

        // Speculative execution was correct: output transactions
        while (!unfinished_queue_.empty()) {
            Transaction* speculative = unfinished_queue_.front();
            result_queue_.push_back(speculative);
            if (!speculative->multiple_partitions()) {
                // finish single partition transactions, freeing any undo buffers they may have
                unfinished_queue_.pop_front();
                if (*speculative->undo() != NULL) {
                    engine_->freeUndo(*speculative->undo());
                }
            } else {
                // Only output the first chunk of a multi-partition transaction that was speculated
                assert(mp_speculated_);
                mp_speculated_ = false;
                break;
            }
        }
    }

    // Free the undo buffer and the transaction state
    cleanUpTransaction(transaction);
}

void OrderedScheduler::abort(Transaction* transaction) {
    // Look for the transaction in the speculative queue
    if (transaction->last_status() != ExecutionEngine::INVALID) {
        bool found_speculative = false;
        for (size_t i = 0; i < unfinished_queue_.size(); ++i) {
            if (unfinished_queue_.at(i) == transaction) {
                // Speculative execution incorrect: put transactions back on the queue
                // and apply undo in LIFO order
                while (unfinished_queue_.size() - 1 > i) {
                    Transaction* speculative = unfinished_queue_.back();
                    unfinished_queue_.pop_back();
                    // undo can be null if it speculatively aborted or if it does no writes
                    if (*speculative->undo() != NULL) {
                        engine_->applyUndo(*speculative->undo());
                    }
                    speculative->prepareReExecute();
                    if (speculative->multiple_partitions()) {
                        multi_partition_queue_.push_front(speculative);
                    } else {
                        single_partition_queue_.push_front(speculative);
                    }
                }

                assert(unfinished_queue_.back() == transaction);
                unfinished_queue_.pop_back();
                if (*transaction->undo() != NULL) {
                    engine_->applyUndo(*transaction->undo());
                }

                found_speculative = true;
                break;
            }
        }
        
        if (!found_speculative) {
            // this is a completed single partition transaction
            assert(!transaction->multiple_partitions() &&
                    transaction->last_status() == ExecutionEngine::ABORT_USER);
            assert(*transaction->undo() == NULL);
        }

        // Look for the transaction in the output queue: can happen with unusual abort timing
        // Add the following check before the loop? if (!transaction->results_sent()) {}
        result_queue_.eraseValue(transaction);
    } else {
        assert(transaction->last_status() == ExecutionEngine::INVALID);
        // this must be a queued transaction
        assert(*transaction->undo() == NULL);
        bool found = false;

        if (transaction->multiple_partitions()) {
            // TODO: support aborting speculatively executed MP transactions
            found = multi_partition_queue_.eraseValue(transaction);
        } else {
            // TODO: support aborting speculatively executed SP transactions
            // TODO: Should we execute aborted single partition transactions anyway?
            found = single_partition_queue_.eraseValue(transaction);
        }
        CHECK(found);
    }

    cleanUpTransaction(transaction);
}
 
 Transaction* OrderedScheduler::getWorkUnitResult() {
    if (result_queue_.empty()) {
        return NULL;
    }
    return result_queue_.dequeue();
}

bool OrderedScheduler::idle() {
    // true if there is more work to do (meaning: we should poll the network, not block)
    bool more = false;
    if (unfinished_queue_.empty()) {
        assert(multi_partition_queue_.empty());
        // Execute single partition transactions without locks or undo
        // NOTE: We limit the number we execute here because when we are replicating, we
        // occasionally need to send replication batches out, read acks and return results to
        // clients. If we don't, we don't get good performance at 0% multi-partition transactions.
        // TODO: Tune this better? For example use a timing deadline?
        int count = 0;
        while (count < MAX_TXNS_PER_LOOP && !single_partition_queue_.empty()) {
            executeFragment(single_partition_queue_.dequeue());
            count += 1;
        }
        more = !single_partition_queue_.empty();
    } else if (cc_mode_ == CC_SPECULATIVE && unfinished_queue_.back()->last_fragment()) {
        // Speculate: all unfinished transactions are on their last fragment
        // Speculate one multi-partition transaction with high priority, if it is on its last
        // fragment. MP transactions are the performance limiter. We want to start the next one
        // ASAP. We can then speculate single partition transactions on top. This is a small
        // improvement when unreplicated, but a large improvement when replicated.
        int count = 0;
        if (!mp_speculated_ && !multi_partition_queue_.empty() &&
                multi_partition_queue_.front()->last_fragment()) {
            mp_speculated_ = true;
            speculate(multi_partition_queue_.dequeue());
            count += 1;
        }

        while (count < MAX_TXNS_PER_LOOP && !single_partition_queue_.empty()) {
            speculate(single_partition_queue_.dequeue());
            count += 1; 
        }

        if (count == 0 && !mp_speculated_ && !multi_partition_queue_.empty()) {
            // This must be a "general" transaction. Since we had nothing else to do, we might
            // as well speculate this. Since this ends speculation, we really want to do this at a
            // low priority.
            assert(!multi_partition_queue_.front()->last_fragment());
            mp_speculated_ = true;
            speculate(multi_partition_queue_.dequeue());
        }

        more = !single_partition_queue_.empty() ||
                (!mp_speculated_ && !multi_partition_queue_.empty());
    }

    return more;
}

void OrderedScheduler::speculate(Transaction* txn) {
    assert(unfinished_queue_.back()->last_fragment());
    // Execute with undo
    ExecutionEngine::Status status = engine_->tryExecute(
            txn->last_work_unit(), txn->output(), txn->undo(), NULL);
    assert(status == ExecutionEngine::OK || status == ExecutionEngine::ABORT_USER);
    txn->setWorkUnitStatus(status);
    if (status == ExecutionEngine::ABORT_USER) {
        // transaction aborted: apply undo immediately so we can continue speculating
        if (*txn->undo() != NULL) {
            engine_->applyUndo(*txn->undo());
            *txn->undo() = NULL;
        }
    }
    unfinished_queue_.push_back(txn);
}

void OrderedScheduler::backupApplyTransaction(const std::vector<std::string>& work_units) {
    string out;
    for (int i = 0; i < work_units.size(); ++i) {
        out.clear();
        ExecutionEngine::Status status = engine_->tryExecute(work_units[i], &out, NULL, NULL);
        CHECK(status == ExecutionEngine::OK);
    }
}

void OrderedScheduler::executeFragment(Transaction* state) {
    ExecutionEngine::Status status = ExecutionEngine::INVALID;
    if (!state->multiple_partitions()) {
        // Execute without undo
        assert(unfinished_queue_.empty());
        status = engine_->tryExecute(state->last_work_unit(), state->output(), NULL, NULL);
    } else {
        assert(state->multiple_partitions());
        assert(unfinished_queue_.empty() || unfinished_queue_.front() == state);
        if (unfinished_queue_.empty()) {
            unfinished_queue_.push_back(state);
        }

        // Execute with undo
        status = engine_->tryExecute(state->last_work_unit(), state->output(), state->undo(), NULL);
    }

    assert(status == ExecutionEngine::OK || status == ExecutionEngine::ABORT_USER);
    state->setWorkUnitStatus(status);
    result_queue_.push_back(state);
}

void OrderedScheduler::cleanUpTransaction(Transaction* transaction) {
    delete transaction;

    // If we have no pending transactions, execute the first multiple partition transaction
    if (unfinished_queue_.empty() && !multi_partition_queue_.empty()) {
        executeFragment(multi_partition_queue_.dequeue());
        assert(!unfinished_queue_.empty());
    }
}

}  // namespace dtxn
