#ifndef DTXN_ORDEREDSCHEDULER_H__
#define DTXN_ORDEREDSCHEDULER_H__

#include "base/circularbuffer.h"
#include "dtxn/transactionscheduler.h"

namespace dtxn {

class ExecutionEngine;

class OrderedScheduler : public TransactionScheduler {
public:
    enum ConcurrencyControl {
        CC_BLOCK,
        CC_SPECULATIVE,
    };

    // Does not own engine.
    OrderedScheduler(ExecutionEngine* engine) : engine_(engine), cc_mode_(CC_BLOCK), mp_speculated_(false) {
        assert(engine_ != NULL);
    }
    virtual ~OrderedScheduler();

    // Set the concurrency control mode.
    void cc_mode(ConcurrencyControl mode);

    virtual Transaction* begin(void* external_state, const std::string& initial_work_unit,
            bool last_fragment, bool multiple_partitions);
    virtual void workUnitAdded(Transaction* transaction);
    virtual void commit(Transaction* transaction);
    virtual void abort(Transaction* transaction);
    virtual Transaction* getWorkUnitResult();
    virtual bool idle();
    virtual void backupApplyTransaction(const std::vector<std::string>& work_units);

    /** The number of single partition transactions that will be executed
    before returning and letting libevent process more network messages. */
    // NOTE: On gigabit ethernet, the round trip time between sending the result of a one shot TXN
    // and receiving the decision was 130/132 us for 2 participants.
    // The htableclientperf transactions execute in ~60 us. Therefore, 2 or 3 transactions is
    // probably close to the right value. In my experiments, 3 transactions performs slightly
    // better than 2 or 4, but this probably needs auto-tuning for different workloads.
    static const int MAX_TXNS_PER_LOOP = 3;

private:
    // Executes the current fragment for state.
    void executeFragment(Transaction* state);

    void cleanUpTransaction(Transaction* transaction);

    // Executes transaction speculatively.
    void speculate(Transaction* transaction);

    ExecutionEngine* engine_;
    ConcurrencyControl cc_mode_;

    typedef CircularBuffer<Transaction*> QueueType;

    // Contains single partition transactions waiting to be executed.
    QueueType single_partition_queue_;

    // Contains multipartition transactions waiting to be executed.
    QueueType multi_partition_queue_;

    // Contains transactions with results can be released to the world via getWorkUnitResult.
    // TODO: Each scheduler ends up implementing queue management crap. Better system?
    QueueType result_queue_;

    // Contains transactions that are pending decision. The head is *actually* being executed.
    // The others are speculative.
    QueueType unfinished_queue_;

    bool mp_speculated_;
};

}  // namespace dtxn

#endif
