#include <algorithm>
#include <limits>

#include "dtxn/executionengine.h"
#include "dtxn/ordered/orderedscheduler.h"
#include "dtxn/transaction.h"
#include "stupidunit/stupidunit.h"

using std::string;
using std::vector;

using namespace dtxn;

class MockExecutionEngine : public ExecutionEngine {
public:
    MockExecutionEngine() : execute_count_(0), commit_(true), undo_count_(0), free_count_(0),
            null_undo_(false) {}

    virtual Status tryExecute(const string& fragment, string* result, void** undo,
            Transaction* transaction) {
        assert(transaction == NULL);
        result->assign(result_);
        execute_count_ += 1;
        if (undo != NULL && !null_undo_) {
            if (*undo != NULL) {
                // this is an existing undo buffer: it must be on top
                assert(!undo_stack_.empty() && undo_stack_.back() == *undo);
                assert(*(int*) *undo == 0);
            } else {
                // We need to provide an undo buffer
                assert(*undo == NULL);
                undo_stack_.push_back(new int(0));
                *undo = undo_stack_.back();
            }
        }

        if (commit_) return OK;
        else return ABORT_USER;
    }

    virtual bool supports_locks() const { return false; }

    virtual void applyUndo(void* undo) {
        assert(undo == undo_stack_.back());
        delete undo_stack_.back();
        undo_stack_.pop_back();
        undo_count_ += 1;
    }

    virtual void freeUndo(void* undo) {
        vector<int*>::iterator i = std::find(undo_stack_.begin(), undo_stack_.end(), undo);
        assert(i != undo_stack_.end());

        delete *i;
        undo_stack_.erase(i);
        free_count_ += 1;
    }

    int execute_count_;
    bool commit_;
    int undo_count_;
    int free_count_;
    string result_;
    vector<int*> undo_stack_;

    // If true, it will not store anything in the undo pointer
    bool null_undo_;
};

class OrderedSchedulerTest : public Test {
public:
    OrderedSchedulerTest() : scheduler_(&engine_) {}

    ~OrderedSchedulerTest() {
        // Clean up transactions that are sitting in the output queue
        Transaction* txn;
        while ((txn = scheduler_.getWorkUnitResult()) != NULL) {
            if (txn->last_status() == ExecutionEngine::OK) {
                scheduler_.commit(txn);
            } else {
                scheduler_.abort(txn);
            }
        }
    }

    MockExecutionEngine engine_;
    OrderedScheduler scheduler_;
};

TEST_F(OrderedSchedulerTest, BadCreate) {
    EXPECT_DEATH(OrderedScheduler(NULL));
}

TEST_F(OrderedSchedulerTest, IdleNothingToDo) {
    // No work queued: idle callback returns false to block on network.
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_FALSE(scheduler_.idle());

    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
}

TEST_F(OrderedSchedulerTest, SinglePartition) {
    engine_.result_ = "out";
    Transaction* txn = scheduler_.begin(this, "foo", true, false);
    EXPECT_EQ(this, txn->external_state());
    EXPECT_EQ("foo", txn->last_work_unit());
    EXPECT_FALSE(txn->multiple_partitions());
    EXPECT_TRUE(txn->last_fragment());

    // Nothing is done until the idle callback
    EXPECT_EQ(0, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // Trigger the idle callback: false = no more work
    EXPECT_FALSE(scheduler_.idle());

    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, txn->last_status());
    EXPECT_EQ("out", *txn->output());

    scheduler_.commit(txn);

    engine_.commit_ = false;
    txn = scheduler_.begin(NULL, "bar", true, false); 
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::ABORT_USER, txn->last_status());

    // can't commit an aborted transaction
    EXPECT_DEATH(scheduler_.commit(txn));
    scheduler_.abort(txn);
}

TEST_F(OrderedSchedulerTest, SinglePartitionLimit) {
    // Queue more transactions than our limit
    for (int i = 0; i < OrderedScheduler::MAX_TXNS_PER_LOOP + 1; ++i) {
        scheduler_.begin(NULL, "foo", true, false);
    }
    // true = more work to be done (poll network)
    EXPECT_TRUE(scheduler_.idle());
    EXPECT_EQ(OrderedScheduler::MAX_TXNS_PER_LOOP, engine_.execute_count_);
}

TEST_F(OrderedSchedulerTest, SpeculateLimit) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);
    // begin a one-shot multi-partition transaction
    /*Transaction* mp = */scheduler_.begin(NULL, "foo", true, true);
    // Queue more single partition transactions than our limit
    for (int i = 0; i < OrderedScheduler::MAX_TXNS_PER_LOOP + 1; ++i) {
        scheduler_.begin(NULL, "foo", true, false);
    }
    // true = more work to be done (poll network)
    EXPECT_TRUE(scheduler_.idle());
    EXPECT_EQ(OrderedScheduler::MAX_TXNS_PER_LOOP+1, engine_.execute_count_);

    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(OrderedScheduler::MAX_TXNS_PER_LOOP+2, engine_.execute_count_);
}

TEST_F(OrderedSchedulerTest, MultiPhaseCommit) {
    Transaction* txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());

    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, txn->last_status());

    // add some work
    txn->resultSent();
    txn->addWorkUnit("bar", false);
    scheduler_.workUnitAdded(txn);
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());

    scheduler_.commit(txn);
    EXPECT_EQ(1, engine_.free_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
}

TEST_F(OrderedSchedulerTest, MultiPhaseAbort) {
    Transaction* txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, txn->last_status());

    // add some work
    engine_.commit_ = false;
    txn->resultSent();
    txn->addWorkUnit("bar", false);
    scheduler_.workUnitAdded(txn);
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    EXPECT_EQ(ExecutionEngine::ABORT_USER, txn->last_status());

    // can't commit aborted transactions
    EXPECT_DEATH(scheduler_.commit(txn));

    scheduler_.abort(txn);
    EXPECT_EQ(1, engine_.undo_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
}

TEST_F(OrderedSchedulerTest, MultiPartitionQueueTransactions) {
    Transaction* txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, txn->last_status());

    // 2 SP transactions are queued
    Transaction* sp1 = scheduler_.begin(NULL, "foo", true, false);
    Transaction* sp2 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // The first transaction commits, the others get executed (after idle)
    scheduler_.commit(txn);
    EXPECT_EQ(1, engine_.execute_count_);

    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(sp1, scheduler_.getWorkUnitResult());
    EXPECT_EQ(sp2, scheduler_.getWorkUnitResult());

    scheduler_.commit(sp1);
    scheduler_.commit(sp2);
}

TEST_F(OrderedSchedulerTest, AbortQueuedSPTransaction) {
    // queue an SP transaction
    Transaction* txn = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_EQ(0, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // trying to commit the queued transaction: death
    EXPECT_DEATH(scheduler_.commit(txn));

    // abort the queued transaction: this works
    scheduler_.abort(txn);

    EXPECT_FALSE(scheduler_.idle());
}

TEST_F(OrderedSchedulerTest, AbortQueuedMPTransaction) {
    Transaction* txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(txn, scheduler_.getWorkUnitResult());

    // an MP transaction is queued
    Transaction* mp = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // trying to commit the queued transaction: death
    EXPECT_DEATH(scheduler_.commit(mp));

    // abort the queued transaction: this works
    scheduler_.abort(mp);

    // The first transaction commits, nothing is executed
    scheduler_.commit(txn);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(1, engine_.execute_count_);

    // Adding another transaction works
    txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(2, engine_.execute_count_);
}

TEST_F(OrderedSchedulerTest, AbortUnsentTransaction) {
    // Start a transaction that is executed
    Transaction* txn = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);

    // Abort the transaction, such as via timeout: removed from the result queue
    // This happens in rare cases when the queue builds up such that the message to start a
    // transaction ends up in the same batch as the message to abort it. We end up executing
    // then immediately aborting. Somewhat wasteful, but this is very rare.
    // TODO: Should DtxnServer handle this, since now each scheduler needs to handle it?
    // TODO: Should be a way to share this test between schedulers?
    scheduler_.abort(txn);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
}

TEST_F(OrderedSchedulerTest, BadCCMode) {
    int bad_cc_value = -1;
    EXPECT_DEATH(scheduler_.cc_mode((OrderedScheduler::ConcurrencyControl) bad_cc_value));

    // Cannot change concurrency while transactions are active
    scheduler_.begin(NULL, "foo", false, true);
    EXPECT_DEATH(scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE));
}

// MP txn, SP txn: both commit
TEST_F(OrderedSchedulerTest, SpeculativeExecutionCommit) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Execute a multiple partition transaction that has the "done" bit set
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    // This will send out a response
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());

    // Receive a transaction from another client: it should be speculatively executed (after idle)
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_EQ(1, engine_.execute_count_);

    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(2, engine_.undo_stack_.size());

    // This will *not* send out a response
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // The transaction commits
    scheduler_.commit(mp);

    // *now* the other client gets a response
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(2, engine_.execute_count_);  // no re-execution
    EXPECT_EQ(0, engine_.undo_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(2, engine_.free_count_);
    scheduler_.commit(sp);
}

// MP txn, SP txn: MP txn aborts, SP txn is reexecuted and commits
TEST_F(OrderedSchedulerTest, SpeculativeExecutionAbort) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Execute a multiple partition transaction that has the "done" bit set
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    // This will send out a response
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());

    // Receive a transaction from another client: it should be speculatively executed (after idle)
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_EQ(1, engine_.execute_count_);

    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(2, engine_.undo_stack_.size());

    // This will *not* send out a response
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // The transaction aborts
    scheduler_.abort(mp);

    // No response or re-execution until idle
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    EXPECT_EQ(2, engine_.execute_count_);

    EXPECT_FALSE(scheduler_.idle());
    // *now* the other client gets a response
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(3, engine_.execute_count_);  // re-execution
    EXPECT_EQ(2, engine_.undo_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    scheduler_.commit(sp);
}

TEST_F(OrderedSchedulerTest, SpeculativeExecutionNullUndo) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Execute a multiple partition transaction that has the "done" bit set, and no undo
    engine_.null_undo_ = true;
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    // This will send out a response
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());

    // Receive a transaction from another client: it should be speculatively executed (after idle)
    // make this transaction speculatively abort
    engine_.commit_ = false;
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    engine_.commit_ = true;
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());

    // The transaction commits
    scheduler_.commit(mp);

    // *now* the other client gets a response
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::ABORT_USER, sp->last_status());
    EXPECT_EQ(2, engine_.execute_count_);  // no re-execution
    EXPECT_EQ(0, engine_.undo_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(0, engine_.free_count_);
    scheduler_.abort(sp);

    // Begin another multi-partition
    mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, mp->last_status());

    // Add a single parition that is speculated on top
    sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(4, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());

    // abort the mp txn
    scheduler_.abort(mp);

    // results should be returned after idle
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(5, engine_.execute_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, sp->last_status());

    // commit it
    scheduler_.commit(sp);
    EXPECT_EQ(0, engine_.free_count_);
}

// MP txn arrives but is not done, SP txn arrives, MP txn finishes, SP is speculatively executed
TEST_F(OrderedSchedulerTest, SpeculativeExecutionQueued) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Start the MP transaction
    Transaction* mp = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    // This will send out a response
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    mp->resultSent();

    // Receive an SP transaction: it must be queued
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());

    // The MP transaction finishes
    mp->addWorkUnit("part2", true);
    scheduler_.workUnitAdded(mp);
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // sp transaction is speculatively executed: no output
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(2, engine_.undo_stack_.size());

    // mp commits: sp is output
    scheduler_.commit(mp);
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    scheduler_.commit(sp);
}

// mp txn 1 starts, mp txn 2 and sp txns are queued and speculated.
TEST_F(OrderedSchedulerTest, SpeculateMPTxn) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Start the one-shot MP transaction
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(1, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    // This will send out a response
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    mp->resultSent();

    // Receive an MP and an SP transaction: MP is speculated then SP speculated.
    Transaction* mp2 = scheduler_.begin(NULL, "foo", true, true);
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(3, engine_.undo_stack_.size());
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // Receive sp and general mp transaction: it is also speculated
    Transaction* sp2 = scheduler_.begin(NULL, "foo", true, false);
    Transaction* mp3 = scheduler_.begin(NULL, "foo", false, true);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(4, engine_.execute_count_);
    EXPECT_EQ(4, engine_.undo_stack_.size());

    // The original MP transaction commits, mp2 is output
    scheduler_.commit(mp);
    EXPECT_EQ(mp2, scheduler_.getWorkUnitResult());
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    EXPECT_EQ(1, engine_.free_count_);
    EXPECT_EQ(3, engine_.undo_stack_.size());

    // received sp; on idle mp3 NOT executed (it is general), but we should poll
    Transaction* sp3 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_TRUE(scheduler_.idle());
    EXPECT_EQ(5, engine_.execute_count_);
    EXPECT_EQ(4, engine_.undo_stack_.size());
    // On idle mp3 is executed
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(6, engine_.execute_count_);
    EXPECT_EQ(5, engine_.undo_stack_.size());

    // another sp: must be queued :(
    Transaction* sp4 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(6, engine_.execute_count_);
    EXPECT_EQ(5, engine_.undo_stack_.size());

    // mp2 commits, sp, sp2, sp3, mp3 are output
    scheduler_.commit(mp2);
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(sp2, scheduler_.getWorkUnitResult());
    EXPECT_EQ(sp3, scheduler_.getWorkUnitResult());
    EXPECT_EQ(mp3, scheduler_.getWorkUnitResult());
    EXPECT_EQ(5, engine_.free_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    scheduler_.commit(sp);
    scheduler_.commit(sp2);
    scheduler_.commit(sp3);

    // still can't execute sp4
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(6, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());

    // commit mp3, sp4 can be executed
    scheduler_.commit(mp3);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    EXPECT_EQ(6, engine_.free_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    EXPECT_EQ(6, engine_.execute_count_);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(7, engine_.execute_count_);
    EXPECT_EQ(sp4, scheduler_.getWorkUnitResult());
    scheduler_.commit(sp4);
}

// mp txn starts, SP txn 1 speculatively aborts, sp txn 2 commits, mp txn commits
TEST_F(OrderedSchedulerTest, SpeculativeAbort) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Start the one-shot MP transaction
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    mp->resultSent();

    // sp transaction speculatively aborts
    engine_.commit_ = false;
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());

    // sp transaction commits
    engine_.commit_ = true;
    Transaction* sp2 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(2, engine_.undo_stack_.size());
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    
    // The MP transaction commits, sp sp2 are output
    scheduler_.commit(mp);
    EXPECT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::ABORT_USER, sp->last_status());
    EXPECT_EQ(sp2, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, sp2->last_status());
    EXPECT_EQ(2, engine_.free_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    
    scheduler_.abort(sp);
    scheduler_.commit(sp2);
}

// mp txn starts, SP txn speculatively aborts, mp txn aborts, sp is re-executed
TEST_F(OrderedSchedulerTest, SpeculativeAbortReexecute) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Start the one-shot MP transaction
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    mp->resultSent();

    // sp transaction speculatively aborts
    engine_.commit_ = false;
    Transaction* sp = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);
    EXPECT_EQ(1, engine_.undo_stack_.size());
    EXPECT_EQ(1, engine_.undo_count_);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    engine_.commit_ = true;

    // The MP transaction aborts, sp is output after idle
    scheduler_.abort(mp);
    EXPECT_EQ(NULL, scheduler_.getWorkUnitResult());
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_EQ(2, engine_.undo_count_);
    EXPECT_EQ(0, engine_.undo_stack_.size());
    ASSERT_EQ(sp, scheduler_.getWorkUnitResult());
    EXPECT_EQ(ExecutionEngine::OK, sp->last_status());   
    scheduler_.commit(sp);
}

// mp txn starts, SP txn speculatively aborts, mp txn aborts, sp is re-executed
TEST_F(OrderedSchedulerTest, AbortSpeculativeMP) {
    scheduler_.cc_mode(OrderedScheduler::CC_SPECULATIVE);

    // Start the one-shot MP transaction
    Transaction* mp = scheduler_.begin(NULL, "foo", true, true);
    EXPECT_EQ(mp, scheduler_.getWorkUnitResult());
    mp->resultSent();

    // sp transaction speculated
    Transaction* sp1 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(2, engine_.execute_count_);

    Transaction* sp2 = scheduler_.begin(NULL, "foo", true, false);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(3, engine_.execute_count_);

    // Abort sp1, sp2 should be re-executed
    scheduler_.abort(sp1);
    EXPECT_EQ(3, engine_.execute_count_);
    EXPECT_FALSE(scheduler_.idle());
    EXPECT_EQ(4, engine_.execute_count_);

    // The MP transaction commits, sp2 is output
    scheduler_.commit(mp);
    EXPECT_EQ(sp2, scheduler_.getWorkUnitResult());
    scheduler_.commit(sp2);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
