#include "htablebenchmark.h"

#include <algorithm>

#include "base/array.h"
#include "base/assert.h"
#include "dtxn/configparser.h"
#include "dtxn/messages.h"
#include "htable/messages.h"
#include "randomgenerator.h"

using std::string;
using std::vector;

namespace htable {

// variables, logical partitions
const DataDefinition DEFAULT_DATA(12, 2);

string makeVariableName(int partition, int client, int variable) {
    string name;
    name.push_back(DataDefinition::idToChar(partition));
    name.push_back(DataDefinition::idToChar(client));
    name.push_back(DataDefinition::idToChar(variable));
    return name;
}

bool LoadTestState::chooseMultiplePartition(RandomGenerator* generator) const {
    int32_t value = generator->random();
    return value <= value_multipartition_;
}

bool LoadTestState::chooseConflict(RandomGenerator* generator) const {
    int32_t value = generator->random();
    return value <= value_conflict_;
}

int32_t LoadTestState::percentToThreshold(int percentage) const {
    assert(0 <= percentage && percentage <= 100);
    int32_t value = static_cast<int32_t>(
            percentage / 100.0 *
            (static_cast<double>(RandomGenerator::maximum()) + 1) - 1);
    assert(-1 <= value && value <= RandomGenerator::maximum());
    return value;
}

static const char WRITE_VALUE[] = "0123";

static void addVariableToMiniTransaction(
        MiniTransaction* mini, int partition, int client, int variable) {
    KeyValuePair p;
    p.key = makeVariableName(partition, client, variable);
    p.hasValue = true;
    p.value = WRITE_VALUE;
    mini->read_set.push_back(p.key);
    mini->write_set.push_back(p);
}

MiniTransaction makeSPTransaction(
        const DataDefinition& data_definition, int partition, int client) {
    MiniTransaction mini;
    // Touches a slice in one partition: num_variables in partition
    for (int i = 0; i < data_definition.num_variables_; ++i) {
        addVariableToMiniTransaction(&mini, partition, client, i);
    }
    return mini;
}

MiniTransaction makeMPTransaction(const DataDefinition& data_definition, int client) {
    const int variables_per_partition =
            data_definition.num_variables_ / data_definition.num_logical_partitions_;
    assert(variables_per_partition * data_definition.num_logical_partitions_ ==
            data_definition.num_variables_);

    // Touches a slice across partitions: variables_per_partition across all partitions
    MiniTransaction mini;
    for (int partition = 0; partition < data_definition.num_logical_partitions_; ++partition) {
        for (int i = 0; i < variables_per_partition; ++i) {
            addVariableToMiniTransaction(&mini, partition, client, i);
        }
    }
    return mini;
}

void removeFromReadSet(
        MiniTransaction* transaction, int partition, int client, int variable) {
    string variable_name = makeVariableName(partition, client, variable);
    vector<string>::iterator key =
            std::find(transaction->read_set.begin(), transaction->read_set.end(), variable_name);
    assert(key != transaction->read_set.end());
    transaction->read_set.erase(key);
}

static void makeConflictingWrite(MiniTransaction* transaction, int partition_id,
        int original_client, int num_logical_partitions) {
    const string original = makeVariableName(partition_id, original_client, 0);

    bool found = false;
    for (int i = 0; i < transaction->write_set.size(); ++i) {
        if (transaction->write_set[i].key == original) {
            // replace the non-conflicting key with the conflicting key
            transaction->write_set[i].key = makeVariableName(partition_id, partition_id, 0);
            found = true;
            break;
        }
    }
    assert(found);
}

static void makeAbort(MiniTransaction* transaction, int partition_id,
        int original_client, int num_logical_partitions) {
    const string original = makeVariableName(partition_id, original_client, 0);

    bool found = false;
    for (int i = 0; i < transaction->write_set.size(); ++i) {
        if (transaction->write_set[i].key == original) {
            // remove the variable from the write set
            transaction->write_set.erase(transaction->write_set.begin() + i);
            found = true;
            break;
        }
    }
    assert(found);

    // add the variable to the conflict set, with a value that will abort
    KeyValuePair pair;
    pair.key = original;
    pair.hasValue = true;
    pair.value = WRITE_VALUE;
    pair.value[0] = '*';
    transaction->compare_set.push_back(pair);
}

// Remove the read to avoid deadlock in conflicting operations (lock upgrades), to ensure
// non-conflicting is the same amount of work. To ensure that a single partition transaction
// is the same work as a multi-partition transaction, and that the work in each partition is
// perfectly even, we remove ONE variable per partition.
static void removeConflictingReadsSingle(
        MiniTransaction* transaction, int partition_id, int client, int num_logical_partitions) {
    // Remove the first num_logical_partition writes
    for (int i = 0; i < num_logical_partitions; ++i) {
        removeFromReadSet(transaction, partition_id, client, i);
    }
}

void removeConflictingReadsMultiple(
        MiniTransaction* transaction, int client, int num_logical_partitions) {
    // Remove one write from each logical partition
    for (int partition = 0; partition < num_logical_partitions; ++partition) {
        removeFromReadSet(transaction, partition, client, 0);
    }
}

void removeConflictingReadsSingles(vector<MiniTransaction>* transactions, int original_client) {
    for (int i = 0; i < transactions->size(); ++i) {
        removeConflictingReadsSingle(
                &(*transactions)[i], i, original_client, (int) transactions->size());
    }
}

vector<MiniTransaction> makeConflictSingles(
        const vector<MiniTransaction>& transactions, int client) {
    // SP and MP transactions must conflict with each other: access the first variable in each
    // partition
    vector<MiniTransaction> conflicts(transactions);
    for (int i = 0; i < conflicts.size(); ++i) {
        makeConflictingWrite(&conflicts[i], i, client, (int) transactions.size());
        removeConflictingReadsSingle(&conflicts[i], i, client, (int) transactions.size());
    }
    return conflicts;
}

vector<MiniTransaction> makeConflictMultiples(
        const vector<MiniTransaction>& transactions, int client) {
    // SP and MP transactions must conflict with each other: access the first variable in each
    // partition
    vector<MiniTransaction> conflicts(transactions);
    for (int i = 0; i < conflicts.size(); ++i) {
        makeConflictingWrite(&conflicts[i], i, client, (int) transactions.size());
        removeConflictingReadsMultiple(&conflicts[i], client, (int) transactions.size());
    }
    return conflicts;
}

static vector<MiniTransaction> makeAbortSingles(
        const vector<MiniTransaction>& transactions, int client) {
    vector<MiniTransaction> aborts(transactions);
    for (int i = 0; i < aborts.size(); ++i) {
        makeAbort(&aborts[i], i, client, (int) aborts.size());
        removeConflictingReadsSingle(&aborts[i], i, client, (int) aborts.size());
    }
    return aborts;
}

static vector<MiniTransaction> makeAbortMultiples(
        const MiniTransaction& multi_partition, int client, int num_partitions) {
    vector<MiniTransaction> aborts(num_partitions, multi_partition);
    for (int i = 0; i < aborts.size(); ++i) {
        makeAbort(&aborts[i], i, client, (int) aborts.size());
        removeConflictingReadsMultiple(&aborts[i], client, (int) aborts.size());
    }
    return aborts;
}

void makeClientWorkload(
        int client, const DataDefinition& data_definition, bool conflicts_enabled,
        vector<MiniTransaction>* single_partitions,
        MiniTransaction* multiple_partition, 
        vector<MiniTransaction>* conflict_singles,
        vector<MiniTransaction>* conflict_multiples,
        vector<MiniTransaction>* abort_singles,
        vector<MiniTransaction>* abort_multiples,
        vector<dtxn::Partition>* partitions) {
    // Create single partition transactions
    for (size_t partition = 0; partition < data_definition.num_logical_partitions_; ++partition) {
        single_partitions->push_back(
                makeSPTransaction(data_definition, (int) partition, client));
    }

    *multiple_partition = makeMPTransaction(DEFAULT_DATA, client);

    *conflict_singles = makeConflictSingles(*single_partitions, client);
    *conflict_multiples = makeConflictMultiples(
            vector<MiniTransaction>(conflict_singles->size(), *multiple_partition),
            client);

    // Add the aborting multi-partition transactions
    *abort_singles = makeAbortSingles(*single_partitions, client);
    *abort_multiples =
            makeAbortMultiples(*multiple_partition, client, data_definition.num_logical_partitions_);

    // Remove the conflicting read from the non-conflicting operation so the work is constant
    removeConflictingReadsSingles(single_partitions, client);
    removeConflictingReadsMultiple(multiple_partition, client,
            data_definition.num_logical_partitions_);
    assert(single_partitions->at(0).read_set.size() == conflict_singles->at(0).read_set.size());
    assert(multiple_partition->read_set.size() == conflict_singles->at(0).read_set.size());
    assert(conflict_multiples->at(0).read_set.size() == conflict_singles->at(0).read_set.size());

    // HACK: Used to direct "conflicting background" transactions to the right partition
    if (conflicts_enabled && client < data_definition.num_logical_partitions_) {
        // the first logical partition clients always write to a single partition
        // Remove all but that partition, and all but requests for that partition
        int physical_partition =
                data_definition.partitionFunction((int) partitions->size(), client);
        dtxn::Partition copy = partitions->at(physical_partition);
        partitions->clear();
        partitions->push_back(copy);

        vector<MiniTransaction>* lists[] = {
            single_partitions, conflict_singles, conflict_multiples,
        };
        for (int i = 0; i < base::arraySize(lists); ++i) {
            assert(lists[i]->size() == data_definition.num_logical_partitions_);
            MiniTransaction mini = (*lists[i])[client];
            lists[i]->clear();
            lists[i]->push_back(mini);
        }
    }

    //~ printf("\nclient %d\n", client);
    //~ printf("sp: %s\n", debugString(single_partitions).c_str());
    //~ printf("mp: %s\n", debugString(multiple_partition).c_str());
    //~ printf("conflict sp: %s\n", debugString(conflict_singles).c_str());
    //~ printf("conflict mp: %s\n", debugString(conflict_multiples).c_str());
}

string toString(const MiniTransaction& mini) {
    string serialize;
    mini.appendToString(&serialize);
    return serialize;
}

vector<string> toString(const vector<MiniTransaction>& v) {
    vector<string> result;
    for (int i = 0; i < v.size(); ++i) {
        result.push_back(toString(v[i]));
    }
    return result;
}

dtxn::Fragment toTransaction(const MiniTransaction& mini) {
    dtxn::Fragment txn;
    mini.appendToString(&txn.transaction);
    return txn;
}

vector<dtxn::Fragment> toTransaction(const vector<MiniTransaction>& v) {
    vector<dtxn::Fragment> result;
    for (int i = 0; i < v.size(); ++i) {
        result.push_back(toTransaction(v[i]));
    }
    return result;
}

string debugString(const MiniTransaction& mini) {
    string s;
    if (!mini.compare_set.empty()) {
        s = "compare:";
        for (int i = 0; i < mini.compare_set.size(); ++i) {
            s.push_back(' ');
            s.append(mini.compare_set[i].key);
        }
        s.append(" ");
    }

    s.append("read:");
    for (int i = 0; i < mini.read_set.size(); ++i) {
        s.push_back(' ');
        s.append(mini.read_set[i]);
    }

    s.append(" write:");
    for (int i = 0; i < mini.write_set.size(); ++i) {
        s.push_back(' ');
        s.append(mini.write_set[i].key);
    }
    return s;
}

string debugString(const std::vector<MiniTransaction>& v) {
    string s;
    for (int i = 0; i < v.size(); ++i) {
        s.append(debugString(v[i]));
        s.append("; ");
    }
    return s;
}

}  // namespace htable
