#include <string>

#include "base/stlutil.h"
#include "benchmark/benchmark.h"
#include "coordinatorworker.h"
#include "dtxn/configparser.h"
#include "dtxn/messages.h"
#include "dtxn/ordered/ordereddistributor.h"
#include "htablebenchmark.h"
#include "htable/messages.h"
#include "io/libeventloop.h"
#include "libevent/event.h"
#include "minitransactiondistributor.h"
#include "randomgenerator.h"

using std::string;
using std::vector;

using namespace htable;

class ClientLoadThread : public benchmark::BenchmarkWorker, public CoordinatorWorker {
public:
    ClientLoadThread(const vector<dtxn::Partition>& partitions,
            LoadTestState* test_state,
            const vector<dtxn::Fragment>& single_partitions,
            const dtxn::Fragment& multiple_partition,
            const vector<dtxn::Fragment>& conflict_single_partitions,
            const vector<dtxn::Fragment>& conflict_multiple_partitions) :

            CoordinatorWorker(partitions, new MiniTransactionDistributor(
                    DEFAULT_DATA.num_logical_partitions_, (int) partitions.size())),
            test_state_(test_state),
            pending_(0),
            measuring_(false),
            id_(0),
            next_client_(0),
            transaction_count_(0),
            did_measure_(false),
            single_partitions_(single_partitions),
            multiple_partition_(multiple_partition),
            conflict_single_partitions_(conflict_single_partitions),
            conflict_multiple_partitions_(conflict_multiple_partitions) {
        assert(test_state_ != NULL);
        assert(single_partitions_.size() > 0);

        // Select a random initial client
        generator_.seed((unsigned int) (time(NULL) + pthread_self()));
        next_client_ = static_cast<int>(generator_.random() % single_partitions_.size());
    }

    virtual void response(const void* data, int length) {
        const char* buffer = reinterpret_cast<const char*>(data);
        dtxn::FragmentResponse result;
        // type, length prefixed
        const char* end = result.parseFromString(buffer + sizeof(int32_t) * 2, buffer + length);
        ASSERT(end == buffer + length);

        // Don't count deadlock aborted transactions
        bool txn_success = result.status == dtxn::ExecutionEngine::OK ||
                result.status == dtxn::ExecutionEngine::ABORT_USER;
        if (txn_success && measuring_) {
            did_measure_ = true;
            transaction_count_ += 1;
        }
        pending_ -= 1;
    }

    virtual void initialize() {}

    virtual void work(bool measure) {
        measuring_ = measure;

        if (pending_ == 0) {
            dtxn::Fragment* txn = NULL;

            if (test_state_->chooseMultiplePartition(&generator_)) {
                if (test_state_->chooseConflict(&generator_)) {
                    txn = &conflict_multiple_partitions_[next_client_];
                } else {
                    txn = &multiple_partition_;
                }
            } else {
                if (test_state_->chooseConflict(&generator_)) {
                    txn = &conflict_single_partitions_[next_client_];
                } else {
                    txn = &single_partitions_[next_client_];
                }
            }

            // This advances the client for non-conflicting multi-partition transactions. Does this
            // matter? It leads to a small but uniformly distributed "double sp txn".
            next_client_ += 1;
            if (next_client_ == single_partitions_.size()) {
                next_client_ = 0;
            }

            txn->id = id_;
            requestReceived(*txn);
            id_ += 1;
            pending_ += 1;

            event_base_loop(event_loop()->base(), EVLOOP_ONCE);
        }

        while (pending_ > 0) {
            event_base_loop(event_loop()->base(), EVLOOP_ONCE);
        }
    }

    virtual void finish() {
        assert(pending_ == 0);
    }

    int transaction_count() const { return transaction_count_; }

private:
    LoadTestState* test_state_;

    int pending_;
    bool measuring_;

    //~ vector<int> client_counts_;
    //~ int distributor_counts_;

    RandomGenerator generator_;

    int32_t id_;
    int next_client_;
    int transaction_count_;
    bool did_measure_;

    vector<dtxn::Fragment> single_partitions_;
    dtxn::Fragment multiple_partition_;

    vector<dtxn::Fragment> conflict_single_partitions_;
    vector<dtxn::Fragment> conflict_multiple_partitions_;
};

int main(int argc, const char* argv[]) {
    if (argc != 5 && argc != 6) {
        fprintf(stderr, "htableclientperf-lock [config file] [clients] [percent multipartition] [percent conflict|0 percent abort]\n");
        return 1;
    }

    // Read and parse the configuration file.
    vector<dtxn::Partition> partitions = dtxn::parseConfigurationFromPath(argv[1]);

    int clients = atoi(argv[2]);
    int percent_multipartition = atoi(argv[3]);
    int percent_conflict = atoi(argv[4]);
    bool scale_conflicts = (argc == 5);
    int percent_aborts = 0;
    if (!scale_conflicts) {
        assert(percent_conflict == 0);
        percent_aborts = atoi(argv[5]);
    }
    LoadTestState test_state(percent_multipartition, scale_conflicts ? percent_conflict : percent_aborts);

    // Need to have enough clients to generate "conflicting" load
    assert(!scale_conflicts || (percent_conflict == 0 || clients >= DEFAULT_DATA.num_logical_partitions_ * 2));

    // Create threads with the partitions
    vector<ClientLoadThread*> threads;
    for (int client = 0; client < clients; ++client) {
        vector<MiniTransaction> single_partitions;
        MiniTransaction multiple_partition;
        vector<MiniTransaction> conflict_singles;
        vector<MiniTransaction> conflict_multiples;
        vector<MiniTransaction> abort_singles;
        vector<MiniTransaction> abort_multiples;
        vector<dtxn::Partition> partitions_copy(partitions);
        makeClientWorkload(client, DEFAULT_DATA, percent_conflict > 0,
                &single_partitions, &multiple_partition, &conflict_singles,
                &conflict_multiples, &abort_singles, &abort_multiples, &partitions_copy);

        // HACK; if we are running the abort test, replace the conficts with aborts
        if (!scale_conflicts) {
            conflict_singles = abort_singles;
            conflict_multiples = abort_multiples;
        }

        threads.push_back(
                new ClientLoadThread(partitions_copy, &test_state,
                        toTransaction(single_partitions), toTransaction(multiple_partition),
                        toTransaction(conflict_singles), toTransaction(conflict_multiples)));
    }

    int64_t us = benchmark::runThreadBench(WARM_UP_SECONDS, TEST_SECONDS, threads);

    // Wait for the workers to finish
    int total_count = 0;
    for (int i = 0; i < threads.size(); ++i) {
        total_count += threads[i]->transaction_count();
    }

    double seconds = (double) us/1000000.0;
    printf("%f\n", total_count/seconds);

    STLDeleteElements(&threads);
    return 0;
}
