#include <cstdio>
#include <cstdlib>

#include <cassert>
#include <string>

#include "base/stlutil.h"
#include "dtxn/common.h"
#include "dtxn/configparser.h"
#include "dtxn/executionengine.h"
#include "dtxn/locks/lockscheduler.h"
#include "dtxn/occ/occscheduler.h"
#include "dtxn/ordered/orderedscheduler.h"
#include "htable/messages.h"
#include "minitransactions/keyvalueengine.h"
#include "minitransactions/lockingkeyvaluetable.h"
#include "minitransactions/minitransactions.h"

using minitransactions::KeyValueEngine;
using std::string;
using std::vector;

// TODO: Add unit tests for this!
class MiniTransactionEngine : public dtxn::ExecutionEngine {
public:
    MiniTransactionEngine(bool locking) : locking_(locking) {
        if (locking) {
            engine_.table(new minitransactions::LockingKeyValueTable());
        } else {
            engine_.table(new minitransactions::SimpleKeyValueTable());
        }
    }

    virtual Status tryExecute(const string& work_unit, string* output, void** undo,
            dtxn::Transaction* transaction) {
        assert(locking_ || transaction == NULL);
        assert(output->empty());

        minitransactions::MiniTransaction miniwork_unit;
        parseFromString(work_unit, &miniwork_unit);

        minitransactions::KeyValueSet result;
        minitransactions::KeyValueSet* undo_record = NULL;
        if (undo != NULL) {
            if (*undo != NULL) {
                undo_record = (minitransactions::KeyValueSet*) *undo;
            } else {
                // TODO: Allocate on demand? Not needed for read-only transactions or if it aborts
                undo_record = new minitransactions::KeyValueSet();
                *undo = undo_record;
            }
        }
        KeyValueEngine::Status status =
                engine_.tryExecute(miniwork_unit, &result, undo_record, transaction);
        if (status != KeyValueEngine::BLOCKED) {
            htable::MiniResult result_message;
            for (minitransactions::KeyValueSet::const_iterator i = result.begin(); i != result.end(); ++i) {
                htable::KeyValuePair p;
                p.key = i->first;
                p.hasValue = i->second != NULL;
                if (p.hasValue) p.value = *i->second;
                result_message.keyvalues.push_back(p);
            }
            result_message.appendToString(output);
        }

        // If we are blocked, this contains partial results we will re-read (yes, wasteful, but...)
        STLDeleteValues(&result);

        if (status == KeyValueEngine::OK) {
            return OK;
        } else if (status == KeyValueEngine::ABORT) {
            return ABORT_USER;
        } else {
            assert(status == KeyValueEngine::BLOCKED);
            return BLOCKED;
        }
    }

    virtual bool supports_locks() const { return locking_; }

    virtual void applyUndo(void* undo) {
        assert(undo != NULL);
        minitransactions::KeyValueSet* undo_record = reinterpret_cast<minitransactions::KeyValueSet*>(undo);
        engine_.applyUndo(*undo_record);
        freeUndo(undo);
    }

    virtual void freeUndo(void* undo) {
        assert(undo != NULL);
        minitransactions::KeyValueSet* undo_record = reinterpret_cast<minitransactions::KeyValueSet*>(undo);
        STLDeleteValues(undo_record);
        delete undo_record;
    }

    void print() const { engine_.print(); }

private:
    /** Parses the string into a MiniTransaction. */
    void parseFromString(const string& data, minitransactions::MiniTransaction* output) {
        // The data in the string is serialized into a htable::MiniTransaction.
        htable::MiniTransaction mini;
        mini.parseFromString(data.data(), data.data() + data.size());

        // Parse out the read, write, and compare sets
        for (int i = 0; i < mini.read_set.size(); ++i) {
            output->read(mini.read_set[i]);
        }
        for (int i = 0; i < mini.write_set.size(); ++i) {
            if (mini.write_set[i].hasValue) {
                output->write(mini.write_set[i].key, mini.write_set[i].value);
            } else {
                output->erase(mini.write_set[i].key);
            }
        }
        for (int i = 0; i < mini.compare_set.size(); ++i) {
            if (mini.compare_set[i].hasValue) {
                output->compare(mini.compare_set[i].key, &(mini.compare_set[i].value));
            } else {
                output->compare(mini.compare_set[i].key, NULL);
            }
        }
    }

    bool locking_;
    minitransactions::KeyValueEngine engine_;
};

int main(int argc, const char* argv[]) {
    if (argc != 5) {
        fprintf(stderr, "htableserver [mode = block|spec|lock|occ] [configuration] [partition index] [replica index]\n");
        return 1;
    }
    const char* mode_string = argv[1];
    const char* config_file = argv[2];
    int partition_index = atoi(argv[3]);
    int replica_index = atoi(argv[4]);

    enum Mode {
        BLOCKING,
        SPECULATE,
        LOCKING,
        OCC,
    };
    Mode mode = LOCKING;
    if (strcmp(mode_string, "block") == 0) {
        mode = BLOCKING;
    } else if (strcmp(mode_string, "spec") == 0) {
        mode = SPECULATE;
    } else if (strcmp(mode_string, "lock") == 0) {
        mode = LOCKING;
    } else if (strcmp(mode_string, "occ") == 0) {
        mode = OCC;
    } else {
        fprintf(stderr, "mode must be block|spec|lock|occ\n");
        return 1;
    }

    vector<dtxn::Partition> partitions = dtxn::parseConfigurationFromPath(config_file);
    assert(0 <= partition_index && partition_index < partitions.size());
    assert(0 <= replica_index && replica_index < partitions[partition_index].numReplicas());

    MiniTransactionEngine engine(mode == LOCKING || mode == OCC);

    dtxn::TransactionScheduler* scheduler = NULL;
    if (mode == LOCKING) {
        scheduler = new dtxn::LockScheduler(&engine);
    } else if (mode == OCC) {
        scheduler = new dtxn::OCCScheduler(&engine);
    } else {
        assert(mode == BLOCKING || mode == SPECULATE);
        scheduler = new dtxn::OrderedScheduler(&engine);
        if (mode == SPECULATE) {
            ((dtxn::OrderedScheduler*) scheduler)->cc_mode(dtxn::OrderedScheduler::CC_SPECULATIVE);
        } else {
            assert(mode == BLOCKING);
        }
    }

    dtxn::runServer(scheduler, partitions[partition_index], replica_index);

    // Print the contents of the table
    engine.print();

    return 0;
}
