#include "io/buffer.h"

#include <cassert>
#include <cstring>

#include <stdint.h>

#include "base/cast.h"
#include "base/stlutil.h"

using std::vector;

namespace io {

static const int PAGE_SIZE = 4096;
struct Page {
    Page() : filled(0) {}

    // Number of bytes of data that have actually been filled
    int filled;
    char data[PAGE_SIZE - sizeof(int)];
};

MemoryOutputStream::MemoryOutputStream() : read_position_(0), available_(0) {
    pages_.push_back(new Page());
}

MemoryOutputStream::~MemoryOutputStream() {
    STLDeleteElements(&pages_);
}

int MemoryOutputStream::write(const char* buffer, size_t length) {
    int remaining_bytes = assert_range_cast<int>(length);

    while (remaining_bytes > 0) {
        // Figure out the space remaining on this page
        Page* current = pages_.back();
        int to_copy = static_cast<int>(sizeof(current->data)) - current->filled;
        assert(to_copy >= 0);
        if (to_copy == 0) {
            // The page is full: add a new one
            current = new Page();
            pages_.push_back(current);
            to_copy = sizeof(current->data);
        }

        // Copy the data into the page
        to_copy = std::min(to_copy, remaining_bytes);
        char* page_end = current->data + current->filled;
        memcpy(page_end, buffer, to_copy);

        // Adjust all the counters
        buffer += to_copy;
        current->filled += to_copy;
        remaining_bytes -= to_copy;
        available_ += to_copy;
        assert(current->filled <= sizeof(current->data));
        assert(remaining_bytes >= 0);
    }
    assert(remaining_bytes == 0);

    return assert_range_cast<int>(length);
}

int MemoryOutputStream::flush(OutputStream* out) {
    int bytes_copied = 0;

    while (available_ > 0) {
        // Try to write everything in the first block
        int remaining_in_block = pages_.front()->filled - read_position_;
        int bytes_written = out->write(pages_.front()->data + read_position_, remaining_in_block);
        assert(bytes_written <= remaining_in_block);
        if (bytes_written == -1) {
            // An error occurred while writing: return the error
            return -1;
        }
        assert(bytes_written >= 0);

        // Adjust the byte counters
        read_position_ += bytes_written;
        assert(read_position_ <= pages_.front()->filled);
        bytes_copied += bytes_written;
        available_ -= bytes_written;
        assert(available_ >= 0);

        if (bytes_written < remaining_in_block) {
            // The other output stream is full: stop writing
            break;
        }
        assert(bytes_written == remaining_in_block);

        if (read_position_ == pages_.front()->filled) {
            // Consumed an entire block: remove it
            if (pages_.size() > 1) {
                delete pages_.front();
                pages_.erase(pages_.begin());
            } else {
                // Last block: just reset the counter
                assert(pages_.size() == 1);
                pages_.front()->filled = 0;
                assert(available_ == 0);
            }
            read_position_ = 0;
        }
    }

    return bytes_copied;
}


struct FIFOBuffer::Page {
    int write_position_;
    Page* next_;
    uint8_t data_[0];
};

const int FIFOBuffer::PAGE_DATA_SIZE = PAGE_SIZE - sizeof(FIFOBuffer::Page);

FIFOBuffer::FIFOBuffer() : read_position_(0), available_(0) {
    Page* p = allocatePage();
    head_ = p;
    tail_ = p;
}

FIFOBuffer::~FIFOBuffer() {
    while (head_ != NULL) {
        popPage();
    }
}

void FIFOBuffer::writeBuffer(void** out_buffer, int* out_length) {
    if (head_ == tail_ && read_position_ == tail_->write_position_) {
        // last read consumed all the data: reset the read and write positions
        read_position_ = 0;
        head_->write_position_ = 0;
    }

    int remaining_bytes = PAGE_DATA_SIZE - tail_->write_position_;
    if (remaining_bytes == 0) {
        Page* p = allocatePage();
        tail_->next_ = p;
        tail_ = p;
        remaining_bytes = PAGE_DATA_SIZE;
    }

    assert(remaining_bytes > 0);
    *out_buffer = tail_->data_ + tail_->write_position_;
    *out_length = remaining_bytes;
    tail_->write_position_ = PAGE_DATA_SIZE;
    available_ += remaining_bytes;
}

void FIFOBuffer::undoWrite(int length) {
    assert(0 <= length && length <= tail_->write_position_);
    tail_->write_position_ -= length;
    assert(0 <= tail_->write_position_ && tail_->write_position_ <= PAGE_DATA_SIZE);
    available_ -= length;
    assert(available_ >= 0);
}

void FIFOBuffer::copyIn(const void* data, int length) {
    assert(length >= 0);

    const uint8_t* bytes = reinterpret_cast<const uint8_t*>(data);
    while (length > 0) {
        void* write;
        int remaining;
        writeBuffer(&write, &remaining);

        int written = std::min(length, remaining);
        memcpy(write, bytes, written);
        length -= written;
        bytes += written;

        // If we didn't fill the write buffer, undo the last bit
        if (written < remaining) {
            assert(length == 0);
            undoWrite(remaining - written);
        }
    }
}

void* FIFOBuffer::writeExact(int length) {
    assert(length > 0);

    int remaining_bytes = PAGE_DATA_SIZE - tail_->write_position_;
    if (remaining_bytes < length) {
        Page* p = allocatePage();
        tail_->next_ = p;
        tail_ = p;
        remaining_bytes = PAGE_DATA_SIZE;
    }

    assert(remaining_bytes >= length);
    void* out_buffer = tail_->data_ + tail_->write_position_;
    tail_->write_position_ += length;
    available_ += length;
    return out_buffer;
}

void FIFOBuffer::readBuffer(const void** out_buffer, int* out_length) {
    // If the current page is empty, reset or discard it
    if (read_position_ == head_->write_position_) {
        read_position_ = 0;
        if (head_ != tail_) {
            // more than one page: discard this page
            popPage();
        } else {
            // reset the write position: empty!
            head_->write_position_ = 0;
        }
    }

    int remaining_bytes = head_->write_position_ - read_position_;
    assert(0 <= remaining_bytes && remaining_bytes <= available_);
    *out_length = remaining_bytes;
    if (remaining_bytes == 0) {
        *out_buffer = NULL;
    } else {
        *out_buffer = head_->data_ + read_position_;
    }
    read_position_ += remaining_bytes;
    available_ -= remaining_bytes;
}

void FIFOBuffer::undoRead(int length) {
    assert(0 <= length && length <= read_position_);
    read_position_ -= length;
    assert(0 <= read_position_ && read_position_ <= PAGE_DATA_SIZE);
    available_ += length;
}

void FIFOBuffer::copyOut(std::string* out) {
    // Read and assign the first chunk
    const void* read;
    int length;
    readBuffer(&read, &length);
    out->assign((const char*) read, length);

    // append any other chunks
    readBuffer(&read, &length);
    while (length > 0) {
        out->append((const char*) read, length);
        readBuffer(&read, &length);
    }
}

void FIFOBuffer::clear() {
    // If we have more than one page, discard them
    while (head_ != tail_) {
        popPage();
    }
    assert(head_ == tail_);
    read_position_ = 0;
    head_->write_position_ = 0;
    available_ = 0;
}

FIFOBuffer::Page* FIFOBuffer::allocatePage() const {
    // Use raw ::operator new and ::operator delete to allocate uninitialized memory
    // TODO: Use std::allocator?
    Page* page = (Page*) ::operator new(PAGE_SIZE);
    page->write_position_ = 0;
    page->next_ = NULL;
    return page;
}

void FIFOBuffer::popPage() {
    Page* previous = head_;
    head_ = head_->next_;
    ::operator delete(previous);
}

}
