#ifndef IO_BUFFER_H__
#define IO_BUFFER_H__

#include <string>
#include <vector>

#include "io/io.h"

namespace io {

struct Page;

// Implements OutputStream to memory blocks.
class MemoryOutputStream : public OutputStream {
public:
    MemoryOutputStream();
    virtual ~MemoryOutputStream();

    // Writes length bytes into the buffer. This always copies all the data.
    virtual int write(const char* buffer, size_t length);

    // Returns a pointer to length bytes in the buffer, and advances the write position by the
    // same amount. This is a very low-level routine that should only be used for writing length
    // prefixed messages. If it is not used carefully, garbage could be written into the buffer.
    char* reserve(size_t length); 

    // Write as much data as possible to out. Returns the number of bytes written,
    // or -1 if out.write() returns -1.
    int flush(OutputStream* out);

    // Number of bytes available to be flushed to another OutputStream.
    int available() const { return available_; }

private:
    std::vector<Page*> pages_;

    // The position that flush() will copy from, in the first page
    int read_position_;
    int available_;
};


// Provides zero-copy reads and writes to a set of internal buffers.
class FIFOBuffer {
public:
    FIFOBuffer();
    ~FIFOBuffer();

    // data will point to the beginning of the next chunk of data to be written in the buffer.
    // length will be the number of bytes available to be written.
    void writeBuffer(void** data, int* length);

    // Push length bytes back into the write end of the buffer. Used when writeBuffer() returns
    // more data than you want to write. length must be <= length returned by the last call to
    // writeBuffer().
    void undoWrite(int length);

    // Returns a pointer to a buffer available for writing exactly length bytes of data. This
    // must only be used to write "small" chunks, since it may leave some wasted bytes at the end
    // of a buffer. This simplifies serialization code at the cost of one int per page to record
    // the last bytes written.
    void* writeExact(int length);

    // Copies length bytes from data into this FIFO.
    void copyIn(const void* data, int length);

    // data will point to the beginning of the next chunk of data in the buffer. length will be the
    // number of bytes available.
    void readBuffer(const void** data, int* length);

    // Push length bytes back into the read end of the buffer. Used when readBuffer() returns more
    // data than you want to consume. length must be <= length returned by the last call to
    // readBuffer().
    void undoRead(int length);

    // Copies all data to out.
    void copyOut(std::string* out);

    // Returns the number of bytes available for reading in the buffer.
    // TODO: Don't cache the number of available bytes?
    int available() const { return available_; }

    // Removes all data from the FIFO, resetting it to empty.
    void clear();

    // Allocate pages of this size
    static const int PAGE_SIZE = 8192;
    static const int PAGE_DATA_SIZE;

private:
    struct Page;
    Page* allocatePage() const;
    void popPage();

    Page* head_;
    Page* tail_;

    int read_position_;
    int available_;
};

}

#endif
