#include "htable/messages.h"
#include "minitransactiondistributor.h"
#include "strings/utils.h"
#include "stupidunit/stupidunit.h"

using dtxn::CoordinatedTransaction;
using std::string;

static const int LOGICAL_PARTITIONS = 2;
static const int PHYSICAL_PARTITIONS = LOGICAL_PARTITIONS;

class MiniTransactionDistributorTest : public Test {
public:
    MiniTransactionDistributorTest() :
        coordinator_(LOGICAL_PARTITIONS, PHYSICAL_PARTITIONS),
        txn_(NULL) {}

    ~MiniTransactionDistributorTest() {
        coordinator_.done(txn_);
        delete txn_;
    }

    void makeTxn() {
        string out;
        mini_.appendToString(&out);
        assert(txn_ == NULL);
        txn_ = new CoordinatedTransaction(2, out);
    }

    void receiveResponse(int partition, bool commit) {
        string out;
        response_.appendToString(&out);
        dtxn::ExecutionEngine::Status status = dtxn::ExecutionEngine::OK;
        if (!commit) status = dtxn::ExecutionEngine::ABORT_USER;
        txn_->receive(partition, out, status);
    }

    MiniTransactionDistributor coordinator_;
    htable::MiniTransaction mini_;
    htable::MiniResult response_;
    CoordinatedTransaction* txn_;
};

static const char FIRST_PARTITION_KEY[] = "A";
static const char SECOND_PARTITION_KEY[] = "B";

TEST_F(MiniTransactionDistributorTest, SinglePartition) {
    mini_.read_set.push_back(FIRST_PARTITION_KEY);
    makeTxn();
    coordinator_.begin(txn_);

    EXPECT_EQ(1, txn_->sent().size());
    EXPECT_EQ(0, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(FIRST_PARTITION_KEY, mini_.read_set[0]);
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    txn_->sentMessages();
    receiveResponse(0, true);

    coordinator_.nextRound(txn_);
    EXPECT_TRUE(txn_->haveClientResponse());
    response_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(0, response_.keyvalues.size());
}

TEST_F(MiniTransactionDistributorTest, MultiplePartition) {
    mini_.read_set.push_back(FIRST_PARTITION_KEY);
    mini_.read_set.push_back(SECOND_PARTITION_KEY);
    makeTxn();
    coordinator_.begin(txn_);

    EXPECT_EQ(2, txn_->sent().size());
    EXPECT_EQ(0, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(FIRST_PARTITION_KEY, mini_.read_set[0]);
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    EXPECT_EQ(1, txn_->sent()[1].first);
    mini_.parseFromString(txn_->sent()[1].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(SECOND_PARTITION_KEY, mini_.read_set[0]);
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    txn_->sentMessages();
    response_.keyvalues.resize(1);
    response_.keyvalues[0].key = FIRST_PARTITION_KEY;
    response_.keyvalues[0].value = "foo";
    receiveResponse(0, true);
    response_.keyvalues[0].key = SECOND_PARTITION_KEY;
    receiveResponse(1, true);

    coordinator_.nextRound(txn_);
    
    EXPECT_TRUE(txn_->haveClientResponse());
    response_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(2, response_.keyvalues.size());
}

TEST_F(MiniTransactionDistributorTest, ResetMax) {
    mini_.read_set.push_back(FIRST_PARTITION_KEY);
    mini_.read_set.push_back(SECOND_PARTITION_KEY);
    mini_.compare_set.resize(1);
    mini_.compare_set[0].key = MiniTransactionDistributor::RESET_MAXIMUM_KEY;
    makeTxn();
    coordinator_.begin(txn_);

    // reads all the variables
    ASSERT_EQ(2, txn_->sent().size());
    EXPECT_EQ(0, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    EXPECT_EQ(1, txn_->sent()[1].first);
    mini_.parseFromString(txn_->sent()[1].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    txn_->sentMessages();
    response_.keyvalues.resize(1);
    response_.keyvalues[0].key = FIRST_PARTITION_KEY;
    response_.keyvalues[0].hasValue = true;
    int32_t value = -2;
    strings::assignBytes(&response_.keyvalues[0].value, value);
    receiveResponse(0, true);
    response_.keyvalues[0].key = SECOND_PARTITION_KEY;
    value = 5;
    strings::assignBytes(&response_.keyvalues[0].value, value);
    receiveResponse(1, true);

    // One message gets sent out
    coordinator_.nextRound(txn_);
    EXPECT_EQ(1, txn_->sent().size());
    EXPECT_EQ(1, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(0, mini_.read_set.size());
    EXPECT_EQ(1, mini_.write_set.size());
    EXPECT_EQ(SECOND_PARTITION_KEY, mini_.write_set[0].key);
    EXPECT_EQ(1, mini_.compare_set.size());

    // ack comes back
    txn_->sentMessages();
    response_.keyvalues.clear();
    receiveResponse(1, true);
    coordinator_.nextRound(txn_);

    EXPECT_TRUE(txn_->haveClientResponse());
    response_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(1, response_.keyvalues.size());
    EXPECT_EQ(SECOND_PARTITION_KEY, response_.keyvalues[0].key);
}

TEST_F(MiniTransactionDistributorTest, TwoRound) {
    coordinator_.set_two_round(true);
    mini_.read_set.push_back(FIRST_PARTITION_KEY);
    mini_.read_set.push_back(SECOND_PARTITION_KEY);
    mini_.write_set.resize(2);
    mini_.write_set[0].key = FIRST_PARTITION_KEY;
    mini_.write_set[0].value = "bar";
    mini_.write_set[1].key = SECOND_PARTITION_KEY;
    mini_.write_set[1].value = "baz";
    makeTxn();
    coordinator_.begin(txn_);

    // Validate the reads
    EXPECT_EQ(2, txn_->sent().size());
    EXPECT_EQ(0, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(FIRST_PARTITION_KEY, mini_.read_set[0]);
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    EXPECT_EQ(1, txn_->sent()[1].first);
    mini_.parseFromString(txn_->sent()[1].second);
    EXPECT_EQ(1, mini_.read_set.size());
    EXPECT_EQ(SECOND_PARTITION_KEY, mini_.read_set[0]);
    EXPECT_EQ(0, mini_.write_set.size());
    EXPECT_EQ(0, mini_.compare_set.size());

    txn_->sentMessages();
    response_.keyvalues.resize(1);
    response_.keyvalues[0].key = FIRST_PARTITION_KEY;
    response_.keyvalues[0].value = "foo";
    receiveResponse(0, true);
    response_.keyvalues[0].key = SECOND_PARTITION_KEY;
    receiveResponse(1, true);

    // Validate the writes
    coordinator_.nextRound(txn_);
    EXPECT_EQ(2, txn_->sent().size());
    EXPECT_EQ(0, txn_->sent()[0].first);
    mini_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(0, mini_.read_set.size());
    EXPECT_EQ(1, mini_.write_set.size());
    EXPECT_EQ(FIRST_PARTITION_KEY, mini_.write_set[0].key);
    EXPECT_EQ("bar", mini_.write_set[0].value);
    EXPECT_EQ(0, mini_.compare_set.size());

    EXPECT_EQ(1, txn_->sent()[1].first);
    mini_.parseFromString(txn_->sent()[1].second);
    EXPECT_EQ(0, mini_.read_set.size());
    EXPECT_EQ(1, mini_.write_set.size());
    EXPECT_EQ(SECOND_PARTITION_KEY, mini_.write_set[0].key);
    EXPECT_EQ("baz", mini_.write_set[0].value);
    EXPECT_EQ(0, mini_.compare_set.size());

    // Receive the last responses: we are done!
    txn_->sentMessages();
    response_.keyvalues.resize(0);
    receiveResponse(0, true);
    receiveResponse(1, true);
    coordinator_.nextRound(txn_);
    ASSERT_TRUE(txn_->haveClientResponse());
    response_.parseFromString(txn_->sent()[0].second);
    EXPECT_EQ(2, response_.keyvalues.size());
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
