#include <cstdio>
#include <cstdlib>
#include <tr1/functional>

#include <netinet/in.h>

#include "base/circularbuffer.h"
#include "base/stlutil.h"
#include "io/libeventloop.h"
#include "io/message.h"
#include "messageconnection.h"
#include "net/messageconnectionutil.h"
#include "net/messageserver.h"
#include "networkaddress.h"
#include "replication/primarybackuplistener.h"

using std::string;
using std::vector;

class SubmitCallbackLog : public replication::FaultTolerantLogCallback {
public:
    // Does not own log.
    SubmitCallbackLog(replication::FaultTolerantLog* log) : log_(log) {
        log_->entry_callback(this);
    }

    virtual void submit(const io::Message& entry, const std::tr1::function<void()>& callback) {
        log_callbacks_.push_back(callback);
        log_->submit(entry, NULL);
    }

    virtual void nextLogEntry(int sequence, const string& entry, void* arg) {
        // HACK: We should register for specific sequence numbers?
        if (log_->isPrimary()) {
            // There should be a callback for each entry at the primary
            assert(!log_callbacks_.empty());
            // Call the callback
            log_callbacks_.front()();
            log_callbacks_.pop_front();
        } else {
            assert(log_callbacks_.empty());
        }

        if (sequence % 1000 == 0) {
            int32_t value;
            assert(entry.size() == sizeof(value));
            memcpy(&value, entry.data(), entry.size());
            printf("%d: %d\n", sequence, value);
        }
    }

private:
    replication::FaultTolerantLog* log_;
    CircularBuffer<std::tr1::function<void()> > log_callbacks_;
};

class NumberMessage : public io::Message {
public:
    NumberMessage(int32_t value) : value_(value) {}

    virtual void serialize(io::FIFOBuffer* buffer) const {
        buffer->copyIn(&value_, sizeof(value_));
    }

private:
    int32_t value_;
};

static void clientLoop(SubmitCallbackLog* log, io::EventLoop* event_loop, int counter) {
    counter += 1;

    NumberMessage counter_message(counter);
    if (counter == 100000) {
        event_loop->exit();
    } else {
        log->submit(counter_message, std::tr1::bind(clientLoop, log, event_loop, counter));
    }
}

int main(int argc, const char* argv[]) {
    if (argc < 2) {
        fprintf(stderr, "replicationdemo [listen port] [replica host_ports (for primary)]\n");
        return 1;
    }
    int port = atoi(argv[1]);

    // Connect to backups
    io::LibEventLoop event_loop;
    vector<MessageConnection*> backup_connections;
    for (int i = 2; i < argc; ++i) {
        NetworkAddress address;
        bool success = address.parse(argv[i]);
        if (!success) {
            fprintf(stderr, "Address %s could not be parsed\n", argv[i]);
            return 1;
        }
        TCPConnection* connection = new TCPConnection(connectTCP(address.sockaddr()));
        backup_connections.push_back(new TCPMessageConnection(event_loop.base(), connection));
    }

    // Start listening on the specified port
    net::MessageServer msg_server;
    MessageListener msg_listener(event_loop.base(), &msg_server);
    if (!msg_listener.listen(port)) {
        fprintf(stderr, "Error binding to port %d\n", port);
        return 1;
    }

    // Create the replica and attach it 
    replication::PrimaryBackupListener listener(&event_loop, &msg_server);
    SubmitCallbackLog log(listener.replica());

    // If we are the primary, connect to the replicas
    if (backup_connections.size() > 0) {
        listener.setPrimary(backup_connections);
        NumberMessage first(-1);
        log.submit(first, std::tr1::bind(clientLoop, &log, &event_loop, 0));
    } else {
        listener.setBackup();
    }

    printf("listening on port %d\n", port);

    // Exit when SIGINT is caught
    event_loop.exitOnSigInt(true);

    // Run the event loop
    event_loop.run();

    return 0;
}
