#!/usr/bin/python

"""A stupid message compiler for C++."""

import struct

def truncateLong(x):
    """Returns the unsigned integer part of a long x."""
    assert x > 0
    bytes = struct.pack("<Q", x)
    return struct.unpack("<I", bytes[:4])[0]


def reinterpretAsInt32(x):
    """Reinterprets an unsigned long as an int32."""
    assert x > 0
    bytes = struct.pack("<Q", x)
    return struct.unpack("<i", bytes[:4])[0]

class Type(object):
    INT32 = 0
    BOOL = 1
    STRING = 2
    STRUCT = 3
    LIST = 4

    def __init__(self, code):
        assert Type.INT32 <= code and code <= Type.LIST
        self.typecode = code

    def typeCode(self):
        return self.typecode

INT32_MAX = 2147483647
INT32_MIN = -2147483648

INT32 = Type(Type.INT32)
BOOL = Type(Type.BOOL)
STRING = Type(Type.STRING)
STRUCT = Type(Type.STRUCT)

class List(Type):
    def __init__(self, subtype):
        Type.__init__(self, Type.LIST)
        self.subtype = subtype

    def typeCode(self):
        return truncateLong(self.typecode * 33) ^ self.subtype.typeCode()

class MessageDefinition(object):
    def __init__(self, name, comment):
        self.typecode = Type.STRUCT
        self.name = name
        self.fields = []
        self.comment = comment

    def addField(self, type_structure, name, default, comment):
        # Validate the default value
        if type_structure.typecode == Type.INT32:
            assert isinstance(default, int)
            assert INT32_MIN <= default and default <= INT32_MAX
        elif type_structure.typecode == Type.BOOL:
            assert isinstance(default, bool)
        elif type_structure.typecode == Type.STRING:
            assert default is None or isinstance(default, str)
        elif type_structure.typecode == Type.STRUCT:
            assert default is None
        elif type_structure.typecode == Type.LIST:
            assert default is None or len(default) >= 0

        self.fields.append((type_structure, name, default, comment))

    def typeCode(self):
        """Generate a "unique" ID using a bad hash function."""
        # TODO: Do something smarter than this.

        def DJBHash(hashval, value):
            return truncateLong((hashval * 33) ^ value)

        def DJBStringHash(hashval, string):
            for c in string:
                hashval = DJBHash(hashval, ord(c))
            return hashval

        hashval = 5381
        hashval = DJBStringHash(hashval, self.name)

        for (type_structure, name, default, comment) in self.fields:
            hashval = DJBHash(hashval, type_structure.typeCode())
            hashval = DJBStringHash(hashval, name)

        return hashval


class CPPOutput(object):
    def __init__(self, filename, namespace):
        self.filename = filename
        self.namespace = namespace

        self.buffer = """/* AUTOMATICALLY GENERATED: DO NOT EDIT */
#ifndef %(guard)s
#define %(guard)s

#include <cassert>
#include <string>
#include <vector>

#include "base/assert.h"
#include "io/message.h"
#include "serialization.h"

namespace io {
class FIFOBuffer;
}

namespace %(namespace)s {\n\n""" % {'guard': self.includeGuard(), 'namespace': namespace}

        self.indent_ = 0

        self.defined_types = {}

    def indent(self):
        self.indent_ += 4

    def unindent(self):
        self.indent_ -= 4
        assert self.indent_ >= 0

    def out(self, text):
        return " " * self.indent_ + text + "\n"

    def includeGuard(self):
        return self.filename.upper().replace(".", "_").replace("/", "_")

    def typeString(self, type_structure):
        if type_structure.typecode == Type.INT32:
            return "int32_t"
        if type_structure.typecode == Type.BOOL:
            return "bool"
        if type_structure.typecode == Type.STRING:
            return "std::string"
        if type_structure.typecode == Type.STRUCT:
            return type_structure.name
        if type_structure.typecode == Type.LIST:
            return "std::vector<" + self.typeString(type_structure.subtype) + ">"
        raise ValueError, "Unknown type: " + str(type_structure)

    def typeValueString(self, type_structure, value):
        if type_structure.typecode == Type.INT32:
            return str(value)
        if type_structure.typecode == Type.BOOL:
            if value: return "true"
            return "false"
        if type_structure.typecode == Type.STRING:
            return "std::string"
        raise ValueError, "Unknown type: " + type_structure

    def serialize(self, type_structure, name):
        if type_structure.typecode == Type.LIST:
            o = "{\n"

            self.indent()
            o += self.out("int32_t _size_ = static_cast<int32_t>(%s.size());" % (name))
            o += self.out(
                    "assert(0 <= _size_ && static_cast<size_t>(_size_) == %s.size());" % (name))
            o += self.out(self.serialize(INT32, "_size_"));

            o += self.out("for (int _i_ = 0; _i_ < _size_; ++_i_) {")
            self.indent()
            o += self.out(self.serialize(type_structure.subtype, name + "[_i_]"));
            self.unindent()

            o += self.out("}")
            self.unindent()
            o += self.out("}")
            # Trim trailing \n
            return o[:-1]
        elif type_structure.typecode == Type.STRUCT:
            return "%s.appendToString(_out_);" % (name)
        else:
            return "serialization::serialize(%s, _out_);" % (name)

    def serializeBuffer(self, type_structure, name):
        if type_structure.typecode == Type.LIST:
            o = "{\n"

            self.indent()
            o += self.out("int32_t _size_ = static_cast<int32_t>(%s.size());" % (name))
            o += self.out(
                    "assert(0 <= _size_ && static_cast<size_t>(_size_) == %s.size());" % (name))
            o += self.out(self.serializeBuffer(INT32, "_size_"));

            o += self.out("for (int _i_ = 0; _i_ < _size_; ++_i_) {")
            self.indent()
            o += self.out(self.serializeBuffer(type_structure.subtype, name + "[_i_]"));
            self.unindent()

            o += self.out("}")
            self.unindent()
            o += self.out("}")
            # Trim trailing \n
            return o[:-1]
        elif type_structure.typecode == Type.STRUCT:
            return "%s.serialize(_out_);" % (name)
        else:
            return "serialization::serialize(%s, _out_);" % (name)

    def deserialize(self, type_structure, name):
        if type_structure.typecode == Type.LIST:
            o = "{\n"
            self.indent()
            o += self.out("int32_t _size_;")
            o += self.out("_start_ = serialization::deserialize(&_size_, _start_, _end_);")
            o += self.out("assert(_size_ >= 0);")
            o += self.out("%s.resize(_size_);" % (name))
            o += self.out("for (int _i_ = 0; _i_ < _size_; ++_i_) {")
            self.indent()
            o += self.out(self.deserialize(type_structure.subtype, name + "[_i_]"));
            self.unindent()
            o += self.out("}")
            self.unindent()
            o += self.out("}")
            # Trim trailing \n
            return o[:-1]
        elif type_structure.typecode == Type.STRUCT:
            return "_start_ = %s.parseFromString(_start_, _end_);" % (name)
        else:
            return "_start_ = serialization::deserialize(&%s, _start_, _end_);" % (name)

    def addMessage(self, message):
        self.buffer += self.out("// " + message.comment)
        self.buffer += self.out("class %s : public io::Message {" % (message.name))
        self.buffer += self.out("public:")
        self.indent()
        self.buffer += self.out("%s() :" % (message.name))

        # Initialize default types
        self.indent()
        for type_structure, name, default, comment in message.fields:
            # Skip field if there is no default value
            if default is None: continue

            self.buffer += self.out("%s(%s)," % (name, self.typeValueString(type_structure, default)))
        # Trim trailing ,
        if self.buffer.endswith(",\n"):
            self.buffer = self.buffer[:-2]
        else:
            assert self.buffer.endswith(" :\n")
            self.buffer = self.buffer[:-3]
        self.buffer += " {}\n"
        self.unindent()

        # Add field definitions
        for type_structure, name, default, comment in message.fields:
            self.buffer += "\n"
            if len(comment) > 0:
                self.buffer += self.out("// " + comment)
            self.buffer += self.out("%s %s;" % (self.typeString(type_structure), name))

        # Add operator== and !=. TODO: Move this to a .cc?
        self.buffer += "\n"
        self.buffer += self.out("bool operator==(const %s& other) const {" % message.name)
        self.indent()
        for type_structure, name, default, comment in message.fields:
            self.buffer += self.out("if (%s != other.%s) return false;" % (name, name))
        self.buffer += self.out("return true;")
        self.unindent()
        self.buffer += self.out("}")
        self.buffer += self.out("bool operator!=(const %s& other) const { return !(*this == other); }" % (message.name))

        # Add appendToString
        self.buffer += "\n"
        self.buffer += self.out("void appendToString(std::string* _out_) const {")
        self.indent()
        for type_structure, name, default, comment in message.fields:
            self.buffer += self.out(self.serialize(type_structure, name))
        self.unindent()
        self.buffer += self.out("}")

        # Add serialize
        self.buffer += "\n"
        self.buffer += self.out("virtual void serialize(io::FIFOBuffer* _out_) const {")
        self.indent()
        for type_structure, name, default, comment in message.fields:
            self.buffer += self.out(self.serializeBuffer(type_structure, name))
        self.unindent()
        self.buffer += self.out("}")

        # Add parseFromString
        self.buffer += "\n"
        self.buffer += self.out("const char* parseFromString(const char* _start_, const char* _end_) {")
        self.indent()
        for type_structure, name, default, comment in message.fields:
            self.buffer += self.out(self.deserialize(type_structure, name))
        self.buffer += self.out("return _start_;")
        self.unindent()
        self.buffer += self.out("}")
        # Add parseFromString(const std::string&) helper wrapper
        self.buffer += "\n"
        self.buffer += self.out("void parseFromString(const std::string& _str_) {")
        self.indent()
        self.buffer += self.out("const char* end = parseFromString(_str_.data(), _str_.data() + _str_.size());")
        self.buffer += self.out("ASSERT(end == _str_.data() + _str_.size());")
        self.unindent()
        self.buffer += self.out("}")

        # Add typeCode
        self.buffer += "\n"
        self.buffer += self.out("static int32_t typeCode() { return %d; }" % (
                reinterpretAsInt32(message.typeCode())))

        # All done
        self.unindent()
        self.buffer += self.out("};") + "\n"

    def output(self):
        self.buffer += "}  // namespace %s\n" % (self.namespace)
        self.buffer += "#endif  // %s\n" % (self.includeGuard())
        return self.buffer


def main(messages, namespace):
    import sys

    if len(sys.argv) != 2:
        sys.stderr.write("%s [output .h]\n" % (sys.argv[0]))
        sys.exit(1)

    output_path = sys.argv[1]
    assert output_path.endswith('.h')  
    out = CPPOutput(output_path, namespace)
    for message in messages:
        out.addMessage(message)

    f = open(output_path, "w")
    f.write(out.output())
    f.close()
