#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <cstdlib>
#include <tr1/functional>

#include "base/assert.h"
#include "base/function_thread.h"
#include "base/stringutil.h"
#include "libevent/event.h"
#include "messageconnection.h"
#include "strings/utils.h"
#include "stupidunit/stupidunit.h"
#include "tcplistener.h"

using std::string;

static const char FOO_STRING[] = "foo";

class TCPListenerTest : public Test, public TCPConnectionCallback, public TCPListenerCallback {
public:
    TCPListenerTest() :
            event_loop_(event_base_new()), listener_(new TCPListener(event_loop_, this)), connection_(NULL), other_(-1),
            write_available_(false) {}

    ~TCPListenerTest() {
        if (other_ >= 0) {
            // Force close the connection
            assert(connection_ != NULL);
            delete connection_;
            connection_ = NULL;
            int error = close(other_);
            ASSERT(error == 0);
        }
        // Need to delete this *before* we delete the event loop, since it
        // unregisters its events
        delete listener_;
        event_base_free(event_loop_);
    }

    virtual void connectionArrived(TCPConnection* connection) {
        assert(connection_ == NULL);
        connection_ = connection;
        connection_->setTarget(event_loop_, this);
    }

    virtual void readAvailable(TCPConnection* connection) {
        assert(data_.empty());

        // Read as much as possible into data_
        data_.resize(4096, '\0');
        assert(connection == connection_);
        int bytes = connection_->read(base::stringArray(&data_), data_.size());
        if (bytes == 0 || bytes == -1) {
            // Connection closed: clean it up
            data_.clear();
            delete connection_;
            connection_ = NULL;
        } else {
            // Typically we would need a loop
            assert(0 < bytes && bytes < 4096);
            data_.resize(bytes);
        }
    }

    virtual void writeAvailable(TCPConnection* connection) {
        assert(!write_available_);
        write_available_ = true;
    }

// disable -Wconversion for this function to work around a bug in htons in glibc
// TODO: Remove this eventually
#if defined(__OPTIMIZE__) && __GNUC__ == 4 && __GNUC_MINOR__ >= 3
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    void connectThread() {
        assert(other_ < 0);
        struct sockaddr_in addr;
        addr.sin_family = AF_INET;
        addr.sin_port = htons(assert_range_cast<uint16_t>(listener_->port()));
        addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
        other_ = connectTCP(addr);
        assert(other_ >= 0);
    }

    bool connect() {
        assert(other_ < 0);
        assert(connection_ == NULL);

        if (!listener_->listen(0)) {
            return false;
        }

        // Connect in another thread
        FunctionThread thread(std::tr1::bind(&TCPListenerTest::connectThread, this));
        int error = event_base_loop(event_loop_, EVLOOP_ONCE);
        thread.join();
        if (error != 0) {
            return false;
        }
        return connection_ != NULL;
    }

    bool closeOtherEnd() {
        // close the other end of the connection
        int error = close(other_);
        other_ = -1;
        if (error != 0) {
            return false;
        }
        return true;
    }

    bool closeConnection() {
        assert(other_ >= 0);
        assert(connection_ != NULL);

        // close the other end of the connection
        if (!closeOtherEnd()) {
            return false;
        }

        int error = event_base_loop(event_loop_, EVLOOP_ONCE);
        return error == 0;
    }

    bool writeFoo() {
        ssize_t bytes = write(other_, FOO_STRING, sizeof(FOO_STRING)-1);
        return bytes == sizeof(FOO_STRING)-1;
    }

    bool fillConnection() {
        char buffer[4096];
        memset(buffer, 0, sizeof(buffer));  // makes valgrind happy
        int bytes = 1;
        while ((bytes = connection_->write(buffer, sizeof(buffer))) > 0) {}

        return bytes == 0;
    }

    bool readFromOther() {
        char buffer[4096];
        ssize_t bytes = read(other_, buffer, sizeof(buffer));
        return bytes > 0;
    }

    bool readAllFromOther(string* out) {
        out->clear();

        TCPConnection connect(other_);
        char buffer[4096];
        int bytes = connect.read(buffer, sizeof(buffer));
        while (bytes > 0) {
            out->append(buffer, bytes);
            bytes = connect.read(buffer, sizeof(buffer));
        }

        other_ = -1;  // connection destructor will close the connection
        return bytes == -1;
    }

protected:
    event_base* event_loop_;
    TCPListener* listener_;
    TCPConnection* connection_;

    string data_;

    int other_;
    bool write_available_;
};

TEST_F(TCPListenerTest, BadCreate) {
    EXPECT_DEATH(TCPListener listener(event_loop_, NULL));
    EXPECT_DEATH(TCPListener listener(NULL, this));
}

TEST_F(TCPListenerTest, NoPort) {
    EXPECT_EQ(0, listener_->port());
}

TEST_F(TCPListenerTest, ConnectClose) {
    ASSERT_EQ(true, connect());
    ASSERT_EQ(true, closeConnection());
    EXPECT_EQ(NULL, connection_);
}

TEST_F(TCPListenerTest, ReadAvailable) {
    // Write into the socket
    ASSERT_EQ(true, connect());
    ASSERT_EQ(true, writeFoo());

    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE));
    EXPECT_EQ(FOO_STRING, data_);
    data_.clear();
}

TEST_F(TCPListenerTest, WriteAvailable) {
    ASSERT_EQ(true, connect());

    // write data into the socket until it is full
    ASSERT_EQ(true, fillConnection());

    // Read some data out the other end: the socket should become available for writing
    ASSERT_EQ(true, readFromOther());

    ASSERT_FALSE(write_available_);
    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE));
    EXPECT_EQ(true, write_available_);
}

TEST_F(TCPListenerTest, ClearTargetPendingEvents) {
    ASSERT_EQ(true, connect());

    // Write into the socket, fill the other end
    ASSERT_EQ(true, writeFoo());
    ASSERT_EQ(true, fillConnection());

    // Disable the callback target, run the event loop. Pending events should not cause problems.
    connection_->setTarget(NULL, NULL);
    ASSERT_EQ(true, readFromOther());
    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE|EVLOOP_NONBLOCK));
    EXPECT_EQ(true, data_.empty());
}

TEST_F(TCPListenerTest, SetNewTarget) {
    ASSERT_EQ(true, connect());

    class OtherTarget : public TCPConnectionCallback {
    public:
        OtherTarget() : read_available_(false) {}

        virtual void connectionArrived(TCPConnection* connection) {
            assert(false);
        }

        virtual void readAvailable(TCPConnection* connection) {
            assert(!read_available_);
            read_available_ = true;
        }

        virtual void writeAvailable(TCPConnection* connection) {
            assert(false);
        }
    
        bool read_available_;
    };

    OtherTarget other;
    connection_->setTarget(event_loop_, &other);

    // Write into the socket
    ASSERT_EQ(true, writeFoo());

    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE));
    EXPECT_EQ(true, other.read_available_);
    EXPECT_EQ(true, data_.empty());
}

TEST_F(TCPListenerTest, WriteOnClosedSocket) {
    ASSERT_EQ(true, listener_->listen(0));

    // Connect in another process which immediately calls exit.
    // This causes a future write to yield ECONNRESET.
    pid_t pid = fork();
    if (pid == 0) {
        connectThread();
        exit(42);
    }
    ASSERT_GT(pid, 0);

    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE));
    int status;
    ASSERT_EQ(pid, waitpid(pid, &status, 0));
    ASSERT_EQ(42, WEXITSTATUS(status));

    // Write into the socket. On Linux we need to write twice to get the error. On Mac OS X, twice
    // *normally* works but occasionally takes longer, particularly when other things are using
    // sockets.
    static const char DATA[] = "m";
    int count = 0;
    while (count < 100) {
        count += 1;
        int result = connection_->write(DATA, sizeof(DATA));
        if (result == -1) {
            break;
        }
        ASSERT_EQ(sizeof(DATA), result);
    }
    EXPECT_LE(2, count);
    EXPECT_LT(count, 100);

    delete connection_;
}

// TODO: Refactor the TCP test infrastructure into some common component. These tests should be in
// messageconnection_test.cc.
TEST_F(TCPListenerTest, TCPMessageConnectionFillConnection) {
    ASSERT_EQ(true, connect());

    // write data into the socket until it is full
    ASSERT_EQ(true, fillConnection());

    TCPMessageConnection m(event_loop_, connection_);

    // write a message
    static const char DATA[] = "message";
    EXPECT_TRUE(m.rawSend(42, DATA, sizeof(DATA) - 1));

    // Read some data out the other end: the socket should become available for writing
    ASSERT_EQ(true, readFromOther());

    // Triggers the TCPMessageConnection callback to write data
    EXPECT_FALSE(write_available_);
    ASSERT_EQ(0, event_base_loop(event_loop_, EVLOOP_ONCE));
    EXPECT_FALSE(write_available_);

    string out;
    readAllFromOther(&out);
    EXPECT_TRUE(strings::endsWith(out, DATA));
}

TEST_F(TCPListenerTest, TCPMessageConnectionWriteHuge) {
    ASSERT_TRUE(connect());

    // write a huge message: it should work
    char data[io::FIFOBuffer::PAGE_DATA_SIZE + 100];
    memset(data, 0x42, sizeof(data));
    TCPMessageConnection m(event_loop_, connection_);
    EXPECT_TRUE(m.rawSend(42, data, sizeof(data)));

    // Read multiple chunks of data from the other end
    string out;
    readAllFromOther(&out);
    ASSERT_EQ(sizeof(data) + 2 * sizeof(int32_t), out.size());
    EXPECT_EQ(0, memcmp(out.data() + 2 * sizeof(int32_t), data, sizeof(data)));
}

TEST_F(TCPListenerTest, TCPMessageConnectionWriteError) {
    ASSERT_EQ(true, connect());

    closeOtherEnd();

    TCPMessageConnection m(event_loop_, connection_);

    // write a message
    static const char DATA[] = "message";
    // See comment in WriteOnClosedSocket.
    bool triggered = false;
    for (int i = 0; !triggered && i < 100; ++i) {
        if (!m.rawSend(42, DATA, sizeof(DATA))) {
            triggered = true;
        }
    }
    EXPECT_TRUE(triggered);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
