#include "tpcccoordinator.h"

#include <tr1/unordered_set>

#include "bench/tpccdb.h"
#include "serialization.h"
#include "tpccengine.h"

using dtxn::CoordinatedTransaction;
using serialization::deserialize;
using std::string;
using std::vector;

struct TransactionState {
    int32_t type;
    int home_partition;
};

void TPCCCoordinator::begin(CoordinatedTransaction* transaction) {
    // Deserialize the type
    int32_t type;
    const string& message = transaction->received()[0].second;
    deserialize(&type, message.data(), message.data() + message.size());
    
    if (type == TPCCEngine::NEW_ORDER) {
        beginNewOrder(transaction);
    } else if (type == TPCCEngine::PAYMENT_BY_ID || type == TPCCEngine::PAYMENT_BY_NAME) {
        beginPayment(transaction);
    } else if (type == TPCCEngine::RESET) {
        beginReset(transaction);
    } else {
        fprintf(stderr, "invalid type code: %d\n", type);
        CHECK(false);
    }

    // All transactions are one shot
    transaction->setAllDone();
}

void TPCCCoordinator::nextRound(CoordinatedTransaction* transaction) {
    assert(transaction->received().size() > 1);
    TransactionState* state = (TransactionState*) transaction->state();
    if (state->type == TPCCEngine::NEW_ORDER) {
        finishNewOrder(transaction);
    } else if (state->type == TPCCEngine::PAYMENT_HOME) {
        finishPayment(transaction);
    } else if (state->type == TPCCEngine::RESET) {
        finishReset(transaction);
    } else {
        fprintf(stderr, "invalid type code: %d\n", state->type);
        CHECK(false);
    }
}

void TPCCCoordinator::done(CoordinatedTransaction* transaction) {
    delete (TransactionState*) transaction->state();
}

void TPCCCoordinator::beginNewOrder(CoordinatedTransaction* transaction) {
    // Deserialize the request
    int32_t warehouse_id;
    int32_t district_id;
    int32_t customer_id;
    vector<NewOrderItem> items;
    string now;
    TPCCEngine::deserializeNewOrder(transaction->received()[0].second, TPCCEngine::NEW_ORDER,
            &warehouse_id, &district_id, &customer_id, &items, &now);

    // Forward the request to the home warehouse
    int home_partition = warehouseToPartition(warehouse_id);
    // HACK: Abuse this to avoid having to reserialize the request
    assert(transaction->received()[0].second[0] == TPCCEngine::NEW_ORDER);
    const_cast<string&>(transaction->received()[0].second)[0] = TPCCEngine::NEW_ORDER_HOME;
    transaction->send(home_partition, transaction->received()[0].second);

    // Serialize the remote request
    string remote_request;
    TPCCEngine::serializeNewOrderRemote(warehouse_id, items, &remote_request);

    // Figure out which remote partitions we need to talk to
    std::tr1::unordered_set<int> remote_partitions;
    TPCCDB::WarehouseSet warehouses = TPCCDB::newOrderRemoteWarehouses(warehouse_id, items);
    assert(warehouses.size() > 0);
    for (TPCCDB::WarehouseSet::const_iterator i = warehouses.begin(); i != warehouses.end(); ++i) {
        // send a request to each partition once
        int partition = warehouseToPartition(*i);
        if (partition != home_partition &&
                remote_partitions.find(partition) == remote_partitions.end()) {
            remote_partitions.insert(partition);
            transaction->send(partition, remote_request);
        }
    }
    assert(remote_partitions.size() > 0);

    // Remember this is a new order transaction
    TransactionState* state = new TransactionState();
    state->type = TPCCEngine::NEW_ORDER;
    state->home_partition = home_partition;
    transaction->state(state);
}
    
void TPCCCoordinator::finishNewOrder(CoordinatedTransaction* transaction) {
    TransactionState* state = (TransactionState*) transaction->state();
    assert(state->type == TPCCEngine::NEW_ORDER);

    // Find the main response
    NewOrderOutput output;
    for (size_t i = 0; i < transaction->received().size(); ++i) {
        if (transaction->received()[i].first == state->home_partition) {
            TPCCEngine::deserializeNewOrderOutput(transaction->received()[i].second, &output);
            break;
        }
    }

    // Combine all other responses
    vector<int32_t> quantities;
    for (size_t i = 0; i < transaction->received().size(); ++i) {
        if (transaction->received()[i].first != state->home_partition) {
            TPCCEngine::deserializeQuantities(transaction->received()[i].second, &quantities);
            TPCCDB::newOrderCombine(quantities, &output);
        }
    }

    // Serialize everything and send it out
    string out;
    TPCCEngine::serializeNewOrderOutput(output, &out);
    transaction->commit(out);
}

void TPCCCoordinator::beginPayment(CoordinatedTransaction* transaction) {
    // Deserialize part of the request: only the bits we need!
    const string& request = transaction->received()[0].second;
    const char* end = request.data() + request.size();
    int32_t type;
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    const char* next = deserialize(&type, request.data(), end);
    assert(type == TPCCEngine::PAYMENT_BY_ID || type == TPCCEngine::PAYMENT_BY_NAME);
    next = deserialize(&warehouse_id, next, end);
    next = deserialize(&district_id, next, end);
    next = deserialize(&c_warehouse_id, next, end);
    assert(warehouse_id != c_warehouse_id);

    // Forward the request to the home warehouse
    int home_partition = warehouseToPartition(warehouse_id);
    
    bool by_id = request[0] == TPCCEngine::PAYMENT_BY_ID;
    assert(by_id || request[0] == TPCCEngine::PAYMENT_BY_NAME);
    // HACK: Abuse this to avoid having to reserialize the request
    const_cast<string&>(request)[0] = TPCCEngine::PAYMENT_HOME;
    transaction->send(home_partition, request);

    // Forward the remote request. string now is usused, but reusing the message avoids copying.
    if (by_id) {
        const_cast<string&>(request)[0] = TPCCEngine::PAYMENT_REMOTE_BY_ID;
    } else {
        const_cast<string&>(request)[0] = TPCCEngine::PAYMENT_REMOTE_BY_NAME;
    }
    int remote_partition = warehouseToPartition(c_warehouse_id);
    assert(remote_partition != home_partition);
    transaction->send(remote_partition, request);

    // Remember this is a payment by id transaction
    TransactionState* state = new TransactionState();
    state->type = TPCCEngine::PAYMENT_HOME;  // we don't care which payment this is
    state->home_partition = home_partition;
    transaction->state(state);
}

void TPCCCoordinator::finishPayment(CoordinatedTransaction* transaction) {
    TransactionState* state = (TransactionState*) transaction->state();
    assert(state->type == TPCCEngine::PAYMENT_HOME);
    assert(transaction->received().size() == 2);

    // Find the main response
    PaymentOutput output;
    bool found = false;
    for (size_t i = 0; i < transaction->received().size(); ++i) {
        if (transaction->received()[i].first == state->home_partition) {
            TPCCEngine::deserializePaymentOutput(transaction->received()[i].second, &output);
            found = true;
            break;
        }
    }
    assert(found);

    // Combine the other response
    PaymentOutput other;
    found = false;
    for (size_t i = 0; i < transaction->received().size(); ++i) {
        if (transaction->received()[i].first != state->home_partition) {
            TPCCEngine::deserializePaymentOutput(transaction->received()[i].second, &other);
            TPCCDB::paymentCombine(other, &output);
            found = true;
        }
    }
    assert(found);

    // Serialize everything and send it out
    string out;
    TPCCEngine::serializePaymentOutput(output, &out);
    transaction->commit(out);
}

void TPCCCoordinator::beginReset(CoordinatedTransaction* transaction) {
    // Forward the request to all partitions
    for (int i = 0; i < num_warehouses_/warehouses_per_partition_; ++i) {
        transaction->send(i, transaction->received()[0].second);
    }
    
    // Remember this is a new order transaction
    TransactionState* state = new TransactionState();
    state->type = TPCCEngine::RESET;
    state->home_partition = -1;
    transaction->state(state);
}

void TPCCCoordinator::finishReset(CoordinatedTransaction* transaction) {
    transaction->commit("");
}
