#include "tpccengine.h"

#include <cstring>

#include "base/cast.h"
#include "bench/tpccdb.h"
#include "bench/tpcctables.h"
#include "serialization.h"

#include "dtxn/transaction.h"

using serialization::deserialize;
using serialization::serialize;
using std::string;
using std::vector;

class MemcpySerializer {
public:
    template <typename T>
    void operator()(const T& value, string* out) const {
        out->append(reinterpret_cast<const char*>(&value), sizeof(value));
    }
};

static const char* memcpyDeserialize(void* output, size_t size, const char* start, const char* end) {
    const char* next = start + size;
    assert(next <= end);
    memcpy(output, start, size);
    return next;    
}

class MemcpyDeserializer {
public:
    template <typename T>
    const char* operator()(T* out, const char* start, const char* end) const {
        return memcpyDeserialize(out, sizeof(*out), start, end);
    }
};

template <typename T, typename Serializer>
void vectorSerialize(const vector<T>& values, Serializer s, string* out) {
    int32_t size = assert_range_cast<int32_t>(values.size());
    serialize(size, out);
    for (size_t i = 0; i < values.size(); ++i) {
        s(values[i], out);
    }
}

template <typename T, typename Deserializer>
const char* vectorDeserialize(vector<T>* out, Deserializer d, const char* start, const char* end) {
    assert(start < end);
    int32_t size;
    start = deserialize(&size, start, end);
    assert(size >= 0);
    out->resize(size);
    for (int i = 0; i < size; ++i) {
        assert(start < end);
        start = d(&(*out)[i], start, end);
    }
    return start;
}

static void cStringSerialize(const char* c_string, string* out) {
    serialize(c_string, strlen(c_string), out);
}

class ItemInfoSerializer {
public:
    void operator()(const NewOrderOutput::ItemInfo& value, string* out) const {
        serialize(value.s_quantity, out);
        serialize(value.i_price, out);
        serialize(value.ol_amount, out);
        serialize(value.brand_generic, out);
        cStringSerialize(value.i_name, out);
    }
};

static const char* strcpyDeserialize(char* destination, const char* start, const char* end) {
    string out;
    const char* next = deserialize(&out, start, end);
    strcpy(destination, out.c_str());
    return next;
}

class ItemInfoDeserializer {
public:
    const char* operator()(NewOrderOutput::ItemInfo* value, const char* start, const char* end) const {
        const char* next = deserialize(&value->s_quantity, start, end);
        next = deserialize(&value->i_price, next, end);
        next = deserialize(&value->ol_amount, next, end);
        next = deserialize(&value->brand_generic, next, end);
        return strcpyDeserialize(value->i_name, next, end);
    }
};

static void serializeTwoInts(int32_t code, int32_t one, int32_t two, string* out) {
    serialize(code, out);
    serialize(one, out);
    serialize(two, out);
}

static void serializeThreeInts(int32_t code, int32_t one, int32_t two, int32_t three, string* out) {
    serializeTwoInts(code, one, two, out);
    serialize(three, out);
}

static const char* deserializeTwoInts(const string& in, int32_t expected_code, int32_t* one,
        int32_t* two) {
    int32_t temp;
    const char* real_end = in.data() + in.size();
    const char* next = deserialize(&temp, in.data(), real_end);
    assert(temp == expected_code);

    next = deserialize(one, next, real_end);
    return deserialize(two, next, real_end);
}

static const char* deserializeThreeInts(const string& in, int32_t expected_code, int32_t* one,
        int32_t* two, int32_t* three) {
    const char* real_end = in.data() + in.size();
    const char* next = deserializeTwoInts(in, expected_code, one, two);    
    return deserialize(three, next, real_end);
}

void TPCCEngine::serializeNewOrder(int32_t code, int32_t warehouse_id, int32_t district_id, int32_t customer_id,
        const std::vector<NewOrderItem>& items, const char* now, std::string* out) {
    assert(code == TPCCEngine::NEW_ORDER || code == TPCCEngine::NEW_ORDER_HOME);
    serializeThreeInts(code, warehouse_id, district_id, customer_id, out);
    vectorSerialize(items, MemcpySerializer(), out);
    cStringSerialize(now, out);
}

void TPCCEngine::deserializeNewOrder(const string& input, int32_t code, int32_t* warehouse_id,
        int32_t* district_id, int32_t* customer_id, vector<NewOrderItem>* items,
        string* now) {
    assert(code == NEW_ORDER || code == NEW_ORDER_HOME);
    const char* const end = input.data() + input.size();
    const char* next = deserializeThreeInts(input, code, warehouse_id,
            district_id, customer_id);
    next = vectorDeserialize(items, MemcpyDeserializer(), next, end);
    next = deserialize(now, next, end);
    assert(next == end);
}

void TPCCEngine::serializeNewOrderRemote(int32_t warehouse_id,
        const std::vector<NewOrderItem>& items, std::string* out) {
    serialize(TPCCEngine::NEW_ORDER_REMOTE, out);
    serialize(warehouse_id, out);
    vectorSerialize(items, MemcpySerializer(), out);
}

static dtxn::ExecutionEngine::Status callNewOrder(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t customer_id;
    vector<NewOrderItem> items;
    string now;
    TPCCEngine::deserializeNewOrder(input, TPCCEngine::NEW_ORDER, &warehouse_id, &district_id,
            &customer_id, &items, &now);

    NewOrderOutput output;
    bool commit = db->newOrder(warehouse_id, district_id, customer_id, items, now.c_str(), &output,
            reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializeNewOrderOutput(output, out);

    if (commit) {
        return dtxn::ExecutionEngine::OK;
    } else {
        return dtxn::ExecutionEngine::ABORT_USER;
    }
}

static dtxn::ExecutionEngine::Status callNewOrderHome(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t customer_id;
    vector<NewOrderItem> items;
    string now;
    TPCCEngine::deserializeNewOrder(input, TPCCEngine::NEW_ORDER_HOME, &warehouse_id, &district_id,
            &customer_id, &items, &now);

    NewOrderOutput output;
    bool commit = db->newOrderHome(warehouse_id, district_id, customer_id, items, now.c_str(),
            &output, reinterpret_cast<TPCCUndo**>(undo));
    if (!commit) {
        TPCCEngine::serializeNewOrderOutput(output, out);
        return dtxn::ExecutionEngine::ABORT_USER;
    }

    // Call newOrderRemote for any local warehouses
    vector<int32_t> quantities;
    TPCCDB::WarehouseSet remote_warehouses = TPCCDB::newOrderRemoteWarehouses(warehouse_id, items);
    for (TPCCDB::WarehouseSet::const_iterator i = remote_warehouses.begin(); i != remote_warehouses.end(); ++i) {
        if (db->hasWarehouse(*i)) {
            bool commit = db->newOrderRemote(warehouse_id, *i, items, &quantities, NULL);
            ASSERT(commit);
            TPCCDB::newOrderCombine(quantities, &output);
        }
    }
    TPCCEngine::serializeNewOrderOutput(output, out);

    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callNewOrderRemote(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    const char* const end = input.data() + input.size();
    int32_t code;
    int32_t warehouse_id;
    vector<NewOrderItem> items;
    const char* next = deserialize(&code, input.data(), end);
    assert(code == TPCCEngine::NEW_ORDER_REMOTE);
    next = deserialize(&warehouse_id, next, end);
    next = vectorDeserialize(&items, MemcpyDeserializer(), next, end);
    assert(next == end);

    // The (int32_t) cast avoids the template from taking a reference, which requires that we put
    // INVALID_QUANTITY in a .cc file.
    vector<int32_t> final_quantities(items.size(), (int32_t) TPCCDB::INVALID_QUANTITY);
    vector<int32_t> temp_quantities(items.size(), (int32_t) TPCCDB::INVALID_QUANTITY);
    TPCCDB::WarehouseSet remote_warehouses = TPCCDB::newOrderRemoteWarehouses(warehouse_id, items);
    for (TPCCDB::WarehouseSet::const_iterator i = remote_warehouses.begin(); i != remote_warehouses.end(); ++i) {
        if (db->hasWarehouse(*i)) {
            bool commit = db->newOrderRemote(warehouse_id, *i, items, &temp_quantities,
                    reinterpret_cast<TPCCUndo**>(undo));
            if (!commit) {
                return dtxn::ExecutionEngine::ABORT_USER;
            }
            
            db->newOrderRemote(warehouse_id, *i, items, &temp_quantities, NULL);
            TPCCDB::newOrderCombine(temp_quantities, &final_quantities);
        }
    }
    vectorSerialize(final_quantities, MemcpySerializer(), out);

    //~ printf("new order remote %d\n", out->size());
    
    return dtxn::ExecutionEngine::OK;
}

void serializeStockLevel(int32_t warehouse_id, int32_t district_id, int32_t threshold,
        string* out) {
    serializeThreeInts(TPCCEngine::STOCK_LEVEL, warehouse_id, district_id, threshold, out);
}

static dtxn::ExecutionEngine::Status callStockLevel(
        const std::string& input, TPCCDB* db, std::string* out) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t threshold;
    const char* const end = input.data() + input.size();
    const char* next = deserializeThreeInts(input, TPCCEngine::STOCK_LEVEL, &warehouse_id,
            &district_id, &threshold);
    ASSERT(next == end);

    int32_t count = db->stockLevel(warehouse_id, district_id, threshold);
    serialize(count, out);
    return dtxn::ExecutionEngine::OK;
}

void serializeOrderStatus(int32_t warehouse_id, int32_t district_id, int32_t customer_id,
        std::string* out) {
    serializeThreeInts(TPCCEngine::ORDER_STATUS_BY_ID, warehouse_id, district_id, customer_id,
            out);
}

void serializeOrderStatus(int32_t warehouse_id, int32_t district_id, const char* c_last,
        std::string* out) {
    serializeTwoInts(TPCCEngine::ORDER_STATUS_BY_NAME, warehouse_id, district_id, out);
    cStringSerialize(c_last, out);
}

class OrderLineSubsetSerializer {
public:
    void operator()(const OrderStatusOutput::OrderLineSubset& value, string* out) {
        serialize(value.ol_i_id, out);
        serialize(value.ol_supply_w_id, out);
        serialize(value.ol_quantity, out);
        serialize(value.ol_amount, out);
        cStringSerialize(value.ol_delivery_d, out);
    }
};

class OrderLineSubsetDeserializer {
public:
    const char* operator()(OrderStatusOutput::OrderLineSubset* value, const char* start,
            const char* end) {
        const char* next = deserialize(&value->ol_i_id, start, end);
        next = deserialize(&value->ol_supply_w_id, next, end);
        next = deserialize(&value->ol_quantity, next, end);
        next = deserialize(&value->ol_amount, next, end);
        return strcpyDeserialize(value->ol_delivery_d, next, end);
    }
};

static void serializeOrderStatusOutput(const OrderStatusOutput& output, string* out) {
    serialize(output.c_id, out);
    serialize(output.c_balance, out);
    serialize(output.o_id, out);
    serialize(output.o_carrier_id, out);
    vectorSerialize(output.lines, OrderLineSubsetSerializer(), out);
    cStringSerialize(output.c_first, out);
    cStringSerialize(output.c_middle, out);
    cStringSerialize(output.c_last, out);
    cStringSerialize(output.o_entry_d, out);
}

static dtxn::ExecutionEngine::Status callOrderStatusById(
        const std::string& input, TPCCDB* db, std::string* out) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t customer_id;
    const char* const end = input.data() + input.size();
    const char* next = deserializeThreeInts(input, TPCCEngine::ORDER_STATUS_BY_ID, &warehouse_id,
            &district_id, &customer_id);
    ASSERT(next == end);

    OrderStatusOutput output;
    db->orderStatus(warehouse_id, district_id, customer_id, &output);
    serializeOrderStatusOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callOrderStatusByName(
        const std::string& input, TPCCDB* db, std::string* out) {
    int32_t warehouse_id;
    int32_t district_id;
    string c_last;
    const char* const end = input.data() + input.size();
    const char* next = deserializeTwoInts(input, TPCCEngine::ORDER_STATUS_BY_NAME, &warehouse_id,
            &district_id);
    next = deserialize(&c_last, next, end);
    assert(next == end);

    OrderStatusOutput output;
    db->orderStatus(warehouse_id, district_id, c_last.c_str(), &output);
    serializeOrderStatusOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

void deserializeOrderStatusOutput(const std::string& in, OrderStatusOutput* output) {
    const char* next = in.data();
    const char* end = in.data() + in.size();
    next = deserialize(&output->c_id, next, end);
    next = deserialize(&output->c_balance, next, end);
    next = deserialize(&output->o_id, next, end);
    next = deserialize(&output->o_carrier_id, next, end);
    next = vectorDeserialize(&output->lines, OrderLineSubsetDeserializer(), next, end);
    // TODO: This makes copies, which is gross, but avoids touching uninitialized data
    next = strcpyDeserialize(output->c_first, next, end);
    next = strcpyDeserialize(output->c_middle, next, end);
    next = strcpyDeserialize(output->c_last, next, end);
    next = strcpyDeserialize(output->o_entry_d, next, end);
    assert(next == end);
}

void serializePayment(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
        int32_t c_district_id, int32_t customer_id, const char* c_last, float h_amount,
        const char* now, std::string* out) {
    assert(customer_id == 0 || c_last == NULL);
    static const char EMPTY[] = "";
    if (c_last == NULL) {
        c_last = EMPTY;
    } else {
        assert(strlen(c_last) > 0);
    }
    assert(customer_id == 0 || strlen(c_last) == 0);
    int32_t type = TPCCEngine::PAYMENT_BY_ID;
    if (customer_id == 0) type = TPCCEngine::PAYMENT_BY_NAME;
    serializeThreeInts(type, warehouse_id, district_id, c_warehouse_id,
            out);
    serialize(c_district_id, out);
    serialize(customer_id, out);
    cStringSerialize(c_last, out);
    serialize(h_amount, out);
    cStringSerialize(now, out);
}

void TPCCEngine::serializePaymentOutput(const PaymentOutput& output, string* out) {
#define SERIALIZE_ADDRESS(src, prefix, output) \
    cStringSerialize(src.prefix ## street_1, output); \
    cStringSerialize(src.prefix ## street_2, output); \
    cStringSerialize(src.prefix ## city, output); \
    cStringSerialize(src.prefix ## state, output); \
    cStringSerialize(src.prefix ## zip, output)

    SERIALIZE_ADDRESS(output, w_, out);
    SERIALIZE_ADDRESS(output, d_, out);

    serialize(output.c_credit_lim, out);
    serialize(output.c_discount, out);
    serialize(output.c_balance, out);
    cStringSerialize(output.c_first, out);
    cStringSerialize(output.c_middle, out);
    cStringSerialize(output.c_last, out);
    SERIALIZE_ADDRESS(output, c_, out);
    cStringSerialize(output.c_phone, out);
    cStringSerialize(output.c_since, out);
    cStringSerialize(output.c_credit, out);
    cStringSerialize(output.c_data, out);

#undef SERIALIZE_ADDRESS
}

static void deserializePayment(const std::string& input, int32_t code,
        int32_t* warehouse_id, int32_t* district_id, int32_t* c_warehouse_id,
        int32_t* c_district_id, int32_t* customer_id, string* c_last, float* h_amount, string* now) {
    const char* const end = input.data() + input.size();
    const char* next = deserializeThreeInts(input, code, warehouse_id, district_id, c_warehouse_id);
    next = deserialize(c_district_id, next, end);
    next = deserialize(customer_id, next, end);
    next = deserialize(c_last, next, end);
    next = deserialize(h_amount, next, end);
    next = deserialize(now, next, end);
    assert(next == end);
    assert(*customer_id == 0 || c_last->empty());
}

static dtxn::ExecutionEngine::Status callPaymentById(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    int32_t c_district_id;
    int32_t customer_id;
    string c_last;
    float h_amount;
    string now;
    deserializePayment(input, TPCCEngine::PAYMENT_BY_ID, &warehouse_id,
            &district_id, &c_warehouse_id, &c_district_id, &customer_id, &c_last, &h_amount, &now);
    assert(customer_id != 0 && c_last.empty());

    PaymentOutput output;
    db->payment(warehouse_id, district_id, c_warehouse_id, c_district_id, customer_id, h_amount,
            now.c_str(), &output, reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializePaymentOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callPaymentHome(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    int32_t c_district_id;
    int32_t customer_id;
    string c_last;
    float h_amount;
    string now;
    deserializePayment(input, TPCCEngine::PAYMENT_HOME, &warehouse_id,
            &district_id, &c_warehouse_id, &c_district_id, &customer_id, &c_last, &h_amount, &now);

    PaymentOutput output;
    db->paymentHome(warehouse_id, district_id, c_warehouse_id, c_district_id, customer_id, h_amount,
            now.c_str(), &output, reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializePaymentOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callPaymentRemoteById(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    int32_t c_district_id;
    int32_t customer_id;
    string c_last;
    float h_amount;
    string now;
    deserializePayment(input, TPCCEngine::PAYMENT_REMOTE_BY_ID, &warehouse_id,
            &district_id, &c_warehouse_id, &c_district_id, &customer_id, &c_last, &h_amount, &now);
    assert(customer_id != 0 && c_last.empty());

    PaymentOutput output;
    db->paymentRemote(warehouse_id, district_id, c_warehouse_id, c_district_id, customer_id,
            h_amount, &output, reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializePaymentOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callPaymentRemoteByName(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    int32_t c_district_id;
    int32_t customer_id;
    string c_last;
    float h_amount;
    string now;
    deserializePayment(input, TPCCEngine::PAYMENT_REMOTE_BY_NAME, &warehouse_id,
            &district_id, &c_warehouse_id, &c_district_id, &customer_id, &c_last, &h_amount, &now);
    assert(customer_id == 0 && !c_last.empty());

    PaymentOutput output;
    db->paymentRemote(warehouse_id, district_id, c_warehouse_id, c_district_id, c_last.c_str(),
            h_amount, &output, reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializePaymentOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

static dtxn::ExecutionEngine::Status callPaymentByName(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t district_id;
    int32_t c_warehouse_id;
    int32_t c_district_id;
    int32_t customer_id;
    string c_last;
    float h_amount;
    string now;
    deserializePayment(input, TPCCEngine::PAYMENT_BY_NAME, &warehouse_id,
            &district_id, &c_warehouse_id, &c_district_id, &customer_id, &c_last, &h_amount, &now);
    assert(customer_id == 0 && !c_last.empty());

    PaymentOutput output;
    db->payment(warehouse_id, district_id, c_warehouse_id, c_district_id, c_last.c_str(), h_amount,
            now.c_str(), &output, reinterpret_cast<TPCCUndo**>(undo));
    TPCCEngine::serializePaymentOutput(output, out);
    return dtxn::ExecutionEngine::OK;
}

void TPCCEngine::deserializePaymentOutput(const std::string& in, PaymentOutput* output) {
#define DESERIALIZE_ADDRESS(dest, prefix) \
    next = strcpyDeserialize(dest->prefix ## street_1, next, end); \
    next = strcpyDeserialize(dest->prefix ## street_2, next, end); \
    next = strcpyDeserialize(dest->prefix ## city, next, end); \
    next = strcpyDeserialize(dest->prefix ## state, next, end); \
    next = strcpyDeserialize(dest->prefix ## zip, next, end)

    const char* end = in.data() + in.size();
    const char* next = in.data();
    DESERIALIZE_ADDRESS(output, w_);
    DESERIALIZE_ADDRESS(output, d_);

    next = deserialize(&output->c_credit_lim, next, end);
    next = deserialize(&output->c_discount, next, end);
    next = deserialize(&output->c_balance, next, end);
    next = strcpyDeserialize(output->c_first, next, end);
    next = strcpyDeserialize(output->c_middle, next, end);
    next = strcpyDeserialize(output->c_last, next, end);
    DESERIALIZE_ADDRESS(output, c_);
    next = strcpyDeserialize(output->c_phone, next, end);
    next = strcpyDeserialize(output->c_since, next, end);
    next = strcpyDeserialize(output->c_credit, next, end);
    next = strcpyDeserialize(output->c_data, next, end);

#undef DESERIALIZE_ADDRESS
}

void serializeDelivery(int32_t warehouse_id, int32_t carrier_id, const char* now, string* out) {
    serializeTwoInts(TPCCEngine::DELIVERY, warehouse_id, carrier_id, out);
    cStringSerialize(now, out);
}

void deserializeDeliveryOutput(const std::string& in, std::vector<DeliveryOrderInfo>* out) {
    const char* end = in.data() + in.size();
    const char* next = vectorDeserialize(out, MemcpyDeserializer(), in.data(), end);
    ASSERT(next == end);
}

static dtxn::ExecutionEngine::Status callDelivery(
        const std::string& input, TPCCDB* db, std::string* out, void** undo) {
    int32_t warehouse_id;
    int32_t carrier_id;
    string now;
    const char* const end = input.data() + input.size();
    const char* next = deserializeTwoInts(input, TPCCEngine::DELIVERY, &warehouse_id,
            &carrier_id);
    next = deserialize(&now, next, end);
    assert(next == end);

    vector<DeliveryOrderInfo> output;
    db->delivery(warehouse_id, carrier_id, now.c_str(), &output,
            reinterpret_cast<TPCCUndo**>(undo));
    vectorSerialize(output, MemcpySerializer(), out);
    return dtxn::ExecutionEngine::OK;
}

void TPCCEngine::serializeNewOrderOutput(const NewOrderOutput& in, std::string* out) {
    serialize(in.w_tax, out);
    serialize(in.d_tax, out);
    serialize(in.o_id, out);
    serialize(in.c_discount, out);
    serialize(in.total, out);
    vectorSerialize(in.items, ItemInfoSerializer(), out);
    cStringSerialize(in.c_last, out);
    cStringSerialize(in.c_credit, out);
    cStringSerialize(in.status, out);
}

void TPCCEngine::deserializeNewOrderOutput(const std::string& in, NewOrderOutput* output) {
    const char* next = in.data();
    const char* end = in.data() + in.size();
    next = deserialize(&output->w_tax, next, end);
    next = deserialize(&output->d_tax, next, end);
    next = deserialize(&output->o_id, next, end);
    next = deserialize(&output->c_discount, next, end);
    next = deserialize(&output->total, next, end);
    next = vectorDeserialize(&output->items, ItemInfoDeserializer(), next, end);
    // TODO: This makes copies, which is gross, but avoids touching uninitialized data
    next = strcpyDeserialize(output->c_last, next, end);
    next = strcpyDeserialize(output->c_credit, next, end);
    next = strcpyDeserialize(output->status, next, end);
    assert(next == end);
}

void TPCCEngine::deserializeQuantities(const std::string& in, std::vector<int32_t>* quantities) {
    const char* end = in.data() + in.size();
    const char* next = vectorDeserialize(quantities, MemcpyDeserializer(), in.data(), end);
    ASSERT(next == end);
}

void TPCCEngine::serializeReset(string* out) {
    serialize(RESET, out);
}

TPCCEngine::~TPCCEngine() {
    delete db_;
}

dtxn::ExecutionEngine::Status TPCCEngine::tryExecute(const string& work_unit, string* output,
        void** undo, dtxn::Transaction* transaction) {
    db_->setTransaction(transaction);
    assert(output->empty());

    // Deserialize the type
    int32_t type;
    deserialize(&type, work_unit.data(), work_unit.data() + work_unit.size());

    assert(transaction == NULL || (output == transaction->output() && undo == transaction->undo()));
    try {
        switch (type) {
            case TPCCEngine::NEW_ORDER:
                return callNewOrder(work_unit, db_, output, undo);
                break;
            case TPCCEngine::STOCK_LEVEL:
                return callStockLevel(work_unit, db_, output);
                break;
            case TPCCEngine::ORDER_STATUS_BY_ID:
                return callOrderStatusById(work_unit, db_, output);
                break;
            case TPCCEngine::ORDER_STATUS_BY_NAME:
                return callOrderStatusByName(work_unit, db_, output);
                break;
            case TPCCEngine::PAYMENT_BY_ID:
                return callPaymentById(work_unit, db_, output, undo);
                break;
            case TPCCEngine::PAYMENT_BY_NAME:
                return callPaymentByName(work_unit, db_, output, undo);
                break;
            case TPCCEngine::DELIVERY:
                return callDelivery(work_unit, db_, output, undo);
                break;
            case TPCCEngine::NEW_ORDER_HOME:
                return callNewOrderHome(work_unit, db_, output, undo);
                break;
            case TPCCEngine::NEW_ORDER_REMOTE:
                return callNewOrderRemote(work_unit, db_, output, undo);
                break;
            case TPCCEngine::PAYMENT_HOME:
                return callPaymentHome(work_unit, db_, output, undo);
                break;
            case TPCCEngine::PAYMENT_REMOTE_BY_ID:
                return callPaymentRemoteById(work_unit, db_, output, undo);
                break;
            case TPCCEngine::PAYMENT_REMOTE_BY_NAME:
                // TODO: Test
                return callPaymentRemoteByName(work_unit, db_, output, undo);
                break;
            case TPCCEngine::RESET:
                // reset our database, if we can
                if (reset_function_ == NULL) {
                    return ExecutionEngine::ABORT_USER;
                } else {
                    // Lock the entire partition to prevent other transactions from accessing bad
                    // state.
                    if (transaction != NULL) {
                        bool result = transaction->tryWriteLock(&partition_lock_);
                        ASSERT(result);
                    }
                    db_ = reset_function_(db_);
                    return ExecutionEngine::OK;
                }
                break;
            default:
                CHECK(false);
        }
    } catch (LockBlockedException& e) {
        // HACK: Apply the undo buffer since these transactions aren't idempotent.
        // (eg. d_next_o_id += 1 cannot be repeated without changing the result).
        if (*undo != NULL) {
            applyUndo(*undo);
            *undo = NULL;
        }
        assert(output->empty());
        //~ printf("%p blocked\n", transaction);
        return ExecutionEngine::BLOCKED;
    }

    CHECK(false);
}

void TPCCEngine::applyUndo(void* undo) {
    db_->applyUndo(reinterpret_cast<TPCCUndo*>(undo));
}

void TPCCEngine::freeUndo(void* undo) {
    db_->freeUndo(reinterpret_cast<TPCCUndo*>(undo));
}
