#ifndef TPCCENGINE_H__
#define TPCCENGINE_H__

#include <vector>

#include "base/assert.h"
#include "dtxn/executionengine.h"
#include "dtxn/locks/lock.h"

class DeliveryOrderInfo;
class NewOrderItem;
class NewOrderOutput;
class OrderStatusOutput;
class PaymentOutput;
class TPCCDB;

class TPCCEngine : public dtxn::ExecutionEngine {
public:
    typedef TPCCDB* (*ResetFPtr)(TPCCDB* old);

    // Owns db. reset_function can be NULL provided the reset functionality is never invoked.
    TPCCEngine(TPCCDB* db, ResetFPtr reset_function) : db_(db), reset_function_(reset_function) {
        assert(db_ != NULL);
    }
    virtual ~TPCCEngine();

    virtual Status tryExecute(const std::string& work_unit, std::string* output, void** undo,
            dtxn::Transaction* transaction);
    virtual bool supports_locks() const { return true; }
    virtual void applyUndo(void* undo);
    virtual void freeUndo(void* undo);

    static const int32_t NEW_ORDER = 0;
    static const int32_t STOCK_LEVEL = 1;
    static const int32_t ORDER_STATUS_BY_ID = 2;
    static const int32_t ORDER_STATUS_BY_NAME = 3;
    static const int32_t PAYMENT_BY_ID = 4;
    static const int32_t PAYMENT_BY_NAME = 5;
    static const int32_t DELIVERY = 6;

    static const int32_t NEW_ORDER_HOME = 7;
    static const int32_t NEW_ORDER_REMOTE = 8;
    static const int32_t PAYMENT_HOME = 9;
    static const int32_t PAYMENT_REMOTE_BY_ID = 10;
    static const int32_t PAYMENT_REMOTE_BY_NAME = 11;

    static const int32_t RESET = 127;

    static void serializeNewOrder(int32_t code, int32_t warehouse_id, int32_t district_id,
            int32_t customer_id, const std::vector<NewOrderItem>& items, const char* now, std::string* out);
    static void deserializeNewOrder(const std::string& in,
            int32_t code, int32_t* warehouse_id, int32_t* district_id, int32_t* customer_id,
            std::vector<NewOrderItem>* items, std::string* now);
    static void serializeNewOrderRemote(int32_t warehouse_id,
            const std::vector<NewOrderItem>& items, std::string* out);
    static void deserializeNewOrderOutput(const std::string& in, NewOrderOutput* output);
    static void deserializeQuantities(const std::string& in, std::vector<int32_t>* quantities);
    static void serializeNewOrderOutput(const NewOrderOutput& in, std::string* out);

    static void serializePaymentOutput(const PaymentOutput& in, std::string* out);
    static void deserializePaymentOutput(const std::string& in, PaymentOutput* output);

    static void serializeReset(std::string* out);

private:
    TPCCDB* db_;
    ResetFPtr reset_function_;

    // Used to reserve access to entire database for things like resetting it.
    // TODO: Make every transaction acquire and release this lock to actually ensure this?
    dtxn::Lock partition_lock_;
};

void serializeStockLevel(int32_t warehouse_id, int32_t district_id, int32_t threshold,
        std::string* out);

void serializeOrderStatus(int32_t warehouse_id, int32_t district_id, int32_t customer_id,
        std::string* out);
void serializeOrderStatus(int32_t warehouse_id, int32_t district_id, const char* c_last,
        std::string* out);
void deserializeOrderStatusOutput(const std::string& in, OrderStatusOutput* output);

// One version: either customer_id is zero or c_last should be NULL.
void serializePayment(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
        int32_t c_district_id, int32_t customer_id, const char* c_last, float h_amount,
        const char* now, std::string* out);

void serializeDelivery(int32_t warehouse_id, int32_t carrier_id, const char* now,
        std::string* out);
void deserializeDeliveryOutput(const std::string& in, std::vector<DeliveryOrderInfo>* out);

#endif
