#include "bench/mocktpccdb.h"
#include "dtxn/locks/locktransaction.h"
#include "dtxn/transaction.h"
#include "serialization.h"
#include "stupidunit/stupidunit.h"
#include "tpccengine.h"

using std::string;
using std::vector;

static const char NOW[DATETIME_SIZE+1] = "20090310083852";

class TPCCEngineTest : public Test {
public:
    TPCCEngineTest() : db_(new tpcc::MockTPCCDB()), engine_(db_, NULL), undo_(NULL), items_(1) {
        items_[0].i_id = 1;
        items_[0].ol_supply_w_id = 2;
        items_[0].ol_quantity = 3; 
    }

    tpcc::MockTPCCDB* db_;
    TPCCEngine engine_;
    string input_;
    string output_;
    void* undo_;

    vector<NewOrderItem> items_;
};

TEST_F(TPCCEngineTest, NoLocks) {
    EXPECT_TRUE(engine_.supports_locks());
}

TEST_F(TPCCEngineTest, NewOrderSerialization) {
    NewOrderOutput output;
    output.items.resize(1);
    output.items[0].s_quantity = 1;
    output.items[0].i_price = 2.0;
    output.items[0].ol_amount = 3.0;
    output.items[0].brand_generic = 'a';
    strcpy(output.items[0].i_name, "foo");

    TPCCEngine::serializeNewOrderOutput(output, &input_);

    NewOrderOutput out;
    TPCCEngine::deserializeNewOrderOutput(input_, &out);
    EXPECT_EQ(1, out.items.size());
    EXPECT_EQ(string("foo"), out.items[0].i_name);
}

TEST_F(TPCCEngineTest, NewOrderSinglePartition) {
    TPCCEngine::serializeNewOrder(TPCCEngine::NEW_ORDER, 0, 1, 2, items_, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(0, db_->w_id_);
    EXPECT_EQ(1, db_->d_id_);
    EXPECT_EQ(2, db_->c_id_);
    ASSERT_EQ(1, db_->items_.size());
    EXPECT_EQ(1, db_->items_[0].i_id);
    EXPECT_EQ(2, db_->items_[0].ol_supply_w_id);
    EXPECT_EQ(3, db_->items_[0].ol_quantity);
    EXPECT_EQ(NOW, db_->now_);

    NewOrderOutput out;
    TPCCEngine::deserializeNewOrderOutput(output_, &out);
    EXPECT_EQ(42, out.d_tax);

    db_->new_order_committed_ = false;
    output_.clear();
    EXPECT_EQ(dtxn::ExecutionEngine::ABORT_USER, engine_.tryExecute(input_, &output_, NULL, NULL));

    out.d_tax = 0;
    TPCCEngine::deserializeNewOrderOutput(output_, &out);
    EXPECT_EQ(42, out.d_tax);
}

TEST_F(TPCCEngineTest, NewOrderApplyUndo) {
    TPCCEngine::serializeNewOrder(TPCCEngine::NEW_ORDER, 0, 1, 2, items_, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_TRUE(undo_ != NULL);
    EXPECT_EQ(1, db_->undo_count_);
    engine_.applyUndo(undo_);
    EXPECT_EQ(0, db_->undo_count_);
}

TEST_F(TPCCEngineTest, NewOrderFreeUndo) {
    TPCCEngine::serializeNewOrder(TPCCEngine::NEW_ORDER, 0, 1, 2, items_, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_TRUE(undo_ != NULL);
    EXPECT_EQ(1, db_->undo_count_);
    engine_.freeUndo(undo_);
    EXPECT_EQ(0, db_->undo_count_);

    NewOrderOutput out;
    TPCCEngine::deserializeNewOrderOutput(output_, &out);
    EXPECT_EQ(42, out.d_tax);
}

TEST_F(TPCCEngineTest, NewOrderRemoteAbort) {
    db_->new_order_committed_ = false;
    TPCCEngine::serializeNewOrderRemote(1, items_, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::ABORT_USER,
            engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_EQ(0, output_.size());
    EXPECT_EQ(NULL, undo_);
    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->remote_warehouse_);
    EXPECT_EQ(items_.size(), db_->items_.size());
}

TEST_F(TPCCEngineTest, NewOrderHomeAbort) {
    // This will actually abort
    db_->new_order_committed_ = false;
    TPCCEngine::serializeNewOrder(TPCCEngine::NEW_ORDER_HOME, 0, 1, 2, items_, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::ABORT_USER,
            engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_EQ(NULL, undo_);
    EXPECT_EQ(0, db_->w_id_);
    EXPECT_EQ(items_.size(), db_->items_.size());
}

TEST_F(TPCCEngineTest, StockLevel) {
    serializeStockLevel(1, 2, 3, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->d_id_);
    EXPECT_EQ(3, db_->stock_level_threshold_);

    int32_t out;
    serialization::deserialize(&out, output_.data(), output_.data() + output_.size());
    EXPECT_EQ(42, out);
}

TEST_F(TPCCEngineTest, OrderStatusById) {
    serializeOrderStatus(1, 2, 3, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->d_id_);
    EXPECT_EQ(3, db_->order_status_c_id_);

    OrderStatusOutput out;
    deserializeOrderStatusOutput(output_, &out);
    EXPECT_EQ(1.23f, out.c_balance);
}

TEST_F(TPCCEngineTest, OrderStatusByName) {
    serializeOrderStatus(1, 2, "foobar", &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->d_id_);
    EXPECT_EQ("foobar", db_->c_last_);

    OrderStatusOutput out;
    deserializeOrderStatusOutput(output_, &out);
    EXPECT_EQ(1.23f, out.c_balance);
}

TEST_F(TPCCEngineTest, PaymentById) {
    serializePayment(1, 2, 3, 4, 5, NULL, 6.01f, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->d_id_);
    EXPECT_EQ(3, db_->c_w_id_);
    EXPECT_EQ(4, db_->c_d_id_);
    EXPECT_EQ(5, db_->c_id_);
    EXPECT_EQ(6.01f, db_->h_amount_);
    EXPECT_EQ(NOW, db_->now_);

    PaymentOutput out;
    TPCCEngine::deserializePaymentOutput(output_, &out);
    EXPECT_EQ(0, strlen(out.w_street_1));
    EXPECT_EQ(1.23f, out.c_balance);
}

TEST_F(TPCCEngineTest, PaymentHomeAbort) {
    serializePayment(1, 2, 3, 4, 5, NULL, 6.01f, NOW, &input_);
    input_[0] = TPCCEngine::PAYMENT_HOME;
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_TRUE(undo_ != NULL);
    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(1, db_->undo_count_);
    engine_.applyUndo(undo_);
    EXPECT_EQ(0, db_->undo_count_);

    PaymentOutput out;
    TPCCEngine::deserializePaymentOutput(output_, &out);
}

TEST_F(TPCCEngineTest, PaymentRemoteAbort) {
    serializePayment(1, 2, 3, 4, 5, NULL, 6.01f, NOW, &input_);
    input_[0] = TPCCEngine::PAYMENT_REMOTE_BY_ID;
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, &undo_, NULL));
    EXPECT_TRUE(undo_ != NULL);
    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(1, db_->undo_count_);
    engine_.applyUndo(undo_);
    EXPECT_EQ(0, db_->undo_count_);

    PaymentOutput out;
    TPCCEngine::deserializePaymentOutput(output_, &out);
}

TEST_F(TPCCEngineTest, Delivery) {
    serializeDelivery(1, 2, NOW, &input_);
    EXPECT_EQ(dtxn::ExecutionEngine::OK, engine_.tryExecute(input_, &output_, NULL, NULL));

    EXPECT_EQ(1, db_->w_id_);
    EXPECT_EQ(2, db_->delivery_carrier_id_);
    EXPECT_EQ(NOW, db_->now_);

    vector<DeliveryOrderInfo> out;
    deserializeDeliveryOutput(output_, &out);
    ASSERT_EQ(1, out.size());
    EXPECT_EQ(42, out[0].o_id);
}

TEST_F(TPCCEngineTest, TransactionsSupported) {
    serializeDelivery(1, 2, NOW, &input_);

    dtxn::LockTransaction txn(NULL, input_, true, false);
    EXPECT_EQ(dtxn::ExecutionEngine::OK,
            engine_.tryExecute(input_, txn.output(), txn.undo(), &txn));

    // executing again will block. NOTE: Due to an idempotence hack, this will undo the
    // transaction.
    dtxn::LockTransaction txn2(NULL, input_, true, false);
    EXPECT_EQ(dtxn::ExecutionEngine::BLOCKED,
            engine_.tryExecute(input_, txn2.output(), txn2.undo(), &txn2));
    EXPECT_EQ(NULL, *txn2.undo());

    // abort transaction 2
    vector<dtxn::LockTransaction*> granted_transactions;
    txn2.dropLocks(&granted_transactions);

    // commit transaction 1
    txn.dropLocks(&granted_transactions);
    engine_.freeUndo(*txn.undo());
}

static TPCCDB* reset_old = NULL;
static TPCCDB* resetter(TPCCDB* old) {
    delete old;
    reset_old = old;
    return new tpcc::MockTPCCDB();
}

TEST_F(TPCCEngineTest, Reset) {
    // No reset function: this doesn't work
    TPCCEngine::serializeReset(&input_);
    EXPECT_EQ(dtxn::ExecutionEngine::ABORT_USER,
            engine_.tryExecute(input_, &output_, NULL, NULL));

    db_ = new tpcc::MockTPCCDB();
    TPCCEngine resettable_engine(db_, resetter);
    EXPECT_EQ(dtxn::ExecutionEngine::OK,
            resettable_engine.tryExecute(input_, &output_, NULL, NULL));
    EXPECT_EQ(db_, reset_old);

    dtxn::MockLockTransaction txn(NULL, input_, true, true);
    EXPECT_EQ(dtxn::ExecutionEngine::OK,
            resettable_engine.tryExecute(input_, txn.output(), txn.undo(), &txn));
    // Make sure reset acquires some lock: needed to avoid an assert, and for
    // correctness.
    EXPECT_NE(NULL, txn.last_lock_);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
