#include <cstring>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include "bench/clock.h"
#include "bench/randomgenerator.h"
#include "bench/tpccgenerator.h"
#include "bench/tpcctables.h"
#include "dtxn/common.h"
#include "dtxn/configparser.h"
#include "dtxn/locks/lockscheduler.h"
#include "dtxn/occ/occscheduler.h"
#include "dtxn/ordered/orderedscheduler.h"
#include "tpccengine.h"

using std::string;
using std::vector;

// TODO: integrate this with the tpcc benchmark?
struct GenerateParameters {
    tpcc::NURandC cLoad;
    tpcc::NURandC cRun;

    int total_warehouses;
    int num_partitions;
    int partition_id;

    int load_seed;

    // TODO: Cache?
    int start_warehouse() {
        return partition_id * warehouses_per_partition() + 1;
    }

    // TODO: Cache?
    int warehouses_per_partition() {
        return total_warehouses /  num_partitions;
    }

    TPCCTables* generate() {
        // Generate cLoad
        SystemClock clock;
        tpcc::RealRandomGenerator* rng = new tpcc::RealRandomGenerator();
        rng->seed(load_seed);
        tpcc::NURandC cLoad = tpcc::NURandC::makeRandom(rng);
        rng->setC(cLoad);

        // Generate the data
        printf("Loading items ...\n");
        char now[Clock::DATETIME_SIZE+1];
        clock.getDateTimestamp(now);
        TPCCGenerator generator(rng, now, Item::NUM_ITEMS, District::NUM_PER_WAREHOUSE,
                Customer::NUM_PER_DISTRICT, NewOrder::INITIAL_NUM_PER_DISTRICT);
        int64_t begin = clock.getMicroseconds();
        TPCCTables* tables = new TPCCTables();
        generator.makeItemsTable(tables);

        // TODO: Vertical partitioning only needs s_dist_xx and s_data. This copies *all* stock.
        // This means that some partitions have incorrect data (eg. s_quantity), but it is simple.
        printf("Loading stock ...\n");
        for (int i = 0; i < total_warehouses; ++i) {
            generator.makeStock(tables, i + 1);
        }

        for (int i = 0; i < warehouses_per_partition(); ++i) {
            printf("Loading warehouse %d ...\n", start_warehouse() + i);
            generator.makeWarehouseWithoutStock(tables, start_warehouse() + i);
        }
        int64_t end = clock.getMicroseconds();
        printf("Loading time %" PRId64 " ms\n", (end-begin)/1000);

        return tables;
    }
};
// TODO: I don't like statics. Fix?
static GenerateParameters params;

static TPCCDB* reset_function(TPCCDB* old) {
    delete old;
    return params.generate();
}

// TODO: Eliminate duplication with htableserver
int main(int argc, const char* argv[]) {
    if (argc != 6) {
        fprintf(stderr, "tpccserver [mode = block|spec|lock|occ] [total warehouses] [configuration] [partition index] [replica index]\n");
        return 1;
    }
    const char* mode_string = argv[1];
    params.total_warehouses = atoi(argv[2]);
    assert(params.total_warehouses >= 1);
    const char* config_file = argv[3];
    params.partition_id = atoi(argv[4]);
    int replica_index = atoi(argv[5]);

    enum Mode {
        BLOCKING,
        SPECULATE,
        LOCKING,
        OCC,
    };
    Mode mode = LOCKING;
    if (strcmp(mode_string, "block") == 0) {
        mode = BLOCKING;
    } else if (strcmp(mode_string, "spec") == 0) {
        mode = SPECULATE;
    } else if (strcmp(mode_string, "lock") == 0) {
        mode = LOCKING;
    } else if (strcmp(mode_string, "occ") == 0) {
        mode = OCC;
    } else {
        fprintf(stderr, "mode must be block|spec|lock|occ\n");
        return 1;
    }

    vector<dtxn::Partition> partitions = dtxn::parseConfigurationFromPath(config_file);
    params.num_partitions = (int) partitions.size();
    assert(0 < params.num_partitions && params.num_partitions <= params.total_warehouses);
    assert(params.total_warehouses % params.num_partitions == 0);
    assert(0 <= params.partition_id && params.partition_id < params.num_partitions);

    // Load the data. TODO: Refactor this to be shared
    // IMPORTANT: everyone must use the same random seed in order to generate
    // the same data. TODO: The "now" field may still differ.
    // TODO: Ideally we would generate the data and "ship" it to the partitions.
    params.load_seed = 0;
    TPCCEngine engine(params.generate(), reset_function);

    dtxn::TransactionScheduler* scheduler = NULL;
    if (mode == LOCKING) {
        scheduler = new dtxn::LockScheduler(&engine);
    } else if (mode == OCC) {
        scheduler = new dtxn::OCCScheduler(&engine);
    } else {
        assert(mode == BLOCKING || mode == SPECULATE);
        scheduler = new dtxn::OrderedScheduler(&engine);
        if (mode == SPECULATE) {
            ((dtxn::OrderedScheduler*) scheduler)->cc_mode(dtxn::OrderedScheduler::CC_SPECULATIVE);
        } else {
            assert(mode == BLOCKING);
        }
    }

    dtxn::runServer(scheduler, partitions[params.partition_id], replica_index);

    return 0;
}
