// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#ifndef TPCC_RANDOMGENERATOR_H__
#define TPCC_RANDOMGENERATOR_H__

#include "../randomgenerator.h"

namespace tpcc {

class RandomGenerator;

// Constant C values for the NURand function.
struct NURandC {
    NURandC() : c_last_(0), c_id_(0), ol_i_id_(0) {}

    int c_last_;
    int c_id_;
    int ol_i_id_;

    // Sets the fields randomly.
    static NURandC makeRandom(RandomGenerator* generator);

    // Sets the fields randomly, in a fashion acceptable for a test run. c_load is the value of
    // c_last_ that was used to generate the tables. See TPC-C 2.1.6.1. (page 20).
    static NURandC makeRandomForRun(RandomGenerator* generator, const NURandC& c_load);
};

class RandomGenerator {
public:
    RandomGenerator() : c_values_(NURandC()) {}
    virtual ~RandomGenerator() {}

    // Return a random integer in the range [lower, upper]. The range is inclusive.
    virtual int number(int lower, int upper) = 0;

    // Return a random integer in the range [lower, upper] excluding excluded. The range is
    // inclusive.
    int numberExcluding(int lower, int upper, int excluding);

    void astring(char* s, int lower_length, int upper_length);
    void nstring(char* s, int lower_length, int upper_length);

    // Fill name with a random last name, generated according to TPC-C rules. Limits the customer
    // id for the generated name to cid.
    void lastName(char* name, int max_cid);

    float fixedPoint(int digits, float lower, float upper);

    // Non-uniform random number function from TPC-C 2.1.6. (page 20).
    int NURand(int A, int x, int y);

    int* makePermutation(int lower, int upper);

    void setC(const NURandC& c) {
        c_values_ = c;
    }

private:
    NURandC c_values_;
};

// A mock RandomGenerator for unit testing.
class MockRandomGenerator : public RandomGenerator {
public:
    MockRandomGenerator() : minimum_(true) {}

    virtual int number(int lower, int upper) {
        if (minimum_) return lower;
        else return upper;
    }

    bool minimum_;
};

static const int MAX_LAST_NAME = 16;

// Generate a last name as defined by TPC-C 4.3.2.3. name must be at least MAX_LAST_NAME+1 bytes.
void makeLastName(int num, char* name);

// A real RandomGenerator that uses random_r.
class RealRandomGenerator : public RandomGenerator {
public:
    // Seeds the generator with the current time.
    RealRandomGenerator() {}

    virtual int number(int lower, int upper);

    // Seed the generator with value.
    void seed(unsigned int value) {
        generator_.seed(value);
    }

private:
    ::RandomGenerator generator_;
};

}  // namespace tpcc

#endif
