// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include <limits>

#include "base/assert.h"
#include "benchmark/benchmark.h"
#include "stupidunit/stupidunit.h"

using std::string;
using std::vector;

class TestWorker : public benchmark::BenchmarkWorker {
public:
    TestWorker() : tid_(0), initialize_count_(0), work_count_(0), no_measure_(false),
            finished_(false) {}

    virtual void initialize() {
        CHECK(tid_ == 0);
        tid_ = pthread_self();
        initialize_count_ += 1;
    }

    virtual void work(bool measure) {
        ASSERT(tid_ == pthread_self());
        if (!measure) no_measure_ = true;
        work_count_ += 1;
        // allows other threads a chance to run instead of running a busy loop
        sched_yield();
    }

    virtual void finish() {
        finished_ = true;
    }

    virtual int result() {
        return work_count_;
    }

    pthread_t tid_;
    int initialize_count_;
    int work_count_;
    bool no_measure_;
    bool finished_;
};

class BenchmarkTest : public Test {
public:
    BenchmarkTest() {
        workers_.push_back(&one_);
        workers_.push_back(&two_);
    }

    TestWorker one_;
    TestWorker two_;
    vector<benchmark::BenchmarkWorker*> workers_;
};

TEST_F(BenchmarkTest, ThreadBench) {
    int total_work = std::numeric_limits<int>::min();
    int64_t microseconds = benchmark::runThreadBench(1, 1, workers_, &total_work);
    EXPECT_LE(1000000, microseconds);
    EXPECT_LT(0, total_work);

    EXPECT_NE(0, one_.tid_);
    EXPECT_LT(0, one_.work_count_);
    EXPECT_TRUE(one_.no_measure_);
    EXPECT_EQ(1, one_.initialize_count_);

    EXPECT_NE(0, two_.tid_);
    EXPECT_LT(0, two_.work_count_);
    EXPECT_TRUE(two_.no_measure_);
    EXPECT_EQ(1, one_.initialize_count_);

    EXPECT_NE(one_.tid_, two_.tid_);

    EXPECT_TRUE(one_.finished_);
    EXPECT_TRUE(two_.finished_);
}

TEST_F(BenchmarkTest, ThreadBenchLatency) {
    vector<int> latencies;
    double requests_per_second = benchmark::runThreadLatencyBench(1, 1, workers_, 2, &latencies);
    EXPECT_GT(2.1, requests_per_second);

    EXPECT_LE(2, latencies.size());
    EXPECT_GT(latencies[0], 0);
    EXPECT_GT(latencies[1], 0);

    EXPECT_NE(0, one_.tid_);
    EXPECT_LT(0, one_.work_count_);
    EXPECT_TRUE(one_.no_measure_);
    EXPECT_EQ(1, one_.initialize_count_);

    EXPECT_NE(0, two_.tid_);
    EXPECT_LT(0, two_.work_count_);
    EXPECT_TRUE(two_.no_measure_);
    EXPECT_EQ(1, one_.initialize_count_);

    EXPECT_NE(one_.tid_, two_.tid_);

    EXPECT_TRUE(one_.finished_);
    EXPECT_TRUE(two_.finished_);
}

TEST_F(BenchmarkTest, ForkBench) {
    double requests_per_second = benchmark::runForkBench(1, 1, workers_);
    EXPECT_LT(0, requests_per_second);

    // The stuff ran in other processes, not in this process
    EXPECT_EQ(0, one_.tid_);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
