// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "benchmark/benchmark.h"

#include <cstdio>
#include <cstring>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/time.h>
#include <time.h>
#include <sys/wait.h>

#include "base/array.h"
#include "base/assert.h"
#include "base/time.h"

using std::vector;

namespace {

using benchmark::BenchmarkWorker;

// man semctl says that this union must be defined this way
union semun {
   int val;
   struct semid_ds* buf;
   unsigned short* array;
   struct seminfo* __buf;
};

/** Contains the state for executing performance tests with multiple processes. */
class ForkBench {
public:
    ForkBench(int num_processes) : num_processes_(num_processes), shared_(NULL) {
        assert(num_processes_ > 0);

        // Create an anonymous shared memory region
        // man mmap on Linux says MAP_ANON is deprecated for MAP_ANONYMOUS but MAP_ANON exists on
        // both Linux and Mac OS X.
        shared_ = (SharedState*) mmap(NULL, sizeof(*shared_), PROT_READ | PROT_WRITE,
                MAP_SHARED | MAP_ANON, -1, 0);
        assert(shared_ != (void*) -1 && shared_ != NULL);

        // Create a semaphore for synchronizing at start
        sem_id_ = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
        assert(sem_id_ != -1);

        // Set the semaphore value to 1
        union semun value;
        value.val = 1;
        int error = semctl(sem_id_, 0, SETVAL, value);
        ASSERT(error == 0);
    }

    ~ForkBench() {
        // unmap the shared memory
        int error = munmap((void*) shared_, sizeof(*shared_));
        ASSERT(error == 0);

        // avoid valgrind warnings
        union semun value;
        memset(&value, 0, sizeof(value));

        // get rid of the semaphore
        error = semctl(sem_id_, 0, IPC_RMID, value);
        ASSERT(error == 0);
    }

    enum State {
        INVALID,
        WARMUP,
        MEASURE,
        DONE,
        EXIT,
    };

    State state() const {
        // See stateChange about volatile
        return shared_->state;
    }

    /** Wait for num_threads to call this. Returns once all the threads have entered. */
    void blockForStart() {
        struct sembuf sem_ops[1];
        memset(sem_ops, 0, sizeof(sem_ops));
        sem_ops[0].sem_num = 0;

        // Do an atomic increment in shared memory to see if we are all ready
        int count = __sync_add_and_fetch(&shared_->ready_count, 1);
        if (count < num_processes_) {
            // block by waiting for zero on the semaphore
            sem_ops[0].sem_op = 0;
            //~ printf("blocking %d ...\n", count);
            int error = semop(sem_id_, sem_ops, base::arraySize(sem_ops));
            ASSERT(error == 0);
            //~ printf("barrier released ...\n");
        } else {
            assert(count == num_processes_);
            assert(state() == INVALID);
            stateChange(WARMUP);

            // We are the last to start! Decrement the semaphore, which releases all processes
            sem_ops[0].sem_op = -1;
            //~ printf("releasing ...\n");
            int error = semop(sem_id_, sem_ops, base::arraySize(sem_ops));
            ASSERT(error == 0);
            //~ printf("released?\n");
        }
        assert(state() == WARMUP);
    }

    void startMeasure() {
        assert(state() == WARMUP);
        stateChange(MEASURE);
    }

    void doneMeasure() {
        assert(state() == MEASURE);
        stateChange(DONE);
        // The master thread must also signal that it is done
        signalDone();
    }

    /** Notify that this worker has entered the done state. */
    void signalDone() {
        assert(state() == DONE);
        int count = __sync_sub_and_fetch(&shared_->ready_count, 1);
        assert(0 <= count && count < num_processes_);
        if (count == 0) {
            // We are the last thread to observe the test finishing
            stateChange(EXIT);
        }
    }

    void setWorkerResult(int worker_id, int value) {
        assert(0 <= worker_id && worker_id < num_processes_ - 1);
        assert(shared_->worker_out[worker_id] == 0);
        shared_->worker_out[worker_id] = value;
    }

    int sumWorkerResults() const {
        int sum = 0;
        for (int i = 0; i < num_processes_ - 1; ++i) {
            sum += shared_->worker_out[i];
        }
        return sum;
    }

private:
    int sharedMemorySize() const {
        return (int)(sizeof(*shared_) + sizeof(shared_->worker_out[0]) * (num_processes_ - 1));
    }

    const int num_processes_;

    struct SharedState {
        volatile State state;
        volatile int ready_count;
        int worker_out[0];
    };

    // Points into shared memory
    SharedState* shared_;

    void stateChange(State next_state) {
        // volatile because we want all reads/writes to actually go to memory
        // TODO: On x86 this should be fine, but it may need memory fences on
        // other platforms to unsure atomic uncached reads and writes.
        //~ printf("changing state %d\n", next_state);
        shared_->state = next_state;
    }

    // The following members are protected by the mutex.
    int sem_id_;
};

void forkChild(int worker_id, ForkBench* test_state, BenchmarkWorker* work) {
    // initialize in parallel
    work->initialize();

    // wait for everyone to start running
    test_state->blockForStart();

    bool seen_done = false;

    ForkBench::State state = test_state->state();
    while (state != ForkBench::EXIT) {
        if (state == ForkBench::DONE && !seen_done) {
            // This is the first time we have observed that the test is done
            // notify the global test state, then continue applying load
            seen_done = true;
            test_state->signalDone();
        }

        // apply load
        work->work(state == ForkBench::MEASURE);
        state = test_state->state();
    }

    // Clean up in the thread
    work->finish();

    // write our result to shared memory
    test_state->setWorkerResult(worker_id, work->result());
}

}  // namespace

namespace benchmark {

double runForkBench(
        int warmup_secs,
        int measure_secs,
        const vector<BenchmarkWorker*>& workers) {
    // + 1 to include this "master" thread
    ForkBench test(static_cast<int>(workers.size() + 1));

    for (size_t i = 0; i < workers.size(); ++i) {
        pid_t pid = fork();
        if (pid == 0) {
            // child: run the child worker then exit
            forkChild((int) i, &test, workers[i]);
            exit(0);
        } else {
            assert(pid > 0);
        }
    }

    // Wait for all the threads to start
    test.blockForStart();

    // Running: wait for the warm up period
    sleep(warmup_secs);

    // Run the measurement period
    //~ printf("measure begin\n");
    struct timeval start;
    int error = gettimeofday(&start, NULL);
    assert(error == 0);

    test.startMeasure();
    sleep(measure_secs);
    test.doneMeasure();
    //~ printf("measure end\n");

    struct timeval end;
    error = gettimeofday(&end, NULL);
    assert(error == 0);

    // Wait for the workers to finish
    int count = 0;
    while (count < workers.size()) {
        int status = 0;
        pid_t pid = wait(&status);
        assert(status == 0);
        ASSERT(pid > 0);
        count += 1;
    }

    int total = test.sumWorkerResults();
    int64_t us = base::timevalDiffMicroseconds(end, start);
    double seconds = (double) us / 1000000.0;
    return total / seconds;
}

}  // namespace benchmark
