// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#ifndef HTABLEBENCHMARK_H__
#define HTABLEBENCHMARK_H__

#include <cassert>
#include <string>
#include <vector>

#include <stdint.h>

class RandomGenerator;

namespace dtxn {
class Fragment;
class Partition;
}

namespace htable {
class MiniTransaction;
}

namespace htable {

class LoadTestState {
public:
    LoadTestState(int percent_multipartition, int percent_conflict) :
            value_multipartition_(percentToThreshold(percent_multipartition)),
            value_conflict_(percentToThreshold(percent_conflict)) {}

    bool chooseMultiplePartition(RandomGenerator* generator) const;
    bool chooseConflict(RandomGenerator* generator) const;

private:
    // Convert percentage into the range [-1, RandomGenerator::maximum]. This ensures that at 0%
    // there are NO multipartition transactions, and at 100 they are *all* multipartition
    int32_t percentToThreshold(int percentage) const;

    // If random() < value_multipartition: send a multipartition transaction
    int32_t value_multipartition_;
    // If random() < value_conflict_: send a conflicting transaction
    int32_t value_conflict_;
};

static const int WARM_UP_SECONDS = 15;
static const int TEST_SECONDS = 60;

inline int partitionFunction(int num_logical_partitions, int num_physical_partitions, int partition_id) {
    assert(num_logical_partitions % num_physical_partitions == 0);
    assert(0 <= partition_id && partition_id < num_logical_partitions);

    int logical_partitions_per_physical = num_logical_partitions / num_physical_partitions;
    return partition_id / logical_partitions_per_physical;
}

/* DATA GENERATION:
* Partitions are "logical" partitions, not "physical" partitions
* Keys have values "%c%c%c" % (partition id, client id, variable id) 
* Single partition transaction: touch num_variables in a partition (vary variable id).
* Multi-partition transaction: touch num_variables across all logical partitions
    (vary partition id).
* Conflict single partition transaction: touch partition id client's first variable.
* Conflict multi-partition transaction: touch one partition client's first variable. This means we
    need num_logical_partitions MP txns for conflicts. This way the transaction only conflicts at
    one partition.

The general idea: we can define 12 variables and 2 partitions and have everything work the way
we had it before. Or we can define 16 variables and 16 partitions, then do 1, 2, 4, 8, 16 physical
partitions.

Conflicts require one "base load" client for each logical partition, then the extra clients need
one transaction for each logical partition.
*/
struct DataDefinition {
    DataDefinition(int num_variables, int num_logical_partitions) :
            num_variables_(num_variables),
            num_logical_partitions_(num_logical_partitions) {
        assert(num_variables_ >= num_logical_partitions_);
        assert(num_logical_partitions_ > 0);
        assert(num_variables_ % num_logical_partitions_ == 0);
    }

    int partitionFunction(int num_physical_partitions, int partition_id) const {
        return htable::partitionFunction(
                num_logical_partitions_, num_physical_partitions, partition_id);
    }

    static char idToChar(int id) {
        assert(0 <= id && id < NUM_ASCII_AFTER_A);
        return static_cast<char>(START_VARIABLE + id);
    }

    static int charToId(char c) {
        int id = c - START_VARIABLE;
        assert(0 <= id && id < NUM_ASCII_AFTER_A);
        return id;
    }

    const int num_variables_;
    const int num_logical_partitions_;

    static const char START_VARIABLE = 'A';
    
    // 'A' + 61 = '~'; next = not ascii
    static const int NUM_ASCII_AFTER_A = 62;
};

extern const DataDefinition DEFAULT_DATA;

inline int partitionFunction(
        int num_logical_partitions, int num_physical_partitions, const std::string& key) {
    int partition_id = DataDefinition::charToId(key[0]);
    return partitionFunction(num_logical_partitions, num_physical_partitions, partition_id);
}

// Returns a string with the variable name combining (partition, client, variable)
std::string makeVariableName(int partition, int client, int variable);

// Fills the parameters with the client workload, given the data definition.
void makeClientWorkload(
        int client_id, const DataDefinition& data_definition, bool conflicts_enabled,
        std::vector<htable::MiniTransaction>* single_partitions,
        htable::MiniTransaction* multi_partition,
        std::vector<htable::MiniTransaction>* conflict_singles,
        std::vector<htable::MiniTransaction>* conflict_multiples,
        std::vector<htable::MiniTransaction>* abort_singles,
        std::vector<htable::MiniTransaction>* abort_multiples,
        std::vector<dtxn::Partition>* partitions);

dtxn::Fragment toTransaction(const htable::MiniTransaction& mini);
std::vector<dtxn::Fragment> toTransaction(const std::vector<htable::MiniTransaction>& v);

std::string toString(const htable::MiniTransaction& mini);
std::vector<std::string> toString(const std::vector<htable::MiniTransaction>& v);

std::string debugString(const htable::MiniTransaction& mini);
std::string debugString(const std::vector<htable::MiniTransaction>& v);

// Returns a MiniTransaction touching num_variables keys in a logical partition.
htable::MiniTransaction makeSPTransaction(
        const htable::DataDefinition& data, int partition, int client);

// Returns a MiniTransaction touching num_variables keys that vary in the most significan byte.
htable::MiniTransaction makeMPTransaction(const htable::DataDefinition& data, int client);

std::vector<htable::MiniTransaction> makeConflictMultiples(
        const std::vector<htable::MiniTransaction>& transactions, int client);
std::vector<htable::MiniTransaction> makeConflictSingles(
        const std::vector<htable::MiniTransaction>& transactions, int client);

}  // namespace htable

#endif
