// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>
#include <tr1/unordered_map>

#include "base/assert.h"
#include "dtxn/dtxnclient.h"
#include "htablebenchmark.h"
#include "io/libeventloop.h"
#include "htable/messages.h"
#include "minitransactions/minitransactions.h"
#include "networkaddress.h"

using std::string;
using std::vector;
using htable::DEFAULT_DATA;

bool execute(
        io::LibEventLoop* event_loop,
        dtxn::DtxnClient* client,
        const htable::MiniTransaction& transaction,
        htable::MiniResult* result) {
    string serialized;
    transaction.appendToString(&serialized);
    string output;
    bool committed = client->blockingExecute(event_loop, serialized, &output);

    if (committed) {
        result->parseFromString(output.data(), output.data() + output.size());
    }
    return committed;
}

typedef std::tr1::unordered_map<string, int32_t> ValueMap;

// Do a CAS increment of a set of variables
void increment(
        io::LibEventLoop* event_loop,
        dtxn::DtxnClient* client,
        const vector<string>& variables,
        ValueMap* values) {
    assert(variables.size() > 0);

    // Partially initialized compare-and-swap transaction
    htable::MiniTransaction cas;
    htable::MiniTransaction read;

    values->clear();
    for (int i = 0; i < variables.size(); ++i) {
        read.read_set.push_back(variables[i]);
        (*values)[variables[i]] = 0;

        htable::KeyValuePair p;
        p.key = variables[i];
        p.hasValue = true;
        cas.write_set.push_back(p);
    }

    int count = -1;
    bool committed = false;
    while (!committed) {
        // read the current values
        htable::MiniResult result;
        committed = execute(event_loop, client, read, &result);
        // The read can abort if there is a deadlock timeout. Retry.
        // TODO: Add a deadlock or timeout status code?
        if (!committed) continue;
        assert(result.keyvalues.size() == values->size());
        for (int i = 0; i < result.keyvalues.size(); ++i) {
            int32_t v = 0;
            assert(values->find(result.keyvalues[i].key) != values->end());
            if (result.keyvalues[i].hasValue) {
                assert(result.keyvalues[i].value.size() == sizeof(v));
                memcpy(&v, result.keyvalues[i].value.data(), sizeof(v));
            }
            (*values)[result.keyvalues[i].key] = v;
        }

        // Update the compare set
        cas.compare_set = result.keyvalues;
        for (int i = 0; i < cas.write_set.size(); ++i) {
            // Increment and update the write set
            int32_t v = (*values)[cas.write_set[i].key] + 1;
            (*values)[cas.write_set[i].key] = v;
            cas.write_set[i].value.assign(reinterpret_cast<const char*>(&v), sizeof(v));
        }

        // Try to execute the transaction
        committed = execute(event_loop, client, cas, &result);
        count += 1;
    }

    if (count > 0) {
        printf("%d conflicts\n", count);
    }
}

// Returns the minimum key from variables, resetting it to zero.
void maximumReset(
        io::LibEventLoop* event_loop,
        dtxn::DtxnClient* client,
        const vector<string>& variables,
        string* key,
        int32_t* value) {
    assert(!variables.empty());
    htable::MiniTransaction minimum;
    htable::KeyValuePair p;
    p.key = "maximum";
    minimum.compare_set.push_back(p);
    minimum.read_set = variables;

    htable::MiniResult result;
    bool committed = execute(event_loop, client, minimum, &result);
    ASSERT(committed);
    assert(result.keyvalues.size() == 1);
    assert(result.keyvalues[0].hasValue);
    key->assign(result.keyvalues[0].key);
    assert(result.keyvalues[0].value.size() == sizeof(*value));
    memcpy(value, result.keyvalues[0].value.data(), sizeof(*value));
}

void maximumTest(io::LibEventLoop* event_loop, dtxn::DtxnClient* client) {
    ValueMap values;

    vector<string> variables;
    variables.push_back("a");
    increment(event_loop, client, variables, &values);
    variables.push_back("e");
    increment(event_loop, client, variables, &values);
    variables.push_back("z");
    increment(event_loop, client, variables, &values);

    string key;
    int32_t value;
    maximumReset(event_loop, client, variables, &key, &value);
    assert(key == "a");
    assert(value == 3);

    maximumReset(event_loop, client, variables, &key, &value);
    assert(key == "e");
    assert(value == 2);

    maximumReset(event_loop, client, variables, &key, &value);
    assert(key == "z");
    assert(value == 1);

    maximumReset(event_loop, client, variables, &key, &value);
    printf("last max = %s\n", key.c_str());
    assert(value == 0);
}

int main(int argc, const char* argv[]) {
    if (argc != 2) {
        fprintf(stderr, "htableclient [distributor]\n");
        return 1;
    }

    io::LibEventLoop event_loop;

    // Connect to the distributor
    NetworkAddress distributor;
    bool success = distributor.parse(argv[1]);
    ASSERT(success);
    dtxn::DtxnClient* client = dtxn::DtxnClient::create(&event_loop, distributor);

    //~ maximumTest(&client);

    ValueMap remaining;
    vector<string> all_variables;
    for (int partition = 0; partition < DEFAULT_DATA.num_logical_partitions_; ++partition) {
        for (int variable = 0; variable < DEFAULT_DATA.num_variables_; ++variable) {
            string variable_name = htable::makeVariableName(partition, variable, 0);
            all_variables.push_back(variable_name);
            remaining[variable_name] = 1000;
        }
    }
    srandom(static_cast<unsigned int>(time(NULL)));

    string max_key;
    int32_t max_value = 0;

    ValueMap last;
    while (!remaining.empty()) {
        // Generate a random vector of variables
        vector<string> variables;
        for (ValueMap::iterator j = remaining.begin(); j != remaining.end();) {
            ValueMap::iterator next = j;
            ++next;

            if (random() % 2) {
                variables.push_back(j->first);
                j->second -= 1;
                if (j->second == 0) {
                    remaining.erase(j);
                }
            }

            j = next;
        }

        if (!variables.empty()) {
            ValueMap out;
            increment(&event_loop, client, variables, &out);

            // Merge the output of this increment with all the other increments
            for (ValueMap::const_iterator i = out.begin(); i != out.end(); ++i) {
                last[i->first] = i->second;
            }
        }

        // Decide if we should clear the max value
        if (max_key.empty() && random() % 1000 == 0) {
            maximumReset(&event_loop, client, all_variables, &max_key, &max_value);
        }
    }

    for (ValueMap::const_iterator i = last.begin(); i != last.end(); ++i) {
        printf("%s -> %d\n", i->first.c_str(), i->second);
    }

    if (!max_key.empty()) {
        printf("\nreset variable %s value %d\n", max_key.c_str(), max_value);
    }

    delete client;
    return 0;
}
