// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include <stdlib.h>

#include <cassert>
#include <string>
#include <tr1/functional>

#include "base/assert.h"
#include "base/stlutil.h"
#include "benchmark/benchmark.h"
#include "dtxn/configparser.h"
#include "dtxn/dtxnclient.h"
#include "dtxn/executionengine.h"
#include "dtxn/messages.h"
#include "htablebenchmark.h"
#include "htable/messages.h"
#include "io/libeventloop.h"
#include "libevent/event.h"
#include "minitransactions/minitransactions.h"
#include "networkaddress.h"
#include "randomgenerator.h"

using std::string;
using std::vector;

using namespace htable;

class ClientLoadThread :
        public benchmark::BenchmarkWorker,
        public dtxn::TransactionCallback {
public:
    ClientLoadThread(NetworkAddress& distributor, const vector<dtxn::Partition>& partitions,
            LoadTestState* test_state,
            int simultaneous_requests, const vector<string>& serialized_single_partitions,
            const string& serialized_multiple_partition,
            const vector<string>& conflict_single_partitions,
            const vector<string>& conflict_multiple_partitions) :

            test_state_(test_state),
            simultaneous_requests_(simultaneous_requests),
            //~ client_counts_(partitions.size(), 0),
            //~ distributor_counts_(0),
            measuring_(false),
            pending_(0),
            transaction_count_(0),
            serialized_single_partitions_(serialized_single_partitions),
            serialized_multiple_partition_(serialized_multiple_partition),
            conflict_single_partitions_(conflict_single_partitions),
            conflict_multiple_partitions_(conflict_multiple_partitions) {
        assert(!partitions.empty());
        assert(test_state_ != NULL);
        assert(simultaneous_requests_ > 0);
        assert(serialized_single_partitions_.size() % partitions.size() == 0);

        // Select a random initial client
        generator_.seed((unsigned int) (time(NULL) + pthread_self()));
        next_sp_txn_ = static_cast<int>(generator_.random() % serialized_single_partitions_.size());

        // Connect to the distributor
        distributor_ = dtxn::DtxnClient::create(&eventloop_, distributor);

        // Create clients
        for (size_t i = 0; i < partitions.size(); ++i) {
            dtxn::DtxnClient* client =
                    dtxn::DtxnClient::create(&eventloop_, partitions[i].replica(0));
            clients_.push_back(client);
        }
    }

    ~ClientLoadThread() {
        STLDeleteElements(&clients_);
        delete distributor_;
    }

    virtual void transactionComplete(const dtxn::FragmentResponse& result) {
        //~ printf("%p transaction complete\n", this);
        // Don't count deadlock aborted transactions
        bool txn_success = result.status == dtxn::ExecutionEngine::OK ||
                result.status == dtxn::ExecutionEngine::ABORT_USER;
        if (txn_success && measuring_) {
            transaction_count_ += 1;
        }

        pending_ -= 1;
    }

    virtual void initialize() {}

    virtual void work(bool measure) {
        measuring_ = measure;

        while (pending_ < simultaneous_requests_) {
            //~ printf("%p dispatching request\n", this);
            if (test_state_->chooseMultiplePartition(&generator_)) {
                if (test_state_->chooseConflict(&generator_)) {
                    distributor_->execute(conflict_multiple_partitions_[next_sp_txn_], this);
                } else {
                    distributor_->execute(serialized_multiple_partition_, this);
                }
                //~ distributor_counts_ += 1;
            } else {
                int txns_per_partition = (int)(serialized_single_partitions_.size() / clients_.size());
                int next_client = next_sp_txn_ / txns_per_partition;
                //~ int next_client = generator_.random() % static_cast<int>(clients_.size());

                if (test_state_->chooseConflict(&generator_)) {
                    clients_[next_client]->execute(conflict_single_partitions_[next_sp_txn_], this);
                } else {
                    clients_[next_client]->execute(serialized_single_partitions_[next_sp_txn_], this);
                }

                //~ client_counts_[next_client_] += 1;
            }

            // This advances the client for non-conflicting multi-partition transactions. Does this
            // matter? It leads to a small but uniformly distributed "double sp txn".
            next_sp_txn_ += 1;
            if (next_sp_txn_ == serialized_single_partitions_.size()) {
                next_sp_txn_ = 0;
            }
            pending_ += 1;
        }

        // Loop until something changes (we receive a response)
        int last = pending_;
        while (pending_ == last) {
            event_base_loop(eventloop_.base(), EVLOOP_ONCE);
        }
    }

    virtual void finish() {
        // If we should exit, loop until all transactions are done
        while (pending_ > 0) {
            event_base_loop(eventloop_.base(), EVLOOP_ONCE);
        }
    }

    int transaction_count() const { return transaction_count_; }

private:
    LoadTestState* test_state_;
    int simultaneous_requests_;

    io::LibEventLoop eventloop_;
    dtxn::DtxnClient* distributor_;
    vector<dtxn::DtxnClient*> clients_;

    bool measuring_;

    //~ vector<int> client_counts_;
    //~ int distributor_counts_;

    RandomGenerator generator_;

    int pending_;
    int next_sp_txn_;
    int transaction_count_;

    vector<string> serialized_single_partitions_;
    string serialized_multiple_partition_;

    vector<string> conflict_single_partitions_;
    vector<string> conflict_multiple_partitions_;
};

int main(int argc, const char* argv[]) {    
    if (argc != 7 && argc != 8) {
        fprintf(stderr, "htableclientperf [distributor] [config file] [dispatch count] [threads] [percent multipartition] [percent conflict|0 percent abort]\n");
        return 1;
    }

    // Parse the distributor's address
    NetworkAddress distributor;
    bool success = distributor.parse(argv[1]);
    ASSERT(success);

    // Read and parse the configuration file.
    vector<dtxn::Partition> partitions = dtxn::parseConfigurationFromPath(argv[2]);
    assert(DEFAULT_DATA.num_variables_ % partitions.size() == 0);
    assert(DEFAULT_DATA.num_logical_partitions_ % partitions.size() == 0);

    int num_threads = atoi(argv[4]);
    int dispatch_count = atoi(argv[3]);
    assert(dispatch_count >= num_threads);
    int percent_multipartition = atoi(argv[5]);
    int percent_conflict = atoi(argv[6]);
    bool scale_conflicts = (argc == 7);
    int percent_aborts = 0;
    if (!scale_conflicts) {
        assert(percent_conflict == 0);
        percent_aborts = atoi(argv[7]);
    }
    LoadTestState test_state(percent_multipartition, scale_conflicts ? percent_conflict : percent_aborts);

    // Need to have enough clients to generate "conflicting" load
    assert(!scale_conflicts || (percent_conflict == 0 || num_threads >= DEFAULT_DATA.num_logical_partitions_ * 2));

    // Create threads with the partitions
    vector<ClientLoadThread*> threads;
    for (int client = 0; client < num_threads; ++client) {
        vector<MiniTransaction> single_partitions;
        MiniTransaction multiple_partition;
        vector<MiniTransaction> conflict_singles;
        vector<MiniTransaction> conflict_multiples;
        vector<MiniTransaction> abort_singles;
        vector<MiniTransaction> abort_multiples;
        vector<dtxn::Partition> partitions_copy(partitions);
        makeClientWorkload(client, DEFAULT_DATA, percent_conflict > 0,
                &single_partitions, &multiple_partition, &conflict_singles,
                &conflict_multiples, &abort_singles, &abort_multiples, &partitions_copy);

        // HACK; if we are running the abort test, replace the conficts with aborts
        if (!scale_conflicts) {
            conflict_singles = abort_singles;
            conflict_multiples = abort_multiples;
        }

        // Figure out how many requests this thread should dispatch
        int dispatch = dispatch_count / (num_threads - client);
        //~ printf("Thread %d dispatching %d requests\n", client, dispatch);
        dispatch_count -= dispatch;
        threads.push_back(
                new ClientLoadThread(distributor, partitions_copy, &test_state, dispatch,
                        toString(single_partitions), toString(multiple_partition),
                        toString(conflict_singles), toString(conflict_multiples)));
    }

    int64_t us = benchmark::runThreadBench(WARM_UP_SECONDS, TEST_SECONDS, threads, NULL);

    int total_count = 0;
    for (int i = 0; i < threads.size(); ++i) {
        total_count += threads[i]->transaction_count();
    }

    double seconds = (double) us/1000000.0;
    printf("%f\n", total_count/seconds);

    STLDeleteElements(&threads);
    return 0;
}
