// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "htableengine.h"

#include "base/stlutil.h"
#include "htable/messages.h"
#include "minitransactions/lockingkeyvaluetable.h"

using minitransactions::KeyValueEngine;
using std::string;

namespace htable {

HTableEngine::HTableEngine(bool locking) : locking_(locking) {
    if (locking) {
        engine_.table(new minitransactions::LockingKeyValueTable());
    } else {
        engine_.table(new minitransactions::SimpleKeyValueTable());
    }
}

dtxn::ExecutionEngine::Status HTableEngine::tryExecute(
        const string& work_unit, string* output, void** undo, dtxn::Transaction* transaction) {
    assert(locking_ || transaction == NULL);
    assert(output->empty());

    minitransactions::MiniTransaction miniwork_unit;
    parseFromString(work_unit, &miniwork_unit);

    minitransactions::KeyValueSet result;
    minitransactions::KeyValueSet* undo_record = NULL;
    if (undo != NULL) {
        if (*undo != NULL) {
            undo_record = (minitransactions::KeyValueSet*) *undo;
        } else {
            // TODO: Allocate on demand? Not needed for read-only transactions or if it aborts
            undo_record = new minitransactions::KeyValueSet();
            *undo = undo_record;
        }
    }
    KeyValueEngine::Status status =
            engine_.tryExecute(miniwork_unit, &result, undo_record, transaction);
    if (status != KeyValueEngine::BLOCKED) {
        htable::MiniResult result_message;
        for (minitransactions::KeyValueSet::const_iterator i = result.begin(); i != result.end(); ++i) {
            htable::KeyValuePair p;
            p.key = i->first;
            p.hasValue = i->second != NULL;
            if (p.hasValue) p.value = *i->second;
            result_message.keyvalues.push_back(p);
        }
        result_message.appendToString(output);
    }

    // If we are blocked, this contains partial results we will re-read (yes, wasteful, but...)
    STLDeleteValues(&result);

    if (status == KeyValueEngine::OK) {
        return OK;
    } else if (status == KeyValueEngine::ABORT) {
        return ABORT_USER;
    } else {
        assert(status == KeyValueEngine::BLOCKED);
        return BLOCKED;
    }
}

void HTableEngine::applyUndo(void* undo) {
    assert(undo != NULL);
    minitransactions::KeyValueSet* undo_record = reinterpret_cast<minitransactions::KeyValueSet*>(undo);
    engine_.applyUndo(*undo_record);
    freeUndo(undo);
}

void HTableEngine::freeUndo(void* undo) {
    assert(undo != NULL);
    minitransactions::KeyValueSet* undo_record = reinterpret_cast<minitransactions::KeyValueSet*>(undo);
    STLDeleteValues(undo_record);
    delete undo_record;
}

void HTableEngine::parseFromString(const string& data, minitransactions::MiniTransaction* output) {
    // The data in the string is serialized into a htable::MiniTransaction.
    htable::MiniTransaction mini;
    mini.parseFromString(data.data(), data.data() + data.size());

    // Parse out the read, write, and compare sets
    for (int i = 0; i < mini.read_set.size(); ++i) {
        output->read(mini.read_set[i]);
    }
    for (int i = 0; i < mini.write_set.size(); ++i) {
        if (mini.write_set[i].hasValue) {
            output->write(mini.write_set[i].key, mini.write_set[i].value);
        } else {
            output->erase(mini.write_set[i].key);
        }
    }
    for (int i = 0; i < mini.compare_set.size(); ++i) {
        if (mini.compare_set[i].hasValue) {
            output->compare(mini.compare_set[i].key, &(mini.compare_set[i].value));
        } else {
            output->compare(mini.compare_set[i].key, NULL);
        }
    }
}

}  // namespace htable
