// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include <sys/time.h>

#include "base/array.h"
#include "base/time.h"
#include "dtxn/locks/locktransaction.h"
#include "htablebenchmark.h"
#include "htableengine.h"
#include "htable/messages.h"

using namespace htable;
using std::string;
using std::vector;

struct Experiment {
    string transaction;
    string title;
    dtxn::ExecutionEngine* engine;
    bool use_undo;
    bool use_locks;
};

double timeWithParameters(const struct Experiment& experiment, int iterations) {
    void* undo_buffer = NULL;
    void** undo_pointer = NULL;
    if (experiment.use_undo) {
        undo_pointer = &undo_buffer;
    }

    dtxn::LockTransaction lock_transaction(NULL, experiment.transaction, false, true);
    vector<dtxn::LockTransaction*> granted_transactions;
    dtxn::Transaction* transaction_pointer = NULL;
    if (experiment.use_locks) {
        CHECK(experiment.use_undo);
        transaction_pointer = &lock_transaction;
    }

    string output;
    timeval start;
    int error = gettimeofday(&start, NULL);
    ASSERT(error == 0);
    for (int i = 0; i < iterations; ++i) {
        output.clear();
        dtxn::ExecutionEngine::Status status = experiment.engine->tryExecute(
                experiment.transaction, &output, undo_pointer, transaction_pointer);
        ASSERT(status == dtxn::ExecutionEngine::OK);
        if (undo_buffer != NULL) {
            experiment.engine->freeUndo(undo_buffer);
            undo_buffer = NULL;

            if (transaction_pointer != NULL) {
                lock_transaction.dropLocks(&granted_transactions);
                assert(granted_transactions.empty());
            }
        } else {
            assert(transaction_pointer == NULL);
        }
    }
    timeval end;
    error = gettimeofday(&end, NULL);
    ASSERT(error == 0);
    int64_t us = base::timevalDiffMicroseconds(end, start);
    return (double) us / iterations;
}

int main(int argc, const char* argv[]) {
    HTableEngine nolocks(false);
    HTableEngine locks(true);

    static const int PARTITION = 0;
    static const int CLIENT = 0;
    static const int NUM_PARTITIONS = 2;
    MiniTransaction single_partition = makeSPTransaction(DEFAULT_DATA, PARTITION, CLIENT);
    MiniTransaction multiple_partition = makeMPTransaction(DEFAULT_DATA, CLIENT);

    // Convert multiple_partition into just the parts that belong to partition 0
    CHECK(multiple_partition.compare_set.empty());
    for (int i = 0; i < multiple_partition.write_set.size(); ++i) {
        int key = DEFAULT_DATA.charToId(multiple_partition.write_set[i].key[0]);
        int partition = DEFAULT_DATA.partitionFunction(NUM_PARTITIONS, key);
        if (partition != PARTITION) {
            multiple_partition.write_set.erase(multiple_partition.write_set.begin() + i);
            --i;
        }
    }
    for (int i = 0; i < multiple_partition.read_set.size(); ++i) {
        int key = DEFAULT_DATA.charToId(multiple_partition.read_set[i][0]);
        int partition = DEFAULT_DATA.partitionFunction(NUM_PARTITIONS, key);
        if (partition != PARTITION) {
            multiple_partition.read_set.erase(multiple_partition.read_set.begin() + i);
            --i;
        }
    }

    // Set up experiments
    Experiment sp = { toString(single_partition), "single partition no undo", &nolocks, false, false };
    Experiment sp_undo = { sp.transaction, "single partition undo", &nolocks, true, false };
    Experiment sp_lock = { sp.transaction, "single partition locks", &locks, true, true };
    Experiment mp = { toString(multiple_partition), "multiple partition no undo", &nolocks, false, false };
    Experiment mp_undo = { mp.transaction, "multiple partition undo", &nolocks, true, false };
    Experiment mp_lock = { mp.transaction, "multiple partition locks", &locks, true, true };

    const Experiment* const experiments[] = {
        &sp, &sp_undo, &sp_lock, &mp, &mp_undo, &mp_lock,
    };
    static const int NUM_EXECUTES = 30000;
    static const int NUM_TRIALS = 3;
    for (int i = 0; i < base::arraySize(experiments); ++i) {
        printf("%s:", experiments[i]->title.c_str());
        // prime the test to warm caches and branch predictors
        timeWithParameters(*experiments[i], 100);
        double times[NUM_TRIALS];
        for (int j = 0; j < NUM_TRIALS; ++j) {
            times[j] = timeWithParameters(*experiments[i], NUM_EXECUTES);
        }
        for (int j = 0; j < NUM_TRIALS; ++j) {
            printf(" %f", times[j]);
        }
        printf("\n");
    }

    return 0;
}
