// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "minitransactiondistributor.h"

#include <cassert>
#include <cstring>

#include "base/stlutil.h"
#include "base/unordered_map.h"
#include "dtxn/dtxndistributor.h"
#include "htablebenchmark.h"
#include "htable/messages.h"
#include "strings/utils.h"

using dtxn::CoordinatedTransaction;
using std::string;

namespace {

struct StateVariant {
    enum Type {
        RESET_MAX,
        TWO_ROUND,
    };

    StateVariant(Type type) : type_(type) {}
    virtual ~StateVariant() {}
    
    Type type_;
};

struct ResetMaxState : public StateVariant {
    ResetMaxState() : StateVariant(RESET_MAX), state(READ) {}

    enum State {
        READ,
        RESET,
    };

    State state;
    string max_key;
    int32_t max_value;
};

typedef base::unordered_map<int, htable::MiniTransaction> RequestMap;

struct TwoRoundState : public StateVariant {
    TwoRoundState() : StateVariant(TWO_ROUND), round_(0) {}

    int round_;
    RequestMap original_requests_;
    htable::MiniResult read_response_;
};

}  // namespace

const char MiniTransactionDistributor::RESET_MAXIMUM_KEY[] = "maximum";

void MiniTransactionDistributor::begin(CoordinatedTransaction* transaction) {
    assert(transaction->received().size() == 1);
    assert(transaction->received()[0].first == CoordinatedTransaction::CLIENT_PARTITION);
    const string& client_message = transaction->received()[0].second;
    assert(!client_message.empty());

    htable::MiniTransaction mini;
    mini.parseFromString(client_message);

    // HACK: This implements the "maximum = 0" transaction
    if (mini.compare_set.size() == 1 && mini.compare_set[0].key == RESET_MAXIMUM_KEY) {
        assert(mini.write_set.size() == 0);
        assert(mini.read_set.size() > 0);
        transaction->state(new ResetMaxState());
        mini.compare_set.clear();
    }

    // Split the transaction using the partition function
    RequestMap requests;

    for (int i = 0; i < mini.compare_set.size(); ++i) {
        int partition = partitionFunction(mini.compare_set[i].key);
        requests[partition].compare_set.push_back(mini.compare_set[i]);
    }
    for (int i = 0; i < mini.write_set.size(); ++i) {
        int partition = partitionFunction(mini.write_set[i].key);
        requests[partition].write_set.push_back(mini.write_set[i]);
    }
    for (int i = 0; i < mini.read_set.size(); ++i) {
        int partition = partitionFunction(mini.read_set[i]);
        requests[partition].read_set.push_back(mini.read_set[i]);
    }

    // Handle the two round variant (general transaction)
    if (use_two_round_variant_ && requests.size() > 1) {
        // initially execute only the reads
        assert(transaction->state() == NULL);
        TwoRoundState* s = new TwoRoundState();
        transaction->state(s);

        // copy the original requests; eliminate the writes
        s->original_requests_ = requests;
        for (RequestMap::iterator i = requests.begin(); i != requests.end(); ++i) {
            i->second.write_set.clear();
            assert(i->second.compare_set.empty());
        }
    }

    // Send the messages
    string out;
    for (RequestMap::const_iterator i = requests.begin(); i != requests.end(); ++i) {
        out.clear();
        i->second.appendToString(&out);
        transaction->send(i->first, out);
    }
    assert(transaction->sent().size() == requests.size());

    if (transaction->state() == NULL) {
        // Normal: we know that we are finished sending work: all are one-shot/single partition
        // TODO: Mark uninvolved partitions as done for reset max?
        transaction->setAllDone();
    }
    sendRound(transaction);
}

int MiniTransactionDistributor::partitionFunction(const string& key) const {
    return htable::partitionFunction(num_logical_partitions_, num_physical_partitions_, key);
}
    
static htable::MiniResult combineResults(const CoordinatedTransaction::MessageList& input) {
    htable::MiniResult result;
    for (int i = 0; i < input.size(); ++i) {
        htable::MiniResult partial;
        partial.parseFromString(input[i].second);
        STLExtend(&result.keyvalues, partial.keyvalues);
    }
    return result;
}

void MiniTransactionDistributor::nextRound(CoordinatedTransaction* transaction) {
    synchronousNextRound(transaction);
    sendRound(transaction);
}

void MiniTransactionDistributor::synchronousNextRound(CoordinatedTransaction* transaction) {
    StateVariant* state = (StateVariant*) transaction->state();

    if (state != NULL && state->type_ == StateVariant::RESET_MAX) {
        ResetMaxState* reset_state = (ResetMaxState*) state;
        if (reset_state->state == ResetMaxState::READ) {
            // Find the maximum of all the reads
            htable::MiniResult result = combineResults(transaction->received());
            assert(!result.keyvalues.empty());

            const htable::KeyValuePair* max = NULL;
            int32_t max_value = 0;
            for (int i = 0; i < result.keyvalues.size(); ++i) {
                if (result.keyvalues[i].hasValue) {
                    int32_t v;
                    strings::assignBytes(&v, result.keyvalues[i].value);
                    if (v > max_value || max == NULL) {
                        max = &result.keyvalues[i];
                        max_value = v;
                    }
                }
            }
            assert(max != NULL);

            // Package up the write
            htable::MiniTransaction txn;
            // compare not strictly needed. It must never fail, so this verifies that.
            txn.compare_set.push_back(*max);
            txn.write_set.push_back(*max);
            strings::assignBytes(&txn.write_set.back().value, (int32_t) 0);

            // Remember the maximum
            reset_state->max_key = max->key;
            reset_state->max_value = max_value;
            reset_state->state = ResetMaxState::RESET;
            string out;
            txn.appendToString(&out);
            transaction->send(partitionFunction(max->key), out);
            transaction->setAllDone();
            return;
        }

        assert(reset_state->state == ResetMaxState::RESET);
        assert(transaction->received().size() == 1);

        htable::MiniResult result;
        result.parseFromString(transaction->received()[0].second);

        // Package up the real result
        assert(result.keyvalues.size() == 0);
        result.keyvalues.resize(1);
        result.keyvalues.back().key = reset_state->max_key;
        result.keyvalues.back().hasValue = true;
        strings::assignBytes(&result.keyvalues.back().value, reset_state->max_value);

        string out;
        result.appendToString(&out);
        transaction->commit(out);
        return;
    } else if (state != NULL && state->type_ == StateVariant::TWO_ROUND) {
        TwoRoundState* two_round_state = (TwoRoundState*) state;
        if (two_round_state->round_ == 0) {
            // Combine and store the read results
            two_round_state->read_response_ = combineResults(transaction->received());

            // Send the next round
            string out;
            for (RequestMap::iterator i = two_round_state->original_requests_.begin(); i != two_round_state->original_requests_.end(); ++i) {
                i->second.read_set.clear();
                out.clear();
                i->second.appendToString(&out);
                transaction->send(i->first, out);
            }
            transaction->setAllDone();
        } else {
            assert(two_round_state->round_ == 1);
            
            // return the results
            string out;
            two_round_state->read_response_.appendToString(&out);
            transaction->commit(out);
        }
        two_round_state->round_ += 1;
        return;
    }
    assert(state == NULL);

    if (transaction->received().size() == 1) {
        // Single response = single partition. Just pass it back to the client
        transaction->commit(transaction->received()[0].second);
    } else {
        assert(transaction->received().size() == 2);
        // Combine the two inputs together
        htable::MiniResult result = combineResults(transaction->received());

        string out;
        result.appendToString(&out);
        transaction->commit(out);
    }
}

void MiniTransactionDistributor::done(CoordinatedTransaction* transaction) {
    delete (StateVariant*) transaction->state();
}
