// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "minitransactions/keyvalueengine.h"

#include <cassert>
#include <cstdio>
#include <cstring>

#include "base/stlutil.h"

using dtxn::Transaction;
using std::string;
using std::pair;

namespace minitransactions {

KeyValueEngine::Status KeyValueEngine::tryExecute(
        const MiniTransaction& minitransaction,
        KeyValueSet* output,
        KeyValueSet* undo,
        Transaction* transaction) {
    assert(output->empty());
    // if transaction is specified, we *must* have an undo buffer for deadlocks
    assert(transaction == NULL || undo != NULL);

    bool equal = false;
    if (!compare(transaction, minitransaction, &equal)) return BLOCKED;
    if (!equal) return ABORT;
    assert(equal);

    if (!read(transaction, minitransaction, output)) return BLOCKED;
    if (!write(transaction, minitransaction, undo)) return BLOCKED;
    return OK;
}

void KeyValueEngine::applyUndo(const KeyValueSet& undo) {
    for (KeyValueSet::const_iterator i = undo.begin(); i != undo.end(); ++i) {
        table_->write(NULL, i->first, i->second, NULL);
    }
}

bool KeyValueEngine::compare(Transaction* transaction, const MiniTransaction& minitransaction, bool* equal) {
    for (KeyValueSet::const_iterator i = minitransaction.compare_set().begin();
            i != minitransaction.compare_set().end(); ++i) {
        const string* result = NULL;
        if (!table_->read(transaction, i->first, &result)) return false;

        if (i->second == NULL) {
            // must not exist
            if (result != NULL) {
                *equal = false;
                return true;
            }
        } else {
            // must match value
            if (result == NULL || *i->second != *result) {
                *equal = false;
                return true;
            }
        }
    }

    *equal = true;
    return true;
}

bool KeyValueEngine::read(Transaction* transaction, const MiniTransaction& minitransaction, KeyValueSet* output) {
    assert(output->empty());  // TODO: Not true if re-executed

    for (MiniTransaction::ReadSet::const_iterator i = minitransaction.read_set().begin();
            i != minitransaction.read_set().end(); ++i) {
        const string* result = NULL;
        if (!table_->read(transaction, *i, &result)) return false;

        string* resultptr = NULL;
        if (result != NULL) {
            resultptr = new string(*result);
        }
        output->insert(std::make_pair(*i, resultptr));
    }
    return true;
}

bool KeyValueEngine::write(Transaction* transaction, const MiniTransaction& minitransaction, KeyValueSet* undo) {
    for (KeyValueSet::const_iterator i = minitransaction.write_set().begin();
            i != minitransaction.write_set().end(); ++i) {
        string* undo_record = NULL;
        string** undo_ptr = NULL;
        if (undo != NULL) {
            undo_ptr = &undo_record;
        }
        // Update the value
        if (!table_->write(transaction, i->first, i->second, undo_ptr)) return false;

        if (undo != NULL) {
            // TODO: Replace with a list of (key, value) pairs, to avoid extra hash lookups?
            pair<KeyValueSet::iterator, bool> result = undo->insert(pair<string, string*>(i->first, NULL));
            if (result.second) {
                // have not written this value before: we need to save it
                result.first->second = undo_record;
            } else {
                delete undo_record;
            }
        }
    }
    return true;
}

SimpleKeyValueTable::~SimpleKeyValueTable() {
    STLDeleteValues(&table_);
}

/** Returns true if the result can be read. False indicates the transaction is blocked. */
bool SimpleKeyValueTable::read(Transaction* transaction, const string& key,
        const string** output) {
    assert(transaction == NULL);
    KeyValueSet::iterator it = table_.find(key);
    if (it == table_.end()) {
        *output = NULL;
    } else {
        *output = it->second;
    }
    return true;
}

/** Returns true if the result can be written. False indicates the transaction is blocked. */
bool SimpleKeyValueTable::write(Transaction* transaction, const string& key,
        const string* value, string** undo) {
    assert(transaction == NULL);
    // TODO: Do not insert keys that are going to be erased?
    pair<KeyValueSet::iterator, bool> result = table_.insert(pair<string, string*>(key, NULL));

    if (undo != NULL) {
        // Save the old value
        if (result.first->second == NULL) {
            *undo = NULL;
        } else {
            *undo = new string(*result.first->second);
        }
    }

    if (value == NULL) {
        // TODO: Erase keys that are NULL?
        delete result.first->second;
        result.first->second = NULL;
    } else if (result.first->second != NULL) {
        result.first->second->assign(*value);
    } else {
        result.first->second = new string(*value);
    }

    return true;
}

void SimpleKeyValueTable::print() {
    for (KeyValueSet::const_iterator i = table_.begin(); i != table_.end(); ++i) {
        int32_t value;
        assert(i->second->size() == sizeof(value));
        memcpy(&value, i->second->data(), sizeof(value));
        printf("%s: %d\n", i->first.c_str(), value);
    }
}

}  // namespace minitransactions
