// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#ifndef MINITRANSACTIONS_KEYVALUETABLE_H__
#define MINITRANSACTIONS_KEYVALUETABLE_H__

#include <cassert>
#include <string>

#include "minitransactions/minitransactions.h"  // for KeyValueSet

namespace dtxn {
class Transaction;
}

namespace minitransactions {

/** Interface that provides access to a key/value table. */
class KeyValueTable {
public:
    virtual ~KeyValueTable() {}

    /** Returns true if the result can be read. False indicates the transaction is blocked. */
    virtual bool read(dtxn::Transaction* transaction, const std::string& key,
            const std::string** output) = 0;

    /** Returns true if the result can be written. False indicates the transaction is blocked. */
    virtual bool write(dtxn::Transaction* transaction, const std::string& key,
            const std::string* value, std::string** undo) = 0;

    /** Prints the contents of the table. Used for debugging. */
    virtual void print() = 0;
};


/** The minitransaction engine. It can use a blocking or a locking VariableTable. */
class KeyValueEngine {
public:
    explicit KeyValueEngine(KeyValueTable* table = NULL) : table_(table) {}

    ~KeyValueEngine() {
        delete table_;
    }

    /** Sets the table used by this engine. May only be called once. Owns table. */
    void table(KeyValueTable* table) {
        assert(table_ == NULL);
        assert(table != NULL);
        table_ = table;
    }

    enum Status {
        /** Transaction executed correctly. */
        OK,
        /** Compare failed. */
        ABORT,
        /** The transaction blocked reading or writing a key. */
        BLOCKED,
    };

    // Execute a minitransaction with an undo log. Returns true if it commits. Output will be
    // filled with the values read from the table. Undo will be filled with the undo data.
    Status tryExecute(const MiniTransaction& minitransaction, KeyValueSet* output,
            KeyValueSet* undo, dtxn::Transaction* transaction);

    // Undoes a transaction by applying the undo buffer.
    void applyUndo(const KeyValueSet& undo);

    // TODO: Replace this with an iterator or a minitransaction iterator
    void print() const { table_->print(); }

private:
    // Sets *equal to true if the compare set values match.
    bool compare(dtxn::Transaction* transaction, const MiniTransaction& minitransaction,
            bool* equal);
    bool read(dtxn::Transaction* transaction, const MiniTransaction& minitransaction,
            KeyValueSet* output);
    bool write(dtxn::Transaction* transaction, const MiniTransaction& minitransaction,
            KeyValueSet* undo);

    KeyValueTable* table_;
};

/** An implementation of VariableTable that never blocks (no concurrency control). */
class SimpleKeyValueTable : public KeyValueTable {
public:
    virtual ~SimpleKeyValueTable();
    virtual bool read(dtxn::Transaction* transaction, const std::string& key,
            const std::string** output);
    virtual bool write(dtxn::Transaction* transaction, const std::string& key,
            const std::string* value, std::string** undo);
    virtual void print();

private:
    KeyValueSet table_;
};


}  // namespace minitransactions

#endif
