// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "base/stlutil.h"
#include "dtxn/locks/locktransaction.h"
#include "minitransactions/lockingkeyvaluetable.h"
#include "minitransactions/keyvalueengine.h"
#include "stupidunit/stupidunit.h"

using dtxn::LockTransaction;
using std::string;
using std::vector;

using namespace minitransactions;

class KeyValueEngineTestBase : public Test {
public:
    KeyValueEngineTestBase(KeyValueTable* table) : table_(table) {}

    virtual ~KeyValueEngineTestBase() {
        clearOutput();
    }

protected:
    void clearOutput() {
        STLDeleteValues(&output_);
    }

    void SingleRead() {
        // Read "foo". It should not exist (NULL).
        transaction_.read("foo");
        transaction_.read("foo");  // add multiple
        EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(NULL, output_["foo"]);

        // Write "foo".
        clearOutput();
        MiniTransaction write;
        write.write("foo", "foo value");
        write.write("foo", "real foo value");  // replaces
        EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(write, &output_, NULL, NULL));
        EXPECT_EQ(0, output_.size());

        // Read "foo."
        clearOutput();
        EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ("real foo value", *output_["foo"]);
    }

    void ReadWrite() {
        // Read "foo" (NULL) and replace it.
        transaction_.read("foo");
        transaction_.write("foo", "replace");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(NULL, output_["foo"]);

        // Read/write foo again
        clearOutput();
        transaction_.write("foo", "value again");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ("replace", *output_["foo"]);
    }

    void Erase() {
        // Create "foo"
        transaction_.write("foo", "bar");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));

        // Read/erase foo.
        clearOutput();
        transaction_.read("foo");
        transaction_.erase("foo");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ("bar", *output_["foo"]);

        // Read/erase foo again: erasing a second time is harmless
        clearOutput();
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(NULL, output_["foo"]);
    }

    void CompareSuccess() {
        // Look for a non-existent key in the table
        const string value("bar");
        transaction_.compare("foo", &value);
        ASSERT_EQ(KeyValueEngine::ABORT, table_.tryExecute(transaction_, &output_, NULL, NULL));

        transaction_.compare("foo", NULL);  // replaces
        transaction_.read("foo");
        transaction_.write("foo", value);

        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(NULL, output_["foo"]);

        clearOutput();
        transaction_.compare("foo", &value);
        transaction_.erase("foo");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.size());
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(value, *output_["foo"]);
    }

    void Undo() {
        // Assign foo = "bar"
        transaction_.write("foo", "bar");
        KeyValueSet undo;
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, &undo, NULL));

        // Verify that it wrote the value
        MiniTransaction read;
        read.read("foo");
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(read, &output_, NULL, NULL));
        ASSERT_EQ(1, output_.count("foo"));
        EXPECT_EQ("bar", *output_["foo"]);

        // Write with a different value
        transaction_.write("foo", "baz");
        clearOutput();
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, &undo, NULL));
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(read, &output_, NULL, NULL));
        ASSERT_EQ(1, output_.count("foo"));
        EXPECT_EQ("baz", *output_["foo"]);

        // Undo the transaction
        table_.applyUndo(undo);

        // Verify that it was undone
        clearOutput();
        ASSERT_EQ(KeyValueEngine::OK, table_.tryExecute(read, &output_, NULL, NULL));
        EXPECT_EQ(1, output_.count("foo"));
        EXPECT_EQ(NULL, output_["foo"]);
    }

    KeyValueEngine table_;
    MiniTransaction transaction_;
    KeyValueSet output_;
};

class SimpleKeyValueEngineTest : public KeyValueEngineTestBase {
public:
    SimpleKeyValueEngineTest() : KeyValueEngineTestBase(new SimpleKeyValueTable()) {}
};

class LockingKeyValueEngineTest : public KeyValueEngineTestBase {
public:
    LockingKeyValueEngineTest() : KeyValueEngineTestBase(new LockingKeyValueTable()) {}
};

TEST_F(SimpleKeyValueEngineTest, InitializeAfterConstructor) {
    // already initialized
    EXPECT_DEATH(table_.table(new SimpleKeyValueTable()));

    KeyValueEngine engine;
    engine.table(new SimpleKeyValueTable());
    EXPECT_DEATH(engine.table(new SimpleKeyValueTable()));
}

#define BOTH(Name) \
TEST_F(SimpleKeyValueEngineTest, Name) { Name(); } \
TEST_F(LockingKeyValueEngineTest, Name) { Name(); }

BOTH(SingleRead)
BOTH(ReadWrite)
BOTH(Erase)
BOTH(CompareSuccess)
BOTH(Undo)

TEST_F(LockingKeyValueEngineTest, BlockedRead) {
    LockTransaction txn(NULL, "foo", false, true);
    transaction_.read("foo");
    KeyValueSet undo;

    EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, &undo, &txn));
    EXPECT_EQ(1, output_.size());
    EXPECT_EQ(1, output_.count("foo"));
    EXPECT_EQ(NULL, output_["foo"]);

    // Write "foo".
    clearOutput();
    MiniTransaction write;
    write.write("foo", "real foo value");
    EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(write, &output_, &undo, &txn));
    EXPECT_EQ(0, output_.size());

    // Read "foo".
    clearOutput();
    EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, &undo, &txn));
    EXPECT_EQ(1, output_.size());
    EXPECT_EQ("real foo value", *output_["foo"]);

    // Read "foo" in a new transaction: blocked
    LockTransaction txn2(NULL, "foo", false, true);
    clearOutput();
    EXPECT_EQ(KeyValueEngine::BLOCKED, table_.tryExecute(transaction_, &output_, &undo, &txn2));
    EXPECT_EQ(0, output_.size());

    // Read "bar" in a new transaction: not blocked
    LockTransaction txn3(NULL, "foo", false, true);
    clearOutput();
    MiniTransaction readOther;
    readOther.read("bar");
    EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(readOther, &output_, &undo, &txn3));
    EXPECT_EQ(NULL, output_["bar"]);
    vector<LockTransaction*> granted_transactions;
    txn3.dropLocks(&granted_transactions);
    ASSERT_EQ(0, granted_transactions.size());
    clearOutput();

    txn.dropLocks(&granted_transactions);
    ASSERT_EQ(1, granted_transactions.size());
    txn2.lockGranted();

    // read now succeeds
    EXPECT_EQ(KeyValueEngine::OK, table_.tryExecute(transaction_, &output_, &undo, &txn2));
    EXPECT_EQ(1, output_.size());
    EXPECT_EQ("real foo value", *output_["foo"]);
    txn2.dropLocks(&granted_transactions);
}

int main() {
    return TestSuite::globalInstance()->runAll();
}
