// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

#include "minitransactions/lockingkeyvaluetable.h"

#include <cstring>

#include "base/stlutil.h"
#include "dtxn/locks/lock.h"
#include "dtxn/transaction.h"

using dtxn::Transaction;
using std::pair;
using std::string;

namespace minitransactions {

class LockingKeyValueTable::LockingString {
public:
    LockingString() : erased_(true) {}

    dtxn::Lock lock_;
    string value_;
    bool erased_;

private:
    LockingString(const LockingString& other);
    LockingString& operator=(const LockingString& other);
};
   
LockingKeyValueTable::~LockingKeyValueTable() {
    STLDeleteValues(&table_);
}

bool LockingKeyValueTable::read(Transaction* transaction, const string& key, const string** output) {
    // Lock for reading
    LockingString* value = getValue(key);
    if (transaction != NULL && !transaction->tryReadLock(&value->lock_)) return false;

    if (value->erased_) {
        *output = NULL;
    } else {
        *output = &(value->value_);
    }
    return true;
}

bool LockingKeyValueTable::write(Transaction* transaction, const string& key, const string* next_value, string** undo) {
    // Lock for writing
    LockingString* value = getValue(key);
    if (transaction != NULL && !transaction->tryWriteLock(&value->lock_)) return false;

    if (undo != NULL) {
        // Save the old value
        if (value->erased_) {
            *undo = NULL;
        } else {
            *undo = new string(value->value_);
        }
    }

    if (next_value == NULL) {
        value->erased_ = true;
        value->value_.clear();
    } else {
        value->value_.assign(*next_value);
        value->erased_ = false;
    }

    return true;
}

void LockingKeyValueTable::print() {
    for (TableType::const_iterator i = table_.begin(); i != table_.end(); ++i) {
        int32_t value;
        assert(i->second->value_.size() == sizeof(value));
        memcpy(&value, i->second->value_.data(), sizeof(value));
        printf("%s: %d\n", i->first.c_str(), value);
    }
}

LockingKeyValueTable::LockingString* LockingKeyValueTable::getValue(const string& key) {
    // TODO: How to handle locking non-existant keys?
    // TODO: Better way to lock erase requests?
    pair<TableType::iterator, bool> result = table_.insert(pair<string, LockingString*>(key, NULL));
    if (result.second) {
        assert(result.first->second == NULL);
        result.first->second = new LockingString();
    }

    return result.first->second;
}

}  // namespace minitransactions
