#!/usr/bin/python

import csv
import os.path
import sys
import time

import frange
from third_party.pystdlib import subprocess

PERF = "./tpccclientperf"
PERF_LOCK = "./tpccclientperf-lock"

TRIALS = 3
CLIENTS = 40
SAMPLE_POINTS = (list(frange.frange(0, 0.021, 0.002)) +
        list(frange.frange(0.03, 0.11, 0.01)) +
        [0.15, 0.2, 0.3, 1])


def newOrderRemoteProbability(p_item_remote):
    """Returns the probability that a new order is remote, given the probability that an item in
    the order is remote. The default value is 0.01."""

    # This is derived as follows:
    # p(order with n items is local) = (1-p)**n
    # p(new order is local) = 1/11*(1-p)**5 + 1/11*(1-p)**6 + ... + 1/11*(1-p)**15
    # p(new order is remote) = 1 - p(new order is local)
    # Solve using the formula for the sum of a geometric series, which is not valid for r == 1

    assert 0 <= p_item_remote and p_item_remote <= 1
    if p_item_remote == 0: return 0
    p_local = ((1-p_item_remote)**5 - (1-p_item_remote)**16)/(11.*p_item_remote)
    return 1 - p_local


def multiPartitionProbability(warehouses, partitions):
    assert warehouses % partitions == 0

    NEW_ORDER_FRACTION = 0.45
    PAYMENT_FRACTION = 0.43
    DEFAULT_ITEM_PROBABILITY = 0.01
    PAYMENT_REMOTE_PROBABILITY = 0.15

    warehouses_per_partition = warehouses / float(partitions)
    remote_warehouse_probability = (warehouses - warehouses_per_partition) / float(warehouses - 1)

    remote_item_probability = DEFAULT_ITEM_PROBABILITY * remote_warehouse_probability
    multi_partition = newOrderRemoteProbability(remote_item_probability) * NEW_ORDER_FRACTION

    multi_partition += PAYMENT_FRACTION* PAYMENT_REMOTE_PROBABILITY * remote_warehouse_probability
    return multi_partition


def runTest(distributor, configuration_file, mode, warehouses):
    results = []

    for p_item_remote in SAMPLE_POINTS:
        # client perf takes new order item value is thousands of probability
        milli_item_remote = int(p_item_remote * 1000)
        percent_multipartition = newOrderRemoteProbability(milli_item_remote/1000.) * 100
        row = [percent_multipartition]
        print "%.1f%% multipartition" % row[0], 
        sys.stdout.flush()
        for _ in range(TRIALS):
            command = (PERF, distributor)
            if mode == "lock":
                command = (PERF_LOCK,)
            child = subprocess.Popen(command + (configuration_file, str(warehouses),
                    str(CLIENTS), str(milli_item_remote)), stdout=subprocess.PIPE)
            output = child.stdout.read()
            result = child.wait()
            assert result == 0

            output = float(output.strip())
            row.append(output)
            print "\t", row[-1],
            sys.stdout.flush()
            # wait 10 seconds for the servers to reload
            time.sleep(10)
        results.append(row)
        print

    output_path = os.path.splitext(configuration_file)[0] + ".csv"
    out = open(output_path, "wb")
    writer = csv.writer(out)
    writer.writerows(results)
    out.close()


if __name__ == "__main__":
    if len(sys.argv) != 5:
        sys.stderr.write("scale_tpcc_mptxns.py [distributor] [config file] [mode] [warehouses]\n")
        sys.exit(1)

    distributor = sys.argv[1]
    configuration_file = sys.argv[2]
    mode = sys.argv[3]
    warehouses = int(sys.argv[4])

    runTest(distributor, configuration_file, mode, warehouses)
