// Copyright 2008,2009,2010 Massachusetts Institute of Technology.
// All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

// Load driver for TPC-C

#include <string>
#include <tr1/functional>
#include <vector>

#include "base/function_thread.h"
#include "base/mutex.h"
#include "base/stlutil.h"
#include "bench/clock.h"
#include "bench/randomgenerator.h"
#include "bench/tpccdb.h"
#include "bench/tpccclient.h"
#include "benchmark/benchmark.h"
#include "dtxn/configparser.h"
#include "dtxn/dtxnclient.h"
#include "io/libeventloop.h"
#include "serialization.h"
#include "tpccengine.h"

using std::string;
using std::vector;

/** Provides a blocking interface backed by dtxn connections.
NOTE: This implements the distribution logic even though technically that belongs in the
distributor. */
class TPCCClientDb : public TPCCDB {
public:
    TPCCClientDb(int num_warehouses, NetworkAddress& distributor,
            const vector<dtxn::Partition>& partitions) : num_warehouses_(num_warehouses) {
        assert(num_warehouses_ > 0);
        assert(!partitions.empty());
        assert(num_warehouses % partitions.size() == 0);

        // Connect to the distributor
        distributor_ = dtxn::DtxnClient::create(&eventloop_, distributor);

        // Create clients
        for (size_t i = 0; i < partitions.size(); ++i) {
            dtxn::DtxnClient* client =
                    dtxn::DtxnClient::create(&eventloop_, partitions[i].replica(0));
            clients_.push_back(client);
        }
    }

    virtual ~TPCCClientDb() {
        delete distributor_;
        STLDeleteElements(&clients_);
    }

    virtual int32_t stockLevel(int32_t warehouse_id, int32_t district_id, int32_t threshold) {
        serializeStockLevel(warehouse_id, district_id, threshold, &request_);
        bool committed = makeRequest(warehouse_id);
        ASSERT(committed);

        int32_t level;
        const char* real_end = result_.data() + result_.size();
        const char* end = serialization::deserialize(&level, result_.data(), real_end);
        ASSERT(real_end == end);
        assert(level >= 0);
        return level;
    }

    virtual void orderStatus(int32_t warehouse_id, int32_t district_id, int32_t customer_id,
            OrderStatusOutput* output) {
        serializeOrderStatus(warehouse_id, district_id, customer_id, &request_);
        bool committed = makeRequest(warehouse_id);
        ASSERT(committed);
        deserializeOrderStatusOutput(result_, output);
    }

    virtual void orderStatus(int32_t warehouse_id, int32_t district_id, const char* c_last,
            OrderStatusOutput* output) {
        serializeOrderStatus(warehouse_id, district_id, c_last, &request_);
        bool committed = makeRequest(warehouse_id);
        ASSERT(committed);
        deserializeOrderStatusOutput(result_, output);
    }

    virtual bool newOrder(int32_t warehouse_id, int32_t district_id, int32_t customer_id,
            const std::vector<NewOrderItem>& items, const char* now,
            NewOrderOutput* output, TPCCUndo** undo) {
        assert(undo == NULL);
        bool cross_partition = false;
        for (size_t i = 0; i < items.size(); ++i) {
            if (partition(items[i].ol_supply_w_id) != partition(warehouse_id)) {
                cross_partition = true;
                break;
            }
        }

        TPCCEngine::serializeNewOrder(TPCCEngine::NEW_ORDER, warehouse_id, district_id, customer_id,
                items, now, &request_);
        int id = warehouse_id;
        if (cross_partition) id = -1;
        bool committed = makeRequest(id);
        TPCCEngine::deserializeNewOrderOutput(result_, output);
        return committed;
    }

    virtual bool newOrderHome(int32_t warehouse_id, int32_t district_id, int32_t customer_id,
            const std::vector<NewOrderItem>& items, const char* now,
            NewOrderOutput* output, TPCCUndo** undo) {
        CHECK(false);
        return false;
    }

    virtual bool newOrderRemote(int32_t home_warehouse, int32_t remote_warehouse,
            const std::vector<NewOrderItem>& items, std::vector<int32_t>* out_quantities,
            TPCCUndo** undo) {
        CHECK(false);
        return false;
    }

    virtual void payment(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, int32_t customer_id, float h_amount, const char* now,
            PaymentOutput* output, TPCCUndo** undo) {
        assert(undo == NULL);
        internalPayment(warehouse_id, district_id, c_warehouse_id, c_district_id, customer_id, NULL,
                h_amount, now, output);
    }

    virtual void payment(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, const char* c_last, float h_amount, const char* now,
            PaymentOutput* output, TPCCUndo** undo) {
        assert(undo == NULL);
        internalPayment(warehouse_id, district_id, c_warehouse_id, c_district_id, 0, c_last,
                h_amount, now, output);
    }

    virtual void paymentHome(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, int32_t c_id, float h_amount, const char* now,
            PaymentOutput* output, TPCCUndo** undo) {
        CHECK(false);
    }
    virtual void paymentRemote(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, int32_t c_id, float h_amount, PaymentOutput* output,
            TPCCUndo** undo) {
        CHECK(false);
    }
    virtual void paymentRemote(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, const char* c_last, float h_amount, PaymentOutput* output,
            TPCCUndo** undo) {
        CHECK(false);
    }

    virtual void delivery(int32_t warehouse_id, int32_t carrier_id, const char* now,
            std::vector<DeliveryOrderInfo>* orders, TPCCUndo** undo) {
        assert(undo == NULL);
        serializeDelivery(warehouse_id, carrier_id, now, &request_);
        bool committed = makeRequest(warehouse_id);
        ASSERT(committed);
        deserializeDeliveryOutput(result_, orders);
    }

    virtual bool hasWarehouse(int32_t warehouse_id) { CHECK(false); }
    virtual void applyUndo(TPCCUndo* undo) { CHECK(false); }
    virtual void freeUndo(TPCCUndo* undo) { CHECK(false); }
    virtual void setTransaction(dtxn::Transaction* txn) { CHECK(false); }

private:
    void internalPayment(int32_t warehouse_id, int32_t district_id, int32_t c_warehouse_id,
            int32_t c_district_id, int32_t customer_id, const char* c_last, float h_amount,
            const char* now, PaymentOutput* output) {
        serializePayment(warehouse_id, district_id, c_warehouse_id, c_district_id,
                customer_id, c_last, h_amount, now, &request_);
        int id = warehouse_id;
        if (partition(warehouse_id) != partition(c_warehouse_id)) id = -1;
        bool committed = makeRequest(id);
        ASSERT(committed);
        TPCCEngine::deserializePaymentOutput(result_, output);
    }

    // The partition function assumes that warehouses are evenly distributed
    int partition(int32_t warehouse_id) {
        assert(warehouse_id <= num_warehouses_);
        int warehouses_per_partition = num_warehouses_ / static_cast<int>(clients_.size());
        int client_index = (warehouse_id - 1) / warehouses_per_partition;
        return client_index;
    }

    bool makeRequest(int32_t warehouse_id) {
        bool committed = false;
        result_.clear();
        while (!committed && result_.empty()) {
            // if this aborts without data, assume it is a deadlock abort and retry. When new order
            // user aborts it returns data.
            if (warehouse_id < 1) {
                // This is a distributor request
                committed = distributor_->blockingExecute(&eventloop_, request_, &result_);
            } else {
                // This is a direct single partition request
                int client = partition(warehouse_id);
                committed = clients_[client]->blockingExecute(&eventloop_, request_, &result_);
            }
        }
        request_.clear();
        return committed;
    }

    int num_warehouses_;

    string request_;
    string result_;

    io::LibEventLoop eventloop_;
    dtxn::DtxnClient* distributor_;
    vector<dtxn::DtxnClient*> clients_;
};

class TPCCPerfTest : public benchmark::BenchmarkWorker {
public:
    // Owns client
    TPCCPerfTest(TPCCClient* client) : client_(client), all_new_order_(false) {
        assert(client_ != NULL);
    }

    virtual ~TPCCPerfTest() {
        delete client_;
    }

    virtual void initialize() {}

    virtual void work(bool measure) {
        if (all_new_order_) {
            client_->doNewOrder();
        } else {
            client_->doOne();
        }
    }

    void all_new_order(bool value) { all_new_order_ = value; }

private:
    TPCCClient* client_;
    bool all_new_order_;
};

static const int WARM_UP = 15;
static const int MEASURE = 60;

int main(int argc, const char* argv[]) {
    if (argc != 5 && argc != 6) {
        fprintf(stderr, "tpccclientperf [distributor] [config file] [warehouses] [client threads] [p new order item remote (0-1000)]\n");
        return 1;
    }

    // Parse the distributor's address
    NetworkAddress distributor;
    bool success = distributor.parse(argv[1]);
    ASSERT(success);

    // Read and parse the configuration file.
    vector<dtxn::Partition> partitions = dtxn::parseConfigurationFromPath(argv[2]);

    int num_warehouses = atoi(argv[3]);
    assert(num_warehouses > 0);

    int threads = atoi(argv[4]);
    assert(threads > 0);
    int remote_item_milli_p = -1;
    if (argc == 6) {
        remote_item_milli_p = atoi(argv[5]);
        assert(0 <= remote_item_milli_p && remote_item_milli_p <= 1000);
    }

    // Generate the same cLoad value as on the servers so we can generate an appropriate cRun value
    tpcc::RealRandomGenerator tempRandom;
    tempRandom.seed(0);
    const tpcc::NURandC cLoad = tpcc::NURandC::makeRandom(&tempRandom);
    tempRandom.setC(cLoad);
    //~ printf("cLoad = %d %d %d\n", cLoad.c_last_, cLoad.c_id_, cLoad.ol_i_id_);

    // All clients must use the same cRun value
    int seed = static_cast<int>(time(NULL));
    tempRandom.seed(seed);
    const tpcc::NURandC cRun = tpcc::NURandC::makeRandomForRun(&tempRandom, cLoad);

    // TODO: Add a command line flag for "bound clients"?
    assert(threads >= num_warehouses);
    double threads_per_warehouse = (double) threads / num_warehouses;
    assert(threads_per_warehouse <= District::NUM_PER_WAREHOUSE);
    //~ if (threads_per_warehouse < District::NUM_PER_WAREHOUSE) {
        //~ printf("WARNING: Binding < %d clients per warehouse; expect delivery warnings\n", District::NUM_PER_WAREHOUSE);
    //~ }

    // Create threads with the partitions
    vector<TPCCPerfTest*> testers;
    for (int i = 0; i < threads; ++i) {
        // Generate appropriate constants for running the test
        tpcc::RealRandomGenerator* rng = new tpcc::RealRandomGenerator();
        rng->seed(seed + i);
        rng->setC(cRun);

        TPCCClientDb* client_db = new TPCCClientDb(num_warehouses, distributor, partitions);
        TPCCClient* client = new TPCCClient(new SystemClock(), rng, client_db,
                Item::NUM_ITEMS, num_warehouses, District::NUM_PER_WAREHOUSE,
                Customer::NUM_PER_DISTRICT);

        // TODO: Command line flag?
        int warehouse = (int) (i / threads_per_warehouse) + 1;
        //~ int district = i - (int) ((warehouse - 1) * threads_per_warehouse) + 1;
        int district = 0;
        client->bindWarehouseDistrict(warehouse, district);

        testers.push_back(new TPCCPerfTest(client));
        if (remote_item_milli_p != -1) {
            client->remote_item_milli_p(remote_item_milli_p);
            testers.back()->all_new_order(true);
        }
    }

    int total_transactions = 0;
    int64_t us = benchmark::runThreadBench(WARM_UP, MEASURE, testers, &total_transactions);
    //~ printf("%d %lld us\n", sum, us);
    double txns_per_s = total_transactions / static_cast<double>(us) * 1000000.0;
    printf("%f\n", txns_per_s);

    STLDeleteElements(&testers);

    // Finally, reset the server
    io::LibEventLoop eventloop;
    dtxn::DtxnClient* client = dtxn::DtxnClient::create(&eventloop, distributor);
    string reset_message;
    TPCCEngine::serializeReset(&reset_message);
    string output;
    // NOTE: This will typically abort because it is really slow. But this doesn't cause problems.
    // However, it is best to disable timeouts in the distributor so the client doesn't exit until
    // this completes
    client->blockingExecute(&eventloop, reset_message, &output);

    return 0;
}
