#!/usr/bin/python
"""
Parses buildrules files for StupidBuilder.
"""

import os.path

def buildDirname(path):
    """Returns os.path.dirname with a trailing "/" appended."""
    out = os.path.dirname(path)
    if out != "/":
        out += "/"
    return out


class Error(Exception):
    pass


class BuildParser(object):
    """Parses buildrules files and loads all the rules."""

    def __init__(self):
        """Creates an empty parser."""
        
        # Stores the build root directory. All absolute paths in buildrules are relative to this.
        self.root = None
        self.rootScope = {}
        # Maps absolute directory names to the parsed rules in that directory
        self.parsedDirectories = {}
        # List of parsed rules
        self.rules = []

    def loadRules(self, directory):
        """Loads the buildRules from directory.
        
        Args:
            directory: the absolute path to the rules file."""
    
        if not directory.startswith("/"):
            raise ValueError("directory must be absolute (was: %s)" % directory)

        if self.root is None:
            self.findRoot(directory)
            assert self.root[-1] != "/"

        # directory must be under the build root for this parser
        if not directory.startswith(self.root):
            raise ValueError("directory must be under the build root (%s)" % self.root)
        buildDirectory = directory[len(self.root):]
        if len(buildDirectory) == 0 or buildDirectory[-1] != "/":
            buildDirectory += "/"
        
        self.parsedDirectories = {}
        self.parseRulesInDirectory(buildDirectory)

    def parseRulesInDirectory(self, buildDirectory):
        """Loads the rules from buildDirectory.
        
        Args:
            buildDirectory: A build path to the directory."""

        assert self.root is not None
        assert buildDirectory[0] == "/"
        assert buildDirectory[-1] == "/"

        # If we have already parsed the rules in this directory, don't do it again
        if buildDirectory in self.parsedDirectories:
            return
        # Mark that we have/will loaded rules for this directory: this is
        # needed to handle cyclic directory dependencies
        # TODO: Cache the per-directory rules to enable reloading?
        self.parsedDirectories[buildDirectory] = True

        # Parse the rules out of this directory
        symbols = self.parse(self.root + buildDirectory)
        if "rules" not in symbols:
            if buildDirectory == "/":
                # No rules is permitted in the root
                symbols["rules"] = []
            else:
                raise Error("buildrules contains no rules in directory: " + buildDirectory)

        for rawRule in symbols["rules"]:
            # Validate the inputs and transform them
            inputs = []
            for input in rawRule[0]:
                if input[0] == "/":
                    # Absolute path: verify it doesn't point to this directory
                    referencedDirectory = buildDirname(input)
                    if referencedDirectory == buildDirectory:
                        raise Error("relative paths must be used for the current directory (bad input: %s)" % input)
                    inputs.append(input)

                    # Parse the rules in the referenced directory as well, since we will need them
                    self.parseRulesInDirectory(referencedDirectory)
                else:
                    if "/" in input:
                        raise Error("inputs must be in the current directory or absolute (bad input: %s)" % input)
                    inputs.append(buildDirectory + input)

            # Validate the outputs and transform them
            outputs = []
            for output in rawRule[1]:
                if "/" in output:
                    raise Error("outputs must be in the current directory (bad output: %s)" % output)
                fullOutput = buildDirectory + output
                if os.path.exists(self.root + fullOutput):
                    raise Error("output exists in source tree: " + fullOutput)
                outputs.append(fullOutput)

            # Load any options
            options = dict()
            if len(rawRule) == 5:
                options = dict(rawRule[4])
            elif len(rawRule) > 5:
                raise Error("too many parameters for rule:", rawRule)

            # Store the transformed rule
            self.rules.append((tuple(inputs), tuple(outputs), rawRule[2], rawRule[3], options))

        # Look for subdirectories
        for subentry in os.listdir(self.root + buildDirectory):
            subdir = buildDirectory + subentry
            if os.path.isdir(self.root + subdir) and os.path.exists(
                    self.root + subdir + "/buildrules"):
                self.parseRulesInDirectory(subdir + "/")


    def findRoot(self, directory):
        """Look for a directory containing a buildrules file with ROOT=True."""

        assert self.root is None

        while len(directory) > 0:
            try:
                symbols = self.parse(directory + "/")
            except NameError:
                # Ignore: NameErrors should be fixed once we have a root scope.
                symbols = {}
            if "ROOT" in symbols and symbols["ROOT"] is True:
                # Copy the symbols into the root scope, plus and minus a few items
                self.root = directory
                for key, value in symbols.iteritems():
                    self.rootScope[key] = value
                if "rules" in self.rootScope: del self.rootScope["rules"]
                self.rootScope["SOURCEROOT"] = self.root
                return

            directory = os.path.dirname(directory)


    def parse(self, directory):
        fileobj = open(directory + "buildrules")
        rules = fileobj.read()
        fileobj.close()

        moduleScope = {}
        exec rules in self.rootScope, moduleScope
        return moduleScope


def parseBuildRules(rules, builder):
    """Fills builder with the parsed build rules by calling buildTarget for each target found."""

    scope = {}
    exec rules in {}, scope

    for arguments in scope["rules"]:
        if len(arguments) != 3 and len(arguments) != 4:
            raise ValueError("Bad number of arguments for rule. Need 3 or 4.")
        builder.buildTarget(*arguments)
