#!/usr/bin/python

import os
import shutil
import tempfile
import unittest

import buildparser


class MockBuilder(object):
    def __init__(self):
        self.inputs = []
        self.outputs = []
        self.buildRules = []
        self.dependencyRules = []

    def buildTarget(self, inputs, outputs, buildRule, dependencyRule):
        # Check that inputs and outputs are lists or tuples
        assert len(inputs) >= 0
        assert len(outputs) >= 0

        # buildRule and dependencyRule must be callable
        assert callable(buildRule)
        assert dependencyRule is None or callable(dependencyRule)

        self.inputs.append(inputs)
        self.outputs.append(outputs)
        self.buildRules.append(buildRule)
        self.dependencyRules.append(dependencyRule)


class ParseBuildRulesTest(unittest.TestCase):
    def testParser(self):
        data = """
def FooBuild(inputs, outputs): pass

rules = [(["foo.cc"], ["foo.o"], FooBuild, None)]"""

        builder = MockBuilder()
        buildparser.parseBuildRules(data, builder)
        assert len(builder.inputs) == 1
        assert builder.inputs[0] == ["foo.cc"]
        assert len(builder.outputs) == 1
        assert builder.outputs[0] == ["foo.o"]
        assert len(builder.buildRules) == 1
        assert len(builder.dependencyRules) == 1
        assert builder.dependencyRules[0] is None


    def testDependencyRule(self):
        data = """
def FooBuild(inputs, outputs): pass

def FooDependency(inputs): return inputs + ("foo",)

rules = [(["foo.cc"], ["foo.o"], FooBuild, FooDependency)]"""

        builder = MockBuilder()
        buildparser.parseBuildRules(data, builder)
        assert len(builder.dependencyRules) == 1
        assert ("foo.cc", "foo") == builder.dependencyRules[0](("foo.cc",))


class TemporaryDirTest(unittest.TestCase):
    """Creates a temporary directory as the current directory in setUp and deletes it in tearDown."""

    def setUp(self):
        # Create a temporary directory in which to create files
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)


class BuildParserTest(TemporaryDirTest):
    def writeBuildRules(self, directory, contents):
        rulesFile = open(directory + "/buildrules", "w")
        rulesFile.write(contents)
        rulesFile.close()
        
    def testLoadRulesErrors(self):
        parser = buildparser.BuildParser()
        # Directories must be absolute
        self.assertRaises(ValueError, parser.loadRules, ".")

        # No buildrules
        self.assertRaises(IOError, parser.loadRules, self.tempdir)

        # A buildrules, but no root
        subdir = self.tempdir + "/subdir"
        os.mkdir(subdir)
        self.writeBuildRules(subdir, "")
        self.assertRaises(IOError, parser.loadRules, subdir)

        # More than one buildrules
        self.writeBuildRules(self.tempdir, "")
        self.assertRaises(IOError, parser.loadRules, subdir)

    def testLoadRulesNoRules(self):
        parser = buildparser.BuildParser()
        # Root with no rules: okay
        self.writeBuildRules(self.tempdir, "ROOT=True\n")
        parser.loadRules(self.tempdir)

        # Subdirectory with no rules: not okay
        one = self.tempdir + "/one"
        os.mkdir(one)
        self.writeBuildRules(one, "")
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

    def testLoadRulesBadInputPath(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["relative/foo.cc"], ["foo.o"], FooBuild, None)]""")

        parser = buildparser.BuildParser()
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

        # Can't use an absolute path to our own directory: only to others
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["/foo.cc"], ["foo.o"], FooBuild, None)]""")
        parser = buildparser.BuildParser()
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

    def testLoadRulesBadOutputPath(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["foo.cc"], ["relative/foo.o"], FooBuild, None)]""")

        parser = buildparser.BuildParser()
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["foo.cc"], ["/absolute/foo.o"], FooBuild, None)]""")

        parser = buildparser.BuildParser()
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

    def testLoadRulesOutputExists(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["foo.cc"], ["foo.o"], FooBuild, None)]""")
        # The output file exists in the source tree: not permitted
        open(self.tempdir + "/foo.o", "w").close()

        parser = buildparser.BuildParser()
        self.assertRaises(buildparser.Error, parser.loadRules, self.tempdir)

    def testLoadRulesSimple(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(["foo.cc"], ["foo.o"], FooBuild, None)]""")

        parser = buildparser.BuildParser()
        parser.loadRules(self.tempdir)
        assert parser.root == self.tempdir
        
        # Check that it actually loaded a rule
        assert len(parser.rules) == 1
        self.assertEquals(("/foo.cc",), parser.rules[0][0])
        self.assertEquals(("/foo.o",), parser.rules[0][1])

        # Parsing a directory that is NOT under the build root is an error
        self.assertRaises(ValueError, parser.loadRules, "/usr")

    def testLoadRulesSetsSourceRoot(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): return SOURCEROOT

rules = [(["foo.cc"], ["foo.o"], FooBuild, None)]""")

        parser = buildparser.BuildParser()
        parser.loadRules(self.tempdir)
        # Calling the build rule returns the SOURCEROOT variable
        assert parser.rules[0][2]([], []) == self.tempdir

    def testLoadRulesAcrossDirs(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass""")

        one = self.tempdir + "/one"
        os.mkdir(one)
        self.writeBuildRules(one, "rules = [(['foo.cc'], ['foo.o'], FooBuild, None)]")

        parser = buildparser.BuildParser()
        parser.loadRules(one)
        assert parser.root == self.tempdir
        assert len(parser.rules) == 1
        self.assertEquals(("/one/foo.cc",), parser.rules[0][0])

        two = self.tempdir + "/two"
        os.mkdir(two)
        self.writeBuildRules(two, """
rules = [
    (['/one/foo.o'], ['foo'], FooBuild, None),
    (['/one/foo.o'], ['foo2'], FooBuild, None),
]""")

        parser = buildparser.BuildParser()
        parser.loadRules(two)

        # Extract the output files from the rules
        targets = []
        for rule in parser.rules:
            for output in rule[1]:
                targets.append(os.path.basename(output))
        self.assertEquals(3, len(targets))
        assert "foo" in targets
        assert "foo.o" in targets
        assert "foo2" in targets

    def testLoadRulesFromRoot(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass
rules = [(['foo.cc'], ['foo.o'], FooBuild, None)]
""")

        one = self.tempdir + "/one"
        os.mkdir(one)
        self.writeBuildRules(one, "rules = [(['bar.o', '/foo.o'], ['bar'], FooBuild, None)]")

        parser = buildparser.BuildParser()
        parser.loadRules(one)
        self.assertEquals(self.tempdir, parser.root)

        # Extract the output files from the rules
        targets = []
        for rule in parser.rules:
            for output in rule[1]:
                targets.append(os.path.basename(output))
        self.assertEquals(2, len(targets))
        assert "bar" in targets
        assert "foo.o" in targets

    def testLoadCircularReferencedDirs(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass""")

        one = self.tempdir + "/one"
        os.mkdir(one)
        self.writeBuildRules(one, """
rules = [
    (['/two/foo.o'], ['baz.a'], FooBuild, None),
    (['bar.cc'], ['bar.o'], FooBuild, None),
]""")

        two = self.tempdir + "/two"
        os.mkdir(two)
        self.writeBuildRules(two, """
rules = [
    (['foo.cc'], ['foo.o'], FooBuild, None),
    (['/one/bar.o'], ['baz2.a'], FooBuild, None),
]""")

        parser = buildparser.BuildParser()
        parser.loadRules(two)

        # Extract the output files from the rules
        targets = []
        for rule in parser.rules:
            for output in rule[1]:
                targets.append(os.path.basename(output))
        self.assertEquals(4, len(targets))
        assert "baz.a" in targets
        assert "bar.o" in targets
        assert "foo.o" in targets
        assert "baz2.a" in targets

    def testLoadFromSubDirs(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass""")

        one = self.tempdir + "/one"
        os.mkdir(one)
        self.writeBuildRules(one, """
rules = [
    (['bar.cc'], ['bar.o'], FooBuild, None),
]""")

        two = self.tempdir + "/two"
        os.mkdir(two)
        self.writeBuildRules(two, """
rules = [
    (['foo.cc'], ['foo.o'], FooBuild, None),
]""")

        three = self.tempdir + "/three"
        os.mkdir(three)

        parser = buildparser.BuildParser()
        parser.loadRules(self.tempdir)

        # Extract the output files from the rules
        targets = []
        for rule in parser.rules:
            for output in rule[1]:
                targets.append(os.path.basename(output))
        self.assertEquals(2, len(targets))
        assert "foo.o" in targets
        assert "bar.o" in targets

    def testLoadOptions(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs, **options): pass

rules = [
    (['foo.cc'], ['foo.o'], FooBuild, None, {'key': 'value'}),
    (['bar.cc'], ['bar.o'], FooBuild, None),
]
""")

        parser = buildparser.BuildParser()
        parser.loadRules(self.tempdir)

        self.assertEquals(2, len(parser.rules))
        self.assertEquals(("/foo.cc",), parser.rules[0][0])
        self.assertEquals(("/foo.o",), parser.rules[0][1])
        self.assertEquals({'key': 'value'}, parser.rules[0][4])

        self.assertEquals(("/bar.cc",), parser.rules[1][0])
        self.assertEquals(("/bar.o",), parser.rules[1][1])
        self.assertEquals({}, parser.rules[1][4])

    def testRuleTooManyParams(self):
        self.writeBuildRules(self.tempdir, """ROOT=True
def FooBuild(inputs, outputs): pass

rules = [(['foo.cc'], ['foo.o'], FooBuild, None, {'key': 'value'}, 42)]
""")

        parser = buildparser.BuildParser()
        try:
            parser.loadRules(self.tempdir)
            self.fail("Expected Error")
        except buildparser.Error, e:
            pass


if __name__ == "__main__":
    unittest.main()
