#!/usr/bin/python
"""
Model: outputs <-- build rule <-- dependency rule <-- inputs

Operations:
1. Build a specific target from scratch.
2. Build all targets from scratch.
3. Clean intermediate output.
4. Precise rebuild of a specific target.

Note: The dependency rule before build rule is not necissarily the most
efficient. Specifically in the case of GCC, the dependency information can be
generated for effectively no cost at the same time as compilation. This saves
~200ms per compilation (dependending on the file, of course). However, I think
the dependency before build system is more generic, and is certainly simple.
"""

import os
import sys

import buildparser


class DuplicateOutputError(Exception):
    """Indicates that an output file is specified as being created by two rules."""
    pass

class SpecialTargetError(Exception):
    """Indicates that an output name is a reserved target name."""
    pass

class InvalidBuildPathError(Exception):
    """Indicates that a build path is not valid."""
    pass

class UnknownInputError(Exception):
    """Indicates that a required input is missing and can't be built."""
    pass


def validateBuildPath(buildPath):
    """Raises an InvalidBuildPath exception if buildPath is not valid."""

    if buildPath[0] != "/":
        raise InvalidBuildPathError("build paths must be absolute (to the build root): (%s)" % buildPath)
    if buildPath[-1] == "/":
        raise InvalidBuildPathError("build paths cannot be directories: (%s)" % buildPath)


class BuildNode(object):
    """A command that takes inputs and produces outputs. It gets dependencies
    from the dependencyRule. If the depedency rule is None, it uses the identity rule,
    which simply uses the list of inputs."""

    def __init__(self, inputs, outputs, buildRule, dependencyRule, options):
        if len(inputs) == 0:
            raise ValueError, "Must specify at least 1 input"
        for input in inputs: validateBuildPath(input)

        if len(outputs) == 0:
            raise ValueError, "Must specify at least 1 output"
        for output in outputs: validateBuildPath(output)
        
        if not callable(buildRule):
            raise TypeError, "buildRule must be callable"

        if dependencyRule is not None and not callable(dependencyRule):
            raise TypeError, "dependencyRule must be None or callable"
        
        self.inputs = tuple(inputs)
        self.outputs = tuple(outputs)
        self.buildRule = buildRule
        self.dependencyRule = dependencyRule
        self.options = options

    def __str__(self):
        return str(list(self.inputs)) + " -> " + str(list(self.outputs)) 

    def dependencies(self, builder):
        """"Returns a list of the dependencies for this rule."""

        if self.dependencyRule is None:
            return self.inputs

        # TODO: cache the dependency list
        inputs = tuple(builder.inputPath(input) for input in self.inputs)
        rawDependencies = self.dependencyRule(inputs)

        # Translate file-based dependencies back into targets
        dependencies = []
        for dependency in rawDependencies:
            # TODO: This is a gross hack. Is there a better way to map from the file
            # back to the "rule" path?
            outputPath = builder.buildRoot + "/build"
            if dependency.startswith(outputPath):
                dependency = dependency[len(outputPath):]
            else:
                assert dependency.startswith(builder.buildRoot)
                dependency = dependency[len(builder.buildRoot):]
            dependencies.append(dependency)
        return dependencies

    def needsBuild(self, builder):
        """Returns (needsBuild, pendingDependencies). needsBuild is True if this
        node needs rebuilding. pendingDependencies is a list of
        dependencies that must be built first."""

        needsBuild = False

        # Find the oldest timestamp of the outputs
        oldestOutputTime = None
        for output in self.outputs:
            outputPath = builder.outputPath(output)
            if not os.path.exists(outputPath):
                # The output does not exist: we need to build it
                needsBuild = True
            else:
                outputTime = os.path.getmtime(outputPath)
                if oldestOutputTime == None or outputTime < oldestOutputTime:
                    oldestOutputTime = outputTime

        # grab the cached dependencies, if possible
        dependencies = builder.getDependencies(self.outputs[0])
        dependenciesRefreshed = False
        if dependencies is None:
            # No cached dependencies: compute them and update them
            dependencies = self.dependencies(builder)
            dependenciesRefreshed = True

        # Check all the dependencies to count how many need to get built
        pendingDependencies = []
        for dependency in dependencies:
            if builder.needsBuild(dependency):
                # A dependency needs rebuilding
                needsBuild = True
                pendingDependencies.append(dependency)
            else:
                # Check if an input has been updated
                dependencyPath = builder.inputPath(dependency)
                inputTime = os.path.getmtime(dependencyPath)
                if inputTime >= oldestOutputTime:
                    needsBuild = True

        # Update the cached dependencies if needed
        if needsBuild or dependenciesRefreshed:
            if not dependenciesRefreshed:
                # Ensure up to date dependencies on build
                dependencies = self.dependencies(builder)
            builder.setDependencies(self.outputs[0], dependencies)

        return (needsBuild, pendingDependencies)

    def buildCommand(self, builder):
        """Returns the command used to run this build rule."""
        
        inputs = tuple(builder.inputPath(input) for input in self.inputs)
        outputs = tuple(builder.outputPath(output) for output in self.outputs)
        return self.buildRule(inputs, outputs, self.options)


class StupidBuilder(object):
    """Contains a directed acyclic graph of the build rules.
    
    This chains rules together where needed."""

    def __init__(self, buildRoot):
        self.buildRoot = buildRoot
        assert self.buildRoot[0] == "/"

        # Maps outputs to build rules
        self.outputs = {}
        # Maps targets to (needsBuild, pendingDependencyCount)
        self.outputBuildState = {}
        # Maps targets to targets that need them as inputs
        self.outputDependencies = {}
        # Maps the first output of a build rule to the cached dependencies
        self.dependencyCache = {}

    def inputPath(self, buildInput):
        validateBuildPath(buildInput)
        if buildInput in self.outputs:
            # This input is also an output: find it in the outputs dir
            return self.outputPath(buildInput)

        # Return the source directory for this file
        return self.buildRoot + buildInput

    def outputPath(self, buildOutput):
        validateBuildPath(buildOutput)
        return self.buildRoot + "/build" + buildOutput

    def targetNode(self, target):
        """Returns the BuildNode for target, if it exists."""

        return self.outputs[target]


    def needsBuild(self, target):
        """Returns True if target needs to be built. This updates the internal build state."""

        # Check for the cached value first
        if target not in self.outputBuildState:
            if target in self.outputs:
                # This node is an output: delegate to the build rule.
                needsBuild, pendingDependencies = self.outputs[target].needsBuild(self)
                self.outputBuildState[target] = [needsBuild, len(pendingDependencies)]

                for dependency in pendingDependencies:
                    # Build a mapping from dependency to rule that needs it
                    if dependency not in self.outputDependencies:
                        self.outputDependencies[dependency] = []
                    self.outputDependencies[dependency].append(target)

            elif not os.path.exists(self.inputPath(target)):
                # An input does not exist and we don't know how to build it
                raise UnknownInputError("Source file " + target + " does not exist and can't be built")
            else:
                # This is a "terminal" source file: it never needs to be built
                self.outputBuildState[target] = (False, 0)

        return self.outputBuildState[target][0]

    def buildTarget(self, inputs, outputs, buildRule, dependencyRule, options):
        # Create the build node
        node = BuildNode(inputs, outputs, buildRule, dependencyRule, options)
        
        # Register it
        for output in outputs:
            if output in self.outputs:
                raise DuplicateOutputError(output + " has already been specified as an output by rule: " + str(self.outputs[output]))
            self.outputs[output] = node
        return node

    def commands(self, target):
        """Returns all build commands in this builder."""

        # Reset the build state
        self.outputBuildState = {}
        self.outputDependencies = {}

        # Build the dependency graph by calling self.needsBuild
        if target is None:
            # Process ALL outputs
            for output, rule in self.outputs.iteritems():
                self.needsBuild(output)
        else:
            # Call self.needsBuild on the one output to prime the dependency graph
            self.needsBuild(target)

        commands = []

        # Step 1: Collect a seed list of targets that are ready to be built
        total = len(self.outputBuildState)
        count = 0
        deleteKeys = []
        for target, (needsBuild, pendingDependencies) in self.outputBuildState.iteritems():
            # If the output directory does not exist, create it.
            # TODO: Does this make sense here? Probaby it should be part of a "command runner"
            outputDirectory = os.path.dirname(self.outputPath(target))
            if not os.path.exists(outputDirectory):
                os.makedirs(outputDirectory)
            
            count += 1
            if needsBuild and pendingDependencies == 0:
                # Ready for building: Add the command to the list
                commands.append((target, self.outputs[target].buildCommand(self)))
        
        # TODO: Remove this
        assert count == total

        # Step 2: Simulate building each target in order until they are all done
        for target, command in commands:
            # The target has been built: update anything that depends on it
            if target in self.outputDependencies:
                for outputDependency in self.outputDependencies[target]:
                    self.outputBuildState[outputDependency][1] -= 1
                    
                    if self.outputBuildState[outputDependency][1] == 0:
                        # The outputDependency now has all its dependencies: build it
                        node = self.outputs[outputDependency]
                        commands.append((outputDependency, node.buildCommand(self)))

        return [command for target, command in commands]

    def cleanFiles(self):
        """Returns an iterator over all the output files."""
        return (self.outputPath(output) for output in self.outputs.iterkeys())

    def getDependencies(self, output):
        """Returns the dependency cache for output, if they exist."""
        return self.dependencyCache.get(output)

    def setDependencies(self, output, dependencies):
        """Update the dependency cache for output."""
        self.dependencyCache[output] = dependencies

if __name__ == "__main__":
    if len(sys.argv) > 2:
        sys.stderr.write('stupidbuild: Too many arguments\n')
        sys.stderr.write('usage: stupidbuild [target to build]\n')
        sys.exit(1)

    target = None
    if len(sys.argv) == 2:
        target = sys.argv[1]

    # Get the current dir then change to root to avoid bugs with paths
    currentDirectory = os.getcwd()
    os.chdir("/")
    
    # Parse the rules for the current directory
    parser = buildparser.BuildParser()
    parser.loadRules(currentDirectory)

    # Parse the rules for the target
    #~ if target is not None:
        #~ if target.startswith("/"):
            #~ targetDir = os.path.join(parser.root, target[1:])
        #~ else:
            #~ targetDir = os.path.join(currentDirectory, target)
        #~ targetDir = os.path.dirname(targetDir)
        #~ parser.loadRules(targetDir)

    # Feed the rules into the build graph
    builder = StupidBuilder(parser.root)
    for rule in parser.rules:
        builder.buildTarget(*rule)

    # Look for cached dependency data
    buildCache = parser.root + "/.stupidbuild_cache"
    if os.path.exists(buildCache):
        f = open(buildCache)
        dependenciesData = f.read()
        f.close()

        # Parse the cached dependency info and load it
        dependencies = eval(dependenciesData)
        builder.dependencyCache = dependencies

    if target == 'clean':
        for filename in builder.cleanFiles():
            if os.path.exists(filename):
                print 'deleting', filename
                os.unlink(filename)
        # TODO: Add this to the clientFiles() list from the builder?
        if os.path.exists(buildCache):
            os.unlink(buildCache)
    else:
        try:
            commands = builder.commands(target)
        except UnknownInputError, e:
            sys.stderr.write('stupidbuild: ' + str(e) + '\n')
            sys.exit(2)
        
        for command in commands:
            print command
            code = os.system(command)
            assert code == 0

    # Update the dependency cache for successful builds
    # TODO: Integrate loading/saving the cache with the StupidBuilder class.
    f = open(buildCache, "w")
    f.write(repr(builder.dependencyCache))
    f.close()

    sys.exit(0)

