#!/usr/bin/python

import tempfile
import time
import os
import shutil
import unittest

import stupidbuild


def NullBuildRule(inputs, outputs, options):
    return outputs[0]


class MockBuilder(object):
    def __init__(self, needsBuildOutput, tempdir):
        self.needsBuildOutput = needsBuildOutput
        self.buildRoot = tempdir
        self.dependencies = None
    
    def needsBuild(self, target):
        if target in self.needsBuildOutput:
            return self.needsBuildOutput[target]
        else:
            return False

    def inputPath(self, path):
        return self.buildRoot + path

    def outputPath(self, path):
        return self.buildRoot + path

    def getDependencies(self, output):
        return self.dependencies

    def setDependencies(self, output, dependencies):
        self.dependencies = dependencies

class TemporaryDirTest(unittest.TestCase):
    """Creates a temporary directory as the current directory in setUp and deletes it in tearDown."""

    def setUp(self):
        # Create a temporary directory in which to create files
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        os.chdir("..")
        shutil.rmtree(self.tempdir)

    def touch(self, filename, timestamp=None):
        fullPath = self.tempdir + filename
        if not os.path.exists(fullPath):
            open(fullPath, 'w').close()

        if timestamp is not None:
            os.utime(fullPath, (timestamp, timestamp))


class BuildNodeTest(TemporaryDirTest):
    def setUp(self):
        TemporaryDirTest.setUp(self)

        self.inputs = ["/foo.c"]
        self.outputs = ["/foo.o"]
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, None, {})
        self.now = time.time()
        self.mockBuilder = MockBuilder({}, self.tempdir)
        self.dependencyRuleCalled = False

    def testConstructor(self):
        # inputs needs to have an item
        self.assertRaises(ValueError, stupidbuild.BuildNode, [], self.outputs, NullBuildRule, None, {})
        # outputs needs to have an item
        self.assertRaises(ValueError, stupidbuild.BuildNode, self.inputs, [], NullBuildRule, None, {})
        # buildRule must be callable
        self.assertRaises(TypeError, stupidbuild.BuildNode, self.inputs, self.outputs, 5, None, {})
        # dependencyRule must be callable
        self.assertRaises(TypeError, stupidbuild.BuildNode, self.inputs, self.outputs, NullBuildRule, 5, {})

        # Check that buildTarget raises an exception for bad paths
        self.assertRaises(stupidbuild.InvalidBuildPathError,
            stupidbuild.BuildNode, ["bar.c"], ["/bar.o"], NullBuildRule, None, {})
        self.assertRaises(stupidbuild.InvalidBuildPathError,
            stupidbuild.BuildNode, ["/bar.c"], ["bar.o"], NullBuildRule, None, {})
        self.assertRaises(stupidbuild.InvalidBuildPathError,
            stupidbuild.BuildNode, ["/bar.c/"], ["bar.o"], NullBuildRule, None, {})
        self.assertRaises(stupidbuild.InvalidBuildPathError,
            stupidbuild.BuildNode, ["/bar.c"], ["/bar.o/"], NullBuildRule, None, {})

    def testSimple(self):
        assert len(self.node.inputs) == 1
        assert "/foo.c" in self.node.inputs
        assert len(self.node.outputs) == 1
        assert "/foo.o" in self.node.outputs
        assert self.node.buildRule == NullBuildRule
        
        # Verify that inputs and outputs get copied
        self.inputs.append("bar")
        assert len(self.node.inputs) == 1
        self.outputs.append("bar")
        assert len(self.node.outputs) == 1

    def testStr(self):
        self.assertEquals("['/foo.c'] -> ['/foo.o']", str(self.node))

    def _dependencyRule(self, inputs):
        self.dependencyRuleCalled = True
        # Inputs come in in the form of absolute paths into the build directory
        for input in inputs:
            assert input.startswith(self.tempdir)
        return inputs + (self.tempdir + "/foo",)

    def testDependencies(self):
        node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})

        dependencies = node.dependencies(self.mockBuilder)
        assert len(dependencies) == 2
        assert self.inputs[0] in dependencies
        assert "/foo" in dependencies

    def _buildRule(self, inputs, outputs, options):
        self._buildInputs = inputs
        self._buildOutputs = outputs
        self._buildOptions = options
        return "foo"

    def testBuildCommand(self):
        node = stupidbuild.BuildNode(self.inputs, self.outputs, self._buildRule, None, {})
        self.assertEquals("foo", node.buildCommand(self.mockBuilder))
       
        assert tuple(self.tempdir + input for input in self.inputs) == self._buildInputs
        assert tuple(self.tempdir + output for output in self.outputs) == self._buildOutputs

    def testBuildWithOptions(self):
        options = {'option': True}
        node = stupidbuild.BuildNode(
                self.inputs, self.outputs, self._buildRule, None, options)
        self.assertEquals("foo", node.buildCommand(self.mockBuilder))
        self.assertEquals(options, self._buildOptions)

    def testNeedsBuildNoInput(self):
        # Raises OSError(2): No such file "foo.c"
        self.assertRaises(OSError, self.node.needsBuild, self.mockBuilder)

    def testNeedsBuildNoOutput(self):
        # This node need to be built because there is no output
        self.touch(self.inputs[0])
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        assert len(pendingDependencies) == 0

    def testNeedsBuildOldOutput(self):
        # The output is the same age or older: needs to be built
        self.touch(self.inputs[0], self.now)
        self.touch(self.outputs[0], self.now)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        assert len(pendingDependencies) == 0

        self.touch(self.inputs[0], self.now+1)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        assert len(pendingDependencies) == 0

    def testNeedsBuildUpToDate(self):
        # The output is newer: does not need to be built
        self.touch(self.inputs[0], self.now)
        self.touch(self.outputs[0], self.now+1)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert not needsBuild
        assert len(pendingDependencies) == 0

    def testNeedsBuildMissingDependency(self):
        # The input exists, but the dependency does not
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now)
        self.assertRaises(OSError, self.node.needsBuild, self.mockBuilder)

    def testNeedsBuildDependencyNewer(self):
        # The dependency is the same age or newer: needs build
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now-1)
        self.touch(self.outputs[0], self.now)
        self.touch("/foo", self.now)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        assert len(pendingDependencies) == 0

        self.touch("/foo", self.now+1)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        assert len(pendingDependencies) == 0

    def testNeedsBuildDependencyUpToDate(self):
        # The dependency is the same age or newer: needs build
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now-1)
        self.touch(self.outputs[0], self.now)
        self.touch("/foo", self.now-1)
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert not needsBuild
        assert len(pendingDependencies) == 0

    def testNeedsBuildCachedDependencies(self):
        # The output is newer: does not need to be built
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now)
        self.touch(self.outputs[0], self.now+1)

        # Add cached dependency information to the mock builder.
        self.mockBuilder.dependencies = self.inputs
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert not needsBuild
        # Should not have called the dependency rule
        assert not self.dependencyRuleCalled

    def testDoesNotNeedsBuildUpdateCachedDependencies(self):
        # The output is newer: does not need to be built
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now)
        self.touch("/foo", self.now)
        self.touch(self.outputs[0], self.now+1)

        # No cached dependency information: It should get updated
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert not needsBuild
        # Should not have called the dependency rule
        assert self.dependencyRuleCalled
        assert len(self.mockBuilder.dependencies) == 2

    def testNeedsBuildUpdateCachedDependencies(self):
        # The output is older: needs build
        self.node = stupidbuild.BuildNode(self.inputs, self.outputs, NullBuildRule, self._dependencyRule, {})
        self.touch(self.inputs[0], self.now)
        self.touch(self.outputs[0], self.now-1)

        # Add cached dependency information to the mock builder.
        self.mockBuilder.dependencies = self.inputs
        needsBuild, pendingDependencies = self.node.needsBuild(self.mockBuilder)
        assert needsBuild
        # Should have called the dependency rule and updated the cache
        assert self.dependencyRuleCalled
        assert len(self.mockBuilder.dependencies) == 2


class BuilderTest(TemporaryDirTest):
    def setUp(self):
        TemporaryDirTest.setUp(self)

        self.input = "/foo.c"
        self.output = "/foo.o"
        self.builder = stupidbuild.StupidBuilder(self.tempdir)
        self.builder.buildTarget([self.input], [self.output], NullBuildRule, None, {})

    def testInputPath(self):
        self.assertEquals(self.tempdir + self.input, self.builder.inputPath(self.input))

        # the input path of an output is in the output directory
        # foo.o -> libfoo.a
        self.builder.buildTarget([self.output], ["/libfoo.a"], NullBuildRule, None, {})
        self.assertEquals(self.tempdir + "/build" + self.output, self.builder.inputPath(self.output))

    def testTargetNode(self):
        # No target not for an input
        self.assertRaises(KeyError, self.builder.targetNode, self.input)

        node = self.builder.targetNode(self.output)
        assert node.inputs[0] == self.input
        assert node.outputs[0] == self.output

    def testBuildTarget(self):
        # Check that buildTarget raises an exception for duplicates
        self.assertRaises(stupidbuild.DuplicateOutputError,
            self.builder.buildTarget, ["/foo2.c"], [self.output], NullBuildRule, None, {})

    def testNeedsBuild(self):
        # No input file: exception
        self.assertRaises(stupidbuild.UnknownInputError,  self.builder.needsBuild, self.input)

        # Input file never needs a build
        self.touch(self.input)
        assert not self.builder.needsBuild(self.input)

    def testCommandsMissingSource(self):
        self.assertRaises(stupidbuild.UnknownInputError,  self.builder.commands, self.output)
    
    def testCommandsSimple(self):
        self.touch(self.input)
        command = self.builder.commands(None)
        assert len(command) == 1
        self.assertEquals(self.tempdir + "/build" + self.output, command[0])

    def testCommandsChain(self):
        # foo.o -> libfoo.a
        intermediateOutput = "/libfoo.a"
        self.builder.buildTarget([self.output], [intermediateOutput], NullBuildRule, None, {})

        # libfoo.a -> foo
        finalOutput = "/foo"
        self.builder.buildTarget([intermediateOutput], [finalOutput], NullBuildRule, None, {})
        self.touch(self.input)
        commands = self.builder.commands(None)

        assert len(commands) == 3
        # The order matters here due to dependencies
        self.assertEquals(self.tempdir + "/build" + self.output, commands[0])
        self.assertEquals(self.tempdir + "/build" + intermediateOutput, commands[1])
        self.assertEquals(self.tempdir + "/build" + finalOutput, commands[2])

    def testCleanFiles(self):
        cleanFiles = list(self.builder.cleanFiles())
        assert len(cleanFiles) == 1
        self.assertEquals(self.builder.outputPath(self.output), cleanFiles[0])

    def testDependsBuildOutput(self):
        self.builder = stupidbuild.StupidBuilder(self.tempdir)

        # foo.proto -> foo.pb.h
        self.builder.buildTarget(["/foo.proto"], ["/foo.pb.h"], NullBuildRule, None, {})

        def DependRule(inputs):
            return inputs + (self.tempdir + "/build/foo.pb.h",)
        self.builder.buildTarget(["/foo.c"], ["/foo.o"], NullBuildRule, DependRule, {})
        self.touch("/foo.proto")
        self.touch("/foo.c")
        commands = self.builder.commands(None)

        self.assertEquals(2, len(commands))
        self.assertEquals(self.tempdir + "/build/foo.pb.h", commands[0])
        self.assertEquals(self.tempdir + "/build" + self.output, commands[1])

    def testCachedDependencies(self):
        # Initially there are no cached dependencies
        self.assertEquals(None, self.builder.getDependencies(self.output))

        # Store cached dependencies
        dependencies = [self.input]
        self.builder.setDependencies(self.output, dependencies)
        self.assertEquals(dependencies, self.builder.getDependencies(self.output))

        # The dependency cache dictionary is accessible
        self.assertEquals(dependencies, self.builder.dependencyCache[self.output])


if __name__ == "__main__":
    unittest.main()
