#!/usr/bin/python

import errno
import os.path
import socket
import sys
import xmlrpclib

import third_party.argparse as argparse

__pychecker__ = "unusednames=host_proxies,args"


class TaskmgrException(BaseException):
    pass


class TaskScheduler(object):
    def __init__(self, hosts):
        self.available_hosts = hosts
        self.reserved_hosts = {}
        self.scheduled_tasks = []

    def reserveHost(self):
        host = self.available_hosts[0]
        del self.available_hosts[0]
        host_name, port = host.split(":")
        port = int(port)
        self.reserved_hosts[host_name] = port
        return host_name

    def schedule(self, command, package):
        host = self.reserveHost()
        self.scheduleHost(host, command, package)
        return host

    def scheduleHost(self, host, command, package):
        port = self.reserved_hosts[host]
        del self.reserved_hosts[host]
        self.scheduled_tasks.append((host, port, command, package))

    def scheduledHostPorts(self):
        for host, port, command, package in self.scheduled_tasks:
            yield host, port


def mapTasksToHosts(host_proxies, tasks):
    # Map one task on each host
    host_tasks = []
    for i, (command, package) in enumerate(tasks):
        host_tasks.append((host_proxies[i], command, package))
    return host_tasks


def start(host_proxies, scheduler, args):
    # TODO: Reduce duplication? This looks a LOT like stop
    args_hosts = {}
    for h in args:
        args_hosts[h] = True

    count = 0
    for host, port, command, package in scheduler.scheduled_tasks:
        if len(args) > 0:
            if host not in args_hosts:
                continue
            else:
                del args_hosts[host]

        print "%s ... " % (host),
        sys.stdout.flush()
        print host_proxies[host].taskStart(command, xmlrpclib.Binary(package))
        count += 1
    print "started %d tasks" % (count)

    if len(args_hosts) > 0:
        raise TaskmgrException("Unused hosts: " + str(args_hosts.keys()))


def status(host_proxies, scheduler, args):
    for host, port, command, package in scheduler.scheduled_tasks:
        tasks = host_proxies[host].taskList()
        for task_id, command, status in tasks:
            print host, task_id, command, status
   

def stop(host_proxies, scheduler, args):
    args_hosts = {}
    for h in args:
        args_hosts[h] = True

    for host, port, command, package in scheduler.scheduled_tasks:
        if len(args) > 0:
            if host not in args_hosts:
                continue
            else:
                del args_hosts[host]

        try:
            tasks = host_proxies[host].taskList()
        except socket.error, e:
            if e.args[0] != errno.ECONNREFUSED:
                raise
            print "WARNING: %s connection refused; skipping" % host
        for task_id, command, status in tasks:
            print "stopping", host, command, "..."
            print host_proxies[host].taskStop(task_id)

    if len(args_hosts) > 0:
        raise TaskmgrException("Unused hosts: " + str(args_hosts.keys()))


def wait(host_proxies, scheduler, args):
    args_hosts = {}
    for h in args:
        args_hosts[h] = True

    for host, port, command, package in scheduler.scheduled_tasks:
        if len(args) > 0:
            if host not in args_hosts:
                continue
            else:
                del args_hosts[host]

        tasks = host_proxies[host].taskList()
        assert len(tasks) == 1
        task_id = tasks[0][0]
        # Bit of a hack: it will "hang" the slave proxy
        code, output = host_proxies[host].taskWait(task_id)
        print output,

    if len(args_hosts) > 0:
        raise TaskmgrException("Unused hosts: " + str(args_hosts.keys()))


def tasks(host_proxies, scheduler, args):
    for host, port, command, package in scheduler.scheduled_tasks:
        print host, command


def parseScript(task_definition, command_hosts, command_port, task_args):
    # Change to the task definition directory
    # TODO: Change back after?
    task_definition_dir = os.path.dirname(task_definition)
    if task_definition_dir != "":
        os.chdir(task_definition_dir)

    # Add the current directory to the python path so imports are found
    pythonpath_before = sys.path
    sys.path = [""] + sys.path

    assert task_definition.endswith(".py")
    module = __import__(os.path.basename(task_definition)[:-3])

    # Restore the python path
    sys.path = pythonpath_before

    # Pull out the hosts from command line options or the script
    hosts = None
    if len(command_hosts) > 0:
        # Combine the hsots and port number
        hosts = ["%s:%d" % (h, command_port) for h in command_hosts]
    elif hasattr(module, "HOST_LIST"):
        # Make a copy so we don't modify the module's variables
        hosts = list(module.HOST_LIST)
    else:
        sys.stderr.write("Error: No slave hosts. Either need HOST_LIST defined in " +
                task_definition + " or specify --slaves\n")
        sys.exit(1)
    assert len(hosts) > 0
    scheduler = TaskScheduler(hosts)

    # Schedule the tasks
    if not hasattr(module, "scheduleTasks"):
        sys.stderr.write("scheduleTasks not defined in " + task_definition + "\n")
        sys.exit(1)
    module.scheduleTasks(scheduler, task_args)

    host_proxies = {}
    for host, port in scheduler.scheduledHostPorts():
        host_proxies[host] = xmlrpclib.ServerProxy("http://%s:%d" % (host, port))
    return host_proxies, scheduler


COMMANDS = (start, status, stop, tasks, wait)


def writeHelpAndExit():
    sys.stderr.write("taskcmd.py [task definition] [command]\n")
    sys.stderr.write("Valid commands:\n")
    for command in COMMANDS:
        sys.stderr.write("\t%s\n" % (command.__name__))
    sys.exit(1)


def argparseList(input):
    value = input.split(",")
    assert len(value) > 0
    return value


def argparseFile(input):
    if not os.path.exists(input):
        raise argparse.ArgumentTypeError("%s does not exist" % (input))
    return input


def main(args):
    parser = argparse.ArgumentParser(
            description = "Start/stop/control tasks across multiple slave hosts."
    )

    # Required positional arguments
    parser.add_argument("task_definition",
            help = "Script defining the tasks to control",
            type = argparseFile)
    parser.add_argument("command",
            help = "Command to execute",
            choices = tuple(c.__name__ for c in COMMANDS))
    parser.add_argument("command_args",
            help = "Optional arguments for the command",
            nargs = "*")

    # Optional arguments
    parser.add_argument("--slaves",
            help = "List of slave machines to control (comma separated)",
            type = argparseList,
            default = [])
    parser.add_argument("--slave_port",
            type = int,
            default = 55555,
            help = "Port for taskmgr on slave machines (default: %(default)d)")
    parser.add_argument("--task_args",
            help = "Arguments for the task definition script",
            nargs = "+",
            default = [])

    options = parser.parse_args()

    # Parse the definition file
    host_proxies, scheduler = parseScript(
            options.task_definition, options.slaves, options.slave_port, options.task_args)

    assert options.command in [c.__name__ for c in COMMANDS]
    for command_function in COMMANDS:
        if command_function.__name__ == options.command:
            command_function(host_proxies, scheduler, options.command_args)
            break


if __name__ == "__main__":
    main(sys.argv)
