#!/usr/bin/python

import cStringIO
import errno
import os
import resource
import shutil
import signal
import sys
from third_party.stdlib import tarfile  # local version
import tempfile
import time

def signalChain(child_pid, signals):
    for i, next_signal in enumerate(signals):
        os.kill(child_pid, next_signal)
        
        # Wait for it to handle the signal
        time.sleep(0.1)

        pid, status = os.waitpid(child_pid, os.WNOHANG)
        if pid > 0:
            # Process has exited: we are done
            assert pid == child_pid
            return status
    return None


def makeTarFromFiles(path_list, dir_pairs={}):
    """Returns the data for a .tar.bz2 containing files in path_list."""

    tar_data = cStringIO.StringIO()
    tar = tarfile.open(mode="w:bz2", fileobj=tar_data)
    for path in path_list:
        tar.add(path, arcname=os.path.basename(path))

    for external_path, tar_path in dir_pairs.iteritems():
        tar.add(external_path, arcname=tar_path)

    tar.close()
    return tar_data.getvalue()


class NamedTemporaryDirectory(object):
    def __init__(self):
        self.name = tempfile.mkdtemp()

    def close(self):
        if self.name is not None:
            shutil.rmtree(self.name)
            self.name = None

    def __del__(self):
        self.close()


class Task(object):
    def __init__(self, command, package_data):
        self.command = command
        self.package = Package(package_data)
        self.output_file = tempfile.NamedTemporaryFile()

        self.child_pid = os.fork()
        if self.child_pid == 0:
            # Child: change to the package directory
            os.chdir(self.package.path())

            # Make us a process group leader: supports killing all children
            # TODO: If the child calls setsid this fails. Fix this.
            os.setsid()

            # Permit unlimited sized core files
            resource.setrlimit(resource.RLIMIT_CORE, (-1, -1))

            # redirect the file descriptors to the output file
            os.dup2(self.output_file.fileno(), sys.stdout.fileno())
            os.dup2(self.output_file.fileno(), sys.stderr.fileno())
            sys.stdin.close()

            # Exec the command
            path = "/bin:/usr/bin"
            env = { "PATH": path }
            os.execvpe(command[0], command, env)
        assert self.child_pid > 0

        self.exit_status = None

    def wait(self):
        if self.poll() is None:
            pid, exit_status = os.waitpid(self.child_pid, 0)
            assert pid == self.child_pid
            self._translateExitStatus(exit_status)
        assert self.exit_status is not None
        return self.exit_status

    def stop(self):
        """Stops the running process and cleans up all the resources."""

        if self.poll() is None:
            # Child still running: gently stop the process
            exit_status = signalChain(self.child_pid, (signal.SIGINT, signal.SIGTERM, signal.SIGKILL))
            if exit_status is not None:
                self._translateExitStatus(exit_status)
        code = self.wait()

        # Send SIGKILL to the entire process group, in case there are children
        # TODO: This won't work if the child 
        try:
            os.killpg(self.child_pid, signal.SIGKILL)
        except OSError, e:
            # Only ignore ESRCH (no such process)
            if e.errno != errno.ESRCH:
                raise

        # Clean up temporary files
        self.package.close()
        self.output_file.close()
        return code

    def output(self):
        f = open(self.output_file.name)
        data = f.read()
        f.close()
        return data

    def status(self):
        code = self.poll()

        if code is None:
            status = "RUNNING"
            msg = ""
        elif code >= 0:
            status = "EXITED"
            if code > 0:
                status = "FAILED"
            msg = "Exit status code: %d" % (code)
        else:
            assert code < 0
            # Build a mapping from signals to names
            signal_map = {}
            # duplicate signals: we ignore the key, using the value instead
            DUP_SIGNALS = {
                    "SIGIOT": "SIGABRT",
                    "SIGCLD": "SIGCHLD",
                    "SIGPOLL": "SIGIO",
            }
            for variable_name in dir(signal):
                if variable_name.startswith("SIG") and variable_name[3].isupper():
                    if (variable_name in DUP_SIGNALS and
                            getattr(signal, variable_name) == getattr(signal, DUP_SIGNALS[variable_name])):
                        # This is a duplicate: ignore it
                        continue

                    if -getattr(signal, variable_name) in signal_map:
                        print "DUP", variable_name, signal_map[-getattr(signal, variable_name)]
                    signal_map[-getattr(signal, variable_name)] = variable_name
            status = "FAILED"
            msg = "Exited due to signal: %s" % (signal_map[code])

        return (status, msg)

    def poll(self):
        if self.exit_status is None:
            pid, status = os.waitpid(self.child_pid, os.WNOHANG)
            if pid != 0:
                assert pid == self.child_pid
                assert status is not None
                self._translateExitStatus(status)
        return self.exit_status

    def _translateExitStatus(self, status):
        assert self.exit_status is None
        if os.WIFEXITED(status):
            self.exit_status = os.WEXITSTATUS(status)
        else:
            assert os.WIFSIGNALED(status)
            self.exit_status = - os.WTERMSIG(status)


class Package(object):
    """Unpackages a set of files into a temporary directory."""

    def __init__(self, data):
        self.directory = NamedTemporaryDirectory()
    
        # Open the tar archive
        in_fileobj = cStringIO.StringIO(data)
        tar = tarfile.open(fileobj=in_fileobj)

        # Validate the members
        for member in tar:
            # Files and directories only
            assert member.isfile() or member.isdir()
            # No file names that might escape the package directory
            assert "../" not in member.name
            assert member.name[0] != "/"

        tar.extractall(path=self.directory.name)

    def path(self):
        return self.directory.name

    def close(self):
        return self.directory.close()


class TaskDescriptor(object):
    def __init__(self, command_line, package_data):
        self.command_line = command_line
        self.package_data = package_data
