#!/usr/bin/python

import cStringIO
import errno
import os.path
import signal
from third_party.stdlib import tarfile  # local version
import time
import unittest

import task


EMPTY_TAR = task.makeTarFromFiles([])


def waitForOk(task):
    while True:
        if task.output() != "":
            assert task.output() == "ok"
            return


class TaskTest(unittest.TestCase):
    def testNoPackage(self):
        t = task.Task(("echo", "foo"), EMPTY_TAR)
        t.wait()
        self.assertEquals("foo\n", t.output())
        self.assertEquals(0, t.stop())

    def testStdoutAndStderr(self):
        package_data = makeTar("one", "ok")

        t = task.Task(("ls", "one", "nosuchfileexists"), package_data)
        t.wait()
        self.assertTrue(t.output().endswith("No such file or directory\none\n"))
        self.assertEquals("FAILED", t.status()[0])
        self.assertTrue(t.status()[1].startswith("Exit status code: "))
        self.assertTrue(t.stop() > 0)

    def testStop(self):
        # A gross embedded script to test handling of signals
        data = """#!/usr/bin/python

import signal
import sys

if sys.argv[1] == "sigint" or sys.argv[1] == "sigterm":
    signal.signal(signal.SIGINT, signal.SIG_IGN)
else:
    # SIGINT will kill the process rather than raising KeyboardInterrupt
    signal.signal(signal.SIGINT, signal.SIG_DFL)

if sys.argv[1] == "sigterm":
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

# Signal to the parent we are running
sys.stdout.write("ok")
sys.stdout.flush()
raw_input()
sys.exit(0)"""
        package_data = makeTar("binary.py", data, mode=0555)

        t = task.Task(("./binary.py", "none"), package_data)
        waitForOk(t)
        self.assertEquals(("RUNNING", ""), t.status())
        self.assertEquals(-signal.SIGINT, t.stop())
        self.assertEquals(("FAILED", "Exited due to signal: SIGINT"), t.status())

        t = task.Task(("./binary.py", "sigint"), package_data)
        waitForOk(t)
        self.assertEquals(-signal.SIGTERM, t.stop())

        t = task.Task(("./binary.py", "sigterm"), package_data)
        waitForOk(t)
        self.assertEquals(-signal.SIGKILL, t.stop())
        
    def testStopAllChildren(self):
        # Start a task that has a child process
        t = task.Task(("sh", "-c",
                "python -c 'import time; time.sleep(60)' & echo $! ; wait"), EMPTY_TAR)

        # Wait for the child PID
        while t.output() == "":
            pass
        child_pid = int(t.output())

        code = t.stop()
        # On Mac OS X, sh catches SIGINT and exits with code 129
        # On Linux, sh dies with SIGINT
        self.assertTrue(code == -signal.SIGINT or code == 129)

        # Verify the child no longer exists (after a short pause: gross!)
        time.sleep(0.01)
        try:
            os.kill(child_pid, 0)
            self.fail("Should have raised OSError: child still running?")
        except OSError, e:
            assert e.errno == errno.ESRCH

    def testCoreFile(self):
        if os.uname()[0] == "Darwin":
            # Mac OS X doesn't support core files in the same way Linux does
            return

        # Start a task that crashes
        t = task.Task(("python", "-c",
                "import os; import signal; os.kill(os.getpid(), signal.SIGSEGV)"), EMPTY_TAR)

        self.assertEquals(-signal.SIGSEGV, t.wait())
        self.assertTrue(os.path.exists(t.package.path() + "/core"))

    def testAbortStatus(self):
        t = task.Task(("python", "-c", "import os; os.abort()"), EMPTY_TAR)
        t.wait()
        status, message = t.status()
        self.assertEquals("FAILED", status)
        self.assertEquals("Exited due to signal: SIGABRT", message)


def makeTarInfo(name, mode):
    info = tarfile.TarInfo()
    info.name = name
    info.size = 0
    info.mtime = time.time()
    info.mode = mode
    info.type = None
    info.uid = 1000
    info.gid = 1000
    info.uname = "user"
    info.gname = "group"
    return info


def makeTarFile(tar, file_name, file_data, mode):
    info = makeTarInfo(file_name, mode)
    info.size = len(file_data)
    info.type = tarfile.REGTYPE

    in_fileobj = cStringIO.StringIO(file_data)
    tar.addfile(info, in_fileobj)


def makeTarSymlink(tar, target, link_name):
    info = makeTarInfo(link_name, 0666)
    info.type = tarfile.SYMTYPE
    info.linkname = target
    tar.addfile(info)


def makeTarDirectory(tar, dir_name):
    info = makeTarInfo(dir_name, 0777)
    info.type = tarfile.DIRTYPE
    tar.addfile(info)


def makeTar(file_name, file_data, mode=0666):
    """Returns the string representing a tar archive containing file_name storing file_data."""
    out = cStringIO.StringIO()
    tar = tarfile.open(mode="w", fileobj=out)
    makeTarFile(tar, file_name, file_data, mode)
    tar.close()

    return out.getvalue()


class PackageTest(unittest.TestCase):
    def testSimple(self):
        tardata = makeTar("one", "ok")

        p = task.Package(tardata)
        path = p.path()
        self.assertTrue(os.path.exists(path))

        f = open(p.path() + "/one")
        data = f.read()
        f.close()
        self.assertEquals("ok", data)

        p.close()
        self.assertEquals(None, p.path())
        self.assertFalse(os.path.exists(path))

    def testBadType(self):
        out = cStringIO.StringIO()
        tar = tarfile.open(mode="w", fileobj=out)
        makeTarSymlink(tar, "/target", "symlink")
        tar.close()
        self.assertRaises(AssertionError, task.Package, out.getvalue())

    def testBadPaths(self):
        # Paths that escape the package are bad. Basically you can't use ..
        tardata = makeTar("../one", "ok")
        self.assertRaises(AssertionError, task.Package, tardata)
        tardata = makeTar("dir/../../one", "ok")
        self.assertRaises(AssertionError, task.Package, tardata)

        # Absolute paths are also bad
        tardata = makeTar("/one", "ok")
        self.assertRaises(AssertionError, task.Package, tardata)

    def testDirectory(self):
        # Directories are okay
        out = cStringIO.StringIO()
        tar = tarfile.open(mode="w", fileobj=out)
        makeTarDirectory(tar, "foo")
        tar.close()
        p = task.Package(out.getvalue())
        self.assertTrue(os.path.isdir(os.path.join(p.path(), "foo")))

    def testEmpty(self):
        p = task.Package(EMPTY_TAR)
        self.assertTrue(os.path.exists(p.path()))

if __name__ == "__main__":
    unittest.main()
