#!/usr/bin/python

import os.path
import tempfile

import taskmgr.task

__pychecker__ = 'unusednames=partition_index,replica_index'

PARTITIONS = 2
REPLICAS = 2
PORT = 12345


SOURCE_PATH = "../hstore"
LOCK_SOURCE_PATH = "../hstore-lock"

MODES = ("block", "spec", "lock")


def makePartitions(scheduler):
    partitions = []
    for partition_index in xrange(PARTITIONS):
        partition_hosts = []
        for replica_index in xrange(REPLICAS):
            partition_hosts.append(scheduler.reserveHost())
        partitions.append(partition_hosts)
    return partitions


def makeConfiguration(partition_lists):
    f = tempfile.NamedTemporaryFile()
    for partition in partition_lists:
        f.write("partition function\n")
        for host in partition:
            f.write("%s:%d\n" % (host, PORT))
        f.write("\n")
    f.flush()
    return f


def serverCommand(server_binary, mode, extra_args, config_tempfile, partition_index, replica_index):
    config_path = os.path.basename(config_tempfile.name)
    return (("taskset", "--cpu-list", "0", "./" + server_binary, mode) +
            extra_args +
            (config_path, str(partition_index), str(replica_index)))


def scheduleServersDistributors(
            scheduler, partitions, server_binary, mode, distributor_extra_args, distributor_binary,
            config):
    # Choose the correct htableserver version: locking is a previous version
    if mode == "lock":
        htableserver_source = LOCK_SOURCE_PATH
    else:
        htableserver_source = SOURCE_PATH
    server_package = taskmgr.task.makeTarFromFiles([
            htableserver_source + "/build/" + server_binary, config.name])

    # Schedule backups first, so they are running when we start the primaries
    PARTITION_EXTRA_ARGS = ()
    for partition_index in xrange(PARTITIONS):
        for replica_index in xrange(1, REPLICAS):
            command = serverCommand(
                    server_binary, mode, PARTITION_EXTRA_ARGS, config,
                    partition_index, replica_index)
            scheduler.scheduleHost(
                    partitions[partition_index][replica_index], command, server_package)

    # Schedule primaries
    for partition_index in xrange(PARTITIONS):
        command = serverCommand(
                    server_binary, mode, PARTITION_EXTRA_ARGS, config, partition_index, 0)
        scheduler.scheduleHost(partitions[partition_index][0], command, server_package)

    # Schedule distributor
    distributor_package = taskmgr.task.makeTarFromFiles([
        SOURCE_PATH + "/build/" + distributor_binary,
        config.name])
    distributor_command = ("taskset", "--cpu-list", "0", "./" + distributor_binary,
            str(PORT), os.path.basename(config.name)) + distributor_extra_args
    distributor_host = scheduler.schedule(distributor_command, distributor_package)
    return distributor_host


def scheduleHTable(scheduler, args, partitions, config):
    if len(args) != 1 and len(args) != 2:
        raise Exception("task_mptxn.py requires one or two arguments: mode [general transaction]")
    mode = args[0]

    distributor_extra_args = ()
    general_transactions = None
    if len(args) == 2:
        general_transactions = args[1]
        assert general_transactions == "general"
        distributor_extra_args = (general_transactions,)
    if mode not in MODES:
        raise Exception("mode '%s' is not valid, must be one of %s" % (mode, ", ".join(MODES)))

    return scheduleServersDistributors(
            scheduler, partitions, "htableserver", mode,
            distributor_extra_args, "htabledistributor", config)


def scheduleAllServers(scheduler, args):
    # Reserve all the hosts for the servers
    partitions = makePartitions(scheduler)

    # Write the config file
    config = makeConfiguration(partitions)

    distributor_host = scheduleHTable(scheduler, args, partitions, config)
    mode = args[0]
    return mode, config, distributor_host


def scheduleClient(scheduler, client_command, config):
    client_package = taskmgr.task.makeTarFromFiles([
        #~ "build/htableclientperf-lock",
        #~ "scale_mptxns.py",
        #~ "htable-rep-debug/htableclientperf-occ",
        SOURCE_PATH + "/build/htableclientperf",
        LOCK_SOURCE_PATH + "/build/htableclientperf-lock",
        SOURCE_PATH + "/perfscripts/scale_mptxns.py",
        SOURCE_PATH + "/perfscripts/scale_mptxns_conflict.py",
        config.name], {
        SOURCE_PATH + "/third_party/__init__.py": "third_party/__init__.py",
        SOURCE_PATH + "/third_party/pystdlib": "third_party/pystdlib",
    })
    #~ client_command = ("./scale_clients.py", os.path.basename(config.name), "%s:12345" % partitions[0][0])
    #~ client_command = ("./scale_mptxns_conflict.py", os.path.basename(config.name))
    client_host = scheduler.schedule(client_command, client_package)
    return client_host


def scheduleTasks(scheduler, args):
    mode, config, distributor_host = scheduleAllServers(scheduler, args)
    client_command = ("./scale_mptxns.py", mode, os.path.basename(config.name),
            "%s:%d" % (distributor_host, PORT))
    scheduleClient(scheduler, client_command, config)
